// Title:   BadNormalRemover.cpp
// Created: Thu Sep 25 15:16:57 2003
// Authors: Simon Heinzle, Tim Weyrich
//
// Copyright (c) 2001, 2002, 2003, Computer Graphics Lab, ETH Zurich
//
// This file is part of the Pointshop3D system.
// See http://www.pointshop3d.com/ for more information.
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later
// version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
// 
// You should have received a copy of the GNU Lesser General
// Public License along with this library; if not, write to the
// Free Software Foundation, Inc., 59 Temple Place, Suite 330,
// Boston, MA 02111-1307 USA
// 
// Contact info@pointshop3d.com if any conditions of this
// licensing are not clear to you.
//

#ifdef _WIN32
	// turn off std::map warnings
	//# pragma warning(push)
	# pragma warning(disable:4786) 
#endif

#include "BadNormalRemover.h"

#include <set>
#include <vector>
#include <qapplication.h>
#include <qvector.h>

#include "../../../../Core/Scene/src/Scene.h"
#include "../../../../Core/DataStructures/src/PriorityQueue.h"
#include "../../../../Core/DataStructures/src/kdTree.h"

#include "Selector.h"

#include "NewOverload.h"

BadNormalRemover::BadNormalRemover() {
	_neighbour_count = 20;	
}

void BadNormalRemover::setMaxCost(float maxcost) {
	_max_cost = maxcost;
}

void BadNormalRemover::setStartPoint(unsigned int index) {
	_start_index = index;
}

void BadNormalRemover::setNeighbours(unsigned int neighbours) {
	_neighbour_count = neighbours;
}

void BadNormalRemover::execute() {
	QApplication::setOverrideCursor(Qt::waitCursor);

	MinPriorityQueue<int, float> pqueue;
	std::set<int> accepted_surfels;
	std::set<int> new_neighbours;
	std::set<int>::iterator new_iterator;
	
	KdTree *kdtree;
	Selector *selector;

	Vector3D accepted_normal;
	Vector3D accepted_position;
	Vector3D test_normal;
	Vector3D test_position;

	float cost;
	int found_index;
	int index;
	bool test_bit;
	
	
	selector = Selector::getInstance();
	kdtree = selector->returnKdTree();
	new_neighbours.insert(_start_index);
	
	pqueue.init();
	pqueue.setSize(selector->getNOfSurfels());

	while (!new_neighbours.empty()) {
		// fill accepted surfels
		// doesn't work -> accepted_surfels.insert(new_neighbours.begin(), new_neighbours.end());
		for (new_iterator = new_neighbours.begin(); new_iterator != new_neighbours.end(); new_iterator++)
			accepted_surfels.insert(*new_iterator);
				
		// check for all new neighbours their neighbours.
		for (new_iterator = new_neighbours.begin(); new_iterator != new_neighbours.end(); new_iterator++) {
			// fetch current one.
			index = *new_iterator;
			test_bit = selector->getSurfelWorldPosition(index,accepted_position);
			test_bit &= selector->getSurfelWorldNormal(index, accepted_normal);
			if (!test_bit)
				continue;

			kdtree->setNOfNeighbours(_neighbour_count);
			kdtree->queryPosition(accepted_position);

			for (int i=1; i<_neighbour_count; i++) {
				found_index = kdtree->getNeighbourPositionIndex(i);
				
				if (accepted_surfels.find(found_index) == accepted_surfels.end() && new_neighbours.find(found_index) == new_neighbours.end()) {
					// found index yet no valid surfel
					test_bit = selector->getSurfelWorldPosition(found_index, test_position);
					test_bit &= selector->getSurfelWorldNormal(found_index, test_normal);
					if (!test_bit)
						continue;
					
					cost = 1.0 - Vector3D::dotProduct(accepted_normal, test_normal);
					pqueue.insert(found_index, cost);
				}
			}
		}
		
		// reset the new neighbours.
		new_neighbours.clear();
		cost = pqueue.getMinWeight();
		index = pqueue.getMinIndex();
		while ( (pqueue.getNofElements() > 0) && (cost < _max_cost)  ) {
			pqueue.removeMin();	
			new_neighbours.insert(index);

			cost = pqueue.getMinWeight();
			index = pqueue.getMinIndex();
		}
	}

	// delete remaining elements in pqueue
	QVector<SurfelInterface> del_me;

	std::vector<unsigned int> delete_this;

	for (int i=0; i<selector->getNOfSurfels(); i++) {
		if (accepted_surfels.find(i) == accepted_surfels.end()) {
			delete_this.push_back(i);
		}
	}
	qDebug("deleting %d surfels",delete_this.size());
	selector->deleteSurfels(delete_this);

	QApplication::restoreOverrideCursor();

}

UndoAction* BadNormalRemover::executeUndo() {
	QApplication::setOverrideCursor(Qt::waitCursor);

	MinPriorityQueue<int, float> pqueue;
	std::set<int> accepted_surfels;
	std::set<int> new_neighbours;
	std::set<int>::iterator new_iterator;

	KdTree *kdtree;
	Selector *selector;

	Vector3D accepted_normal;
	Vector3D accepted_position;
	Vector3D test_normal;
	Vector3D test_position;

	float cost;
	int found_index;
	int index;
	bool test_bit;
	
	selector = Selector::getInstance();
	kdtree = selector->returnKdTree();
	new_neighbours.insert(_start_index);
	
	pqueue.init();
	pqueue.setSize(selector->getNOfSurfels());

	while (!new_neighbours.empty()) {
		// fill accepted surfels
		for (new_iterator = new_neighbours.begin(); new_iterator != new_neighbours.end(); new_iterator++)
			accepted_surfels.insert(*new_iterator);
				
		// check for all new neighbours their neighbours.
		for (new_iterator = new_neighbours.begin(); new_iterator != new_neighbours.end(); new_iterator++) {
			// fetch current one.
			index = *new_iterator;
			test_bit = selector->getSurfelWorldPosition(index,accepted_position);
			test_bit &= selector->getSurfelWorldNormal(index, accepted_normal);
			if (!test_bit)
				continue;

			kdtree->setNOfNeighbours(_neighbour_count);
			kdtree->queryPosition(accepted_position);

			for (int i=1; i<_neighbour_count; i++) {
				found_index = kdtree->getNeighbourPositionIndex(i);
				
				if (accepted_surfels.find(found_index) == accepted_surfels.end() && new_neighbours.find(found_index) == new_neighbours.end()) {
					// found index yet no valid surfel
					test_bit = selector->getSurfelWorldPosition(found_index, test_position);
					test_bit &= selector->getSurfelWorldNormal(found_index, test_normal);
					if (!test_bit)
						continue;
					
					cost = 1.0 - Vector3D::dotProduct(accepted_normal, test_normal);
					pqueue.insert(found_index, cost);
				}
			}
		}
		
		// reset the new neighbours.
		new_neighbours.clear();
		cost = pqueue.getMinWeight();
		index = pqueue.getMinIndex();
		while ( (pqueue.getNofElements() > 0) && (cost < _max_cost)  ) {
			pqueue.removeMin();	
			new_neighbours.insert(index);

			cost = pqueue.getMinWeight();
			index = pqueue.getMinIndex();
		}
	}

	// delete remaining elements in pqueue
	QVector<SurfelInterface> del_me;
		
	std::vector<unsigned int> delete_this;

	for (int i=0; i<selector->getNOfSurfels(); i++) {
		if (accepted_surfels.find(i) == accepted_surfels.end()) {
			delete_this.push_back(i);
		}
	}
	qDebug("deleting %d surfels",delete_this.size());
	
	UndoAction *undo;
	undo = selector->deleteSurfelsUndo(delete_this);

	QApplication::restoreOverrideCursor();

	return undo;
}

// Some Emacs-Hints -- please don't remove:
//
//  Local Variables:
//  mode:C++
//  tab-width:4
//  End:
