// Title:   BoxSelector.h
// Created: Thu Sep 25 15:16:52 2003
// Authors: Simon Heinzle, Tim Weyrich
//
// Copyright (c) 2001, 2002, 2003, Computer Graphics Lab, ETH Zurich
//
// This file is part of the Pointshop3D system.
// See http://www.pointshop3d.com/ for more information.
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later
// version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
// 
// You should have received a copy of the GNU Lesser General
// Public License along with this library; if not, write to the
// Free Software Foundation, Inc., 59 Temple Place, Suite 330,
// Boston, MA 02111-1307 USA
// 
// Contact info@pointshop3d.com if any conditions of this
// licensing are not clear to you.
//

#ifndef __BOXSELECTOR_H
#define __BOXSELECTOR_H

#ifdef _WIN32
	// turn off std::map warnings
	# pragma warning(push)
	# pragma warning(disable:4786) // identifier truncated to 255 char
#endif

// constant definitions

// selector - shape
#define SHAPE_BOX 50
#define SHAPE_SPHERE 51

// mouse & keyboard button constants
#define LEFTBUTTON 1
#define MIDBUTTON 2
#define RIGHTBUTTON 4
#define ALTBUTTON 8
#define SHIFTBUTTON 16
#define CTRLBUTTON 32

// box alignment mode constants
#define ALIGN_MEDIAN 1
#define ALIGN_MEDIUM 2
#define ALIGN_LOWEST 3

// box movement mode constants
#define MOUSE_PICKING 1
#define MOUSE_ALIGN 2
#define MOUSE_FREEMOVE 3

// selection constants
#define EDGE 99
#define VERT 200
#define SIDE 300

// rotation constants (also selection constans)
#define XROT 100
#define YROT 101
#define ZROT 102

#include <qgl.h>
#include <gsl/gsl_blas.h>
#include <gsl/gsl_eigen.h>
#include "../../../../DataTypes/src/Vector3D.h"
#include "../../../../Utilities/src/Matrix.h"
#include "cQuaternion.h"

/**
 * Box Selector
 * This class provides a box/sphere that can be moved around.
 * The box works in OpenGL preview renderer only.
 * 
 * It has 3 modi
 *  1) Picking mode:   box doesn't move, by clicking on a side you can resize it along this side's normal.
 *                     By clicking on an edge you can rotate the box around the axis parallel to the edge
 *  2) Align mode:     box aligns itself to the surface of the object using the depth values in the z-buffer.
 *                     3 sub modi: a. uses median z value
 *                                 b. uses medium of z values
 *                                 c. uses lowest z value
 *  3) Free move mode: box moves parallel to screen plane. By holding mid mouse button it moves orthogonal to screen plane.
 *
 * In modus 2 & 3 holding the right mouse button resizes the box.
 * Clicking the left mouse button selects surfels inside the box.
 *
 * The box has 2 shapes:
 *   1) box shape
 *   2) sphere shape
 *
 *
 * The following constants are defined:
 *
 * - Selector Shape Constants: 
 *      SHAPE_BOX.....our selector has the shape of a box
 *      SHAPE_SPHERE..our selector has the shape of a sphere
 * - Mouse Button & Keyboard Constants: (this constants can be OR'ed together)
 *      LEFTBUTTON...left mouse button
 *      MIDBUTTON....middle mouse button
 *      RIGHTBUTTON..right mouse button
 *      ALTBUTTON....keyboard alt key
 *      SHIFTBUTTON..keyboard shift key
 *      CTRLBUTTON...keyboard control key
 * - Box Alignment Mode Constants:
 *      ALIGN_MEDIAN..uses median z buffer value for alignment
 *      ALIGN_MEDIUM..uses medium of z buffer values for alignment
 *      ALIGN_LOWEST..uses lowest z buffer value
 * - Box Movement Constants:
 *      MOUSE_PICKING...box doesn't move, by clicking on certain box elements the box can be resized / rotated
 *      MOUSE_ALIGN.....box aligns itself to the object
 *      MOUSE_FREEMOVE..box can be moved independent of object but parallel/orthogonal to viewing plane
 * - Selection Constants (part 1): (usually not needed for user)   
 *      EDGE...selects an edge (between EDGE and EDGE+12)
 *      VERT...selects a vertex (between VERT and VERT+8)
 *      SIDE...selects a side (between SIDE and SIDE+6)
 * - Rotation and Selection Constants (part 2): (usually only needed for rotation)
 *      XROT..an edge parallel to the normal of the x side
 *      YROT..an edge parallel to the normal of the y side
 *      ZROT..an edge parallel to the normal of the z side
 *      
 *      
 * @author Simon Heinzle
 * @version 1.0
 */
class BoxSelector {
private:		
	// data how to paint
	Vector3D _the_box[8];
	int _the_faces[6][4];
	Vector3D _the_face_normals[6];
	int _the_x_lines[4][2];
	int _the_y_lines[4][2];
	int _the_z_lines[4][2];
	// current normals
	cQuaternion _x_normal;
	cQuaternion _y_normal;
	cQuaternion _z_normal;
	// current rotation matrix
	float _rotation_matrix[4][4];
	
	// current selection
	int _selection;
	// current shape
	int _shape;
	// sphere quadric
	GLUquadricObj *_quadric_sphere;


	Vector3D _center;
	Vector3D _edgelength;
	bool _box_position_valid; // is current position valid position?

	// color constants
	float _color_normal[4];
	float _color_normal_transparent[4];
	float _color_selected[4];

	// mouse movement
	Vector3D _old_mouse_projection;
	int _old_mouse_x, _old_mouse_y;
	bool _freeze_box;
	// to restore the camera
	MyDataTypes::TransformationMatrix16f _old_scene_matrix;
	bool _old_scene_set;

	// mouse modes
	bool _picking_mode; // internal flag for rendering apropriate cube
	int _mouse_mode;

	// selection related variables
	float _selection_index;
	bool _aa_box; // is box axis aligned
	
	// align mode related variables
	int _align_mode;
	float _align_rect[25];

	// ******************************
	// * private painting functions *
	// *                            *
	inline void paintVertex(Vector3D mid, float size = 0.2);
	inline void paintVertices(float size = 0.2);
	inline void paintLines(float thickness = 0.1);
	inline void paintFaces();
	inline void paintSphere();
	
	// ***************************
	// * private align functions *
	// *                         *
	inline void mapPosToMouse(void);
	inline void getAlignRect(void);
	inline float alignUseMedian(); // use only the middle 5 values
	inline float alignUseMedium(); // use only the middle 5 values
	inline float alignUseLowest(); // using the lowest value !=0
	
	// ***************************
	// * private resize function *
	inline void resizeAfterMouseMovement(int len);

public:
	/**
	 * Constructor, center(0,0,0), edge_length(100,100,100)
	 */
	BoxSelector();
	/**
	 * Constructor with user defined center <code>Vector3D init_center</code> and edge length vector<code>Vector3D init_edge_length</code>.
	 */
	BoxSelector(Vector3D init_center , Vector3D init_edge_length);
	/**
	 * Destructor
	 */
	~BoxSelector();
	/**
	 * Paints the box
	 */
	void paintBox();
	/**
	 * Paints the box with names defined. This is used in picking mode when a mouse button is pressed.
	 */
	void paintBoxSelect();
	/**
	 * Resizes all edge lengths uniformly by adding <code>float additional_length</code>.
	 */
	void resize(float additional_length = 0.0);
	/**
	 * Resizes edge lengths by adding <code>Vector3D additional_length</code>.
	 */
	void resize(Vector3D additional_length);
	/**
	 * Rotates the box by <code>float angle</code> in degree around the axis specified with <code>int axis</code>.
	 * <code>int axis</code> can contain either XROT, YROT or ZROT, the x,y or z axis of the box.
	 */
	void rotate(float angle, int axis);
	/**
	 * Translates the box by <code>Vector3D translation</code>.
	 */
	void translate(Vector3D translation);
	/**
	 * Translates the box along it's face normals, where the x,y and z component of <code>Vector3D translation</code>
	 * define the the translation along the x, y and respectively z face normals.
	 */
	void translateAlongFaceNormals(Vector3D translation);
	/**
	 * Resets the box in the constructor state.
	 */
	void reset(void);
	/**
	 * Gets the normal of the x face
	 *
	 * @return <code>cQuaternion</code>defining the normal of the x face
	 */
	cQuaternion getX(void);
	/**
	 * Gets the normal of the y face
	 *
	 * @return <code>cQuaternion</code>defining the normal of the y face
	 */
	cQuaternion getY(void);
	/**
	 * Gets the normal of the z face
	 *
	 * @return <code>cQuaternion</code>defining the normal of the z face
	 */
	cQuaternion getZ(void);
	
	/**
	 * Passes the x,y coordinates of the mouse click to the box to determine the underlying element
	 * of the box.
	 *
	 * @return <code>int</code>defining the picked element. The selection constants define the element. 0 is no selection.
	 */
	int getMouseSelection(int x, int y);
	/** 
	 * Sets the current selection to <code>int sel</code>. <code>int sel</code> must be one of the selection constants.
	 */
	void setSelection(int sel);
	/** 
	 * Gets the current selection.
	 *
	 * @return current selection, which is one of the selection constants or 0 for no selection.
	 */
	int getSelection(void);
	/** 
	 * Deselects current selection.
	 */
	void deselect(void);
	/**
	 * Gets the current edge length
	 *
	 * @return <code>Vector3D</code> defining the 3 edge lengths.
	 */
	Vector3D getEdgeLength(void);
	/**
	 * Sets the new edge length to <code>Vector3D newlen</code>.
	 */
	void setEdgeLength(Vector3D newlen);
	/**
	 * Gets the current center.
	 *
	 * @return <code>Vector3D</code> defining the current center.
	 */
	Vector3D getCenter(void);
	/**
	 * Sets the new center to <code>Vector3D newcen</code>.
	 */
	void setCenter(Vector3D newcen);
	/** 
	 * Sets the new shape to <code>int new_shape</code>, either SHAPE_BOX or SHAPE_SPHERE.
	 */
	void setShape(int new_shape);	
	/**
	 * Get current shape.
	 *
	 * @return <code>int</code> defining the shape of the box, either SHAPE_BOX or SHAPE_SPHERE.
	 */
	int getShape(void);
	/**
	 * Gets the current freeze state.
	 *
	 * @return true if box cannot be moved, false otherwise.
	 */
	bool isFreezed();
	/**
	 * Sets the current moude. <code>int mode</mode> must be either MOUSE_PICKING, MOUSE_ALIGN or MOUSE_FREEMOVE.
	 */
	void setMouseMode(int mode);
	/**
	 * Handles a mouse press, <code>int button</code> is one of the mouse button constants.
	 */
	void handleMousePress(int x, int y, int button);
	/**
	 * Handles a mouse release, <code>int button</code> is one of the mouse button constants.
	 */
	void handleMouseRelease(int x, int y, int button);
	/**
	 * Handles a mouse movement, <code>int button</code> is one of the mouse button constants.
	 */
	void handleMouseMovementMidpoint(int x, int y, int button);
	/**
	 * Handles a key press, <code>int key</code> defined by the Qt::Key_**** constants.
	 */
	void handleKeyPressed(int key);
	/**
	 * Handles a key release, <code>int key</code> defined by the Qt::Key_**** constants.
	 */
	void handleKeyReleased(int key);
	/**
	 * Gets the current transformation matrix in GSL form.
	 *
	 * @return <code>gsl_matrix*</code> containing the object transformation.
	 */
	gsl_matrix* getTransformationMatrixGSL();
	/** 
	 * Aligns mouse position to nearest surfel position
	 */
	void alignToNearest();
	/**
	 * Aligns mouse to object, according to z-buffer values.
	 */
	void alignToScreenPos();
	/**
	 * Sets current align mode.
	 * <code>int mode</code> must be either ALIGN_MEDIAN, ALIGN_MEDIUM or ALIGN_LOWEST 
	 */
	void setAlignMode(int mode);
	/**
	 * Selects the surfels inside the box & marks them according to the <code>int selection_index</code>
	 */
	inline void selectBox(int selection_index);
	/**
	 * Sets the current selection index to <code>int selection_index</code>
	 */
	void setSelectionIndex(int selection_index); 
	/**
	 * Sets the colors of the box.
	 * <code>float color[4]</code> contains the normal colors, color[0..2] used for opaque and color [0..3] 
	 * used for transparent elements.
	 * <code>float selection_color[4]</code> is used to highlight the current element on which the user clicked on.
	 */
	void setColors(float color[4], float selection_color[4]);
};

#ifdef _WIN32
	# pragma warning(pop)
#endif

#endif 
//__SCANTOOLSGLEXTENSIONS_H

// Some Emacs-Hints -- please don't remove:
//
//  Local Variables:
//  mode:C++
//  tab-width:4
//  End:
