// Title:   ColorModifier.cpp
// Created: Thu Sep 25 15:16:51 2003
// Authors: Simon Heinzle, Tim Weyrich
//
// Copyright (c) 2001, 2002, 2003, Computer Graphics Lab, ETH Zurich
//
// This file is part of the Pointshop3D system.
// See http://www.pointshop3d.com/ for more information.
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later
// version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
// 
// You should have received a copy of the GNU Lesser General
// Public License along with this library; if not, write to the
// Free Software Foundation, Inc., 59 Temple Place, Suite 330,
// Boston, MA 02111-1307 USA
// 
// Contact info@pointshop3d.com if any conditions of this
// licensing are not clear to you.
//

#include "ColorModifier.h"

#include <qapplication.h>

#include "UndoRedoSpecialisations.h"
#include "Selector.h"


#ifdef _WIN32
	// turn off std::map warnings
	//# pragma warning(push)
	# pragma warning(disable:4786) 
#endif

#include "NewOverload.h"

ColorModifier::ColorModifier() {
	QColor tmp;

	tmp.setRgb(200,200,200);
	_diffuse = tmp.rgb();
	_specular = tmp.rgb();

	_knext = 10;
	_action_mode = COLOR_UNI;
}

void ColorModifier::setParamsFilter(unsigned int knext, float coeff) {
	_coeff = coeff * coeff;
	_knext = knext;
}

void ColorModifier::setParamsUni(QRgb diffuse, QRgb specular) {
	_diffuse = diffuse;
	_specular = specular;
}

void ColorModifier::setAction(int action_mode) {
	_action_mode = action_mode;
}

void ColorModifier::execute() {
	if (_action_mode == COLOR_UNI) {
		QApplication::setOverrideCursor (Qt::waitCursor);
		
		std::vector<unsigned int> *selection;
		Selector *selector;
		SurfelInterface *surfel;
		Scene *scene;

		scene = Scene::getInstance();
		selector = Selector::getInstance();
		selection = selector->getSelectionIndicesStdVec(Selector::intToFlag(_selection_index));
		if (selection->size() == 0) {
			QApplication::restoreOverrideCursor();
			_MCdelete((unsigned long)selection);
			delete selection;
			return;
		}

		for (unsigned int i=0; i<selection->size(); i++) {
			surfel = selector->getSurfel((*selection)[i]);
			scene->setSurfelDiffuseColor(surfel,_diffuse, false);
			scene->setSurfelSpecularColor(surfel, _specular, false);
		}

		scene->emitObjectModifiedSignal();
		scene->emitRenderFrameSignal();
		
		_MCdelete((unsigned long)selection);
		delete selection;
		

		QApplication::restoreOverrideCursor();
	} else if (_action_mode == COLOR_FILTER) {
		QApplication::setOverrideCursor(Qt::waitCursor);

		std::vector<unsigned int> *selection;
		std::vector<QRgb> diffuse;
		std::vector<QRgb> specular;
		Selector *selector;
		SurfelInterface *surfel;
		Scene *scene;
		KdTree *kdtree;
		Vector3D position;
		QColor tmpcol;
		float dr,dg,db;
		float sr,sg,sb;
		float weight;
		float total_weight;
		float distance;
		unsigned int index;

		scene = Scene::getInstance();
		selector = Selector::getInstance();
		selection = selector->getSelectionIndicesStdVec(Selector::intToFlag(_selection_index));
		kdtree = selector->returnKdTree();

		if (selection->size() == 0) {
			QApplication::restoreOverrideCursor();
			_MCdelete((unsigned long)selection);
			delete selection;
			return;
		}

		kdtree->setNOfNeighbours(_knext);
		for (unsigned int i=0; i<selection->size(); i++) {
			position = selector->getSurfelWorldPosition((*selection)[i]); 
			kdtree->queryPosition(position);
			
			dr = dg = db = 0.0f;
			sr = sg = sb = 0.0f;
			total_weight = 0.0f;

			for (unsigned int j=1; j<_knext; j++) {
				index = kdtree->getNeighbourPositionIndex(j);
				surfel = selector->getSurfel(index);
				
				if (_coeff != 0.0f) {
					distance = Vector3D::distance(position, selector->getSurfelWorldPosition(index));
					weight = exp (distance * distance / _coeff);
				} else {
					weight = 1.f;
				}

				total_weight += weight;
				
				tmpcol = surfel->getDiffuseColor();
				dr += tmpcol.red() * weight;
				dg += tmpcol.green() * weight;
				db += tmpcol.blue() * weight;

				tmpcol = surfel->getSpecularColor();
				sr += tmpcol.red() * weight;
				sg += tmpcol.green() * weight;
				sb += tmpcol.blue() * weight;
			}

			dr /= total_weight;
			dg /= total_weight;
			db /= total_weight;
			sr /= total_weight;
			sg /= total_weight;
			sb /= total_weight;
			
			surfel = selector->getSurfel((*selection)[i]);
			tmpcol.setRgb(dr,dg,db);
			scene->setSurfelDiffuseColor(surfel,tmpcol.rgb(),false);
			tmpcol.setRgb(sr,sg,sb);
			scene->setSurfelSpecularColor(surfel,tmpcol.rgb(),false);
		}
		scene->emitObjectModifiedSignal();
		scene->emitRenderFrameSignal();
		
		_MCdelete((unsigned long)selection);
		delete selection;
		
		QApplication::restoreOverrideCursor();
	}
}

UndoAction* ColorModifier::executeUndo() {
	std::vector<unsigned int> *selection;
	Selector *selector;
	SurfelInterface *surfel;
	std::vector<SurfelInterface*> sfls;
	std::vector<QRgb> diff;
	std::vector<QRgb> spec;

	selector = Selector::getInstance();
	selection = selector->getSelectionIndicesStdVec(Selector::intToFlag(_selection_index));
		
	for (unsigned int i=0; i<selection->size(); i++) {
		surfel = selector->getSurfel((*selection)[i]);
		sfls.push_back(surfel);
		diff.push_back(surfel->getDiffuseColor());
		spec.push_back(surfel->getSpecularColor());
	}

	execute();

	UndoAction *undo = new ColorUndo(sfls,diff,spec);
	_MCnew( (unsigned long)undo, sizeof(ColorUndo(sfls,diff,spec)), __FILE__, __LINE__);
	
	_MCdelete((unsigned long)selection);
	delete selection;
		
	return undo;
}
// Some Emacs-Hints -- please don't remove:
//
//  Local Variables:
//  mode:C++
//  tab-width:4
//  End:
