// Title:   LocalParticleSimulation.cpp
// Created: Thu Sep 25 15:16:41 2003
// Authors: Simon Heinzle, Tim Weyrich
//
// Copyright (c) 2001, 2002, 2003, Computer Graphics Lab, ETH Zurich
//
// This file is part of the Pointshop3D system.
// See http://www.pointshop3d.com/ for more information.
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later
// version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
// 
// You should have received a copy of the GNU Lesser General
// Public License along with this library; if not, write to the
// Free Software Foundation, Inc., 59 Temple Place, Suite 330,
// Boston, MA 02111-1307 USA
// 
// Contact info@pointshop3d.com if any conditions of this
// licensing are not clear to you.
//

#ifdef _WIN32
	// turn off std::map warnings
	# pragma warning(disable:4786) 
#endif

#include "UndoRedoSpecialisations.h"

/**
 * old one in tmp.cpp
 */

#include <qvector.h>
#include <qapplication.h>
#include <algorithm>
#include <set>
#include "LocalParticleSimulation.h"
#include "Selector.h"
#include "MovingLeastSquares.h"
#include "MovingLeastSquaresSmoother.h"

/** 
 * rand includes
 */
#include <stdio.h>
#include <stdlib.h>
#include <time.h>

#include "NewOverload.h"

LocalParticleSimulation::LocalParticleSimulation() {
	_positions_array = 0;
	_new_positions_array = 0;
	_normals_array = 0;
	_surfel_indices = 0;
	_surfel_indices_uIntPtr = 0;
	_old_positions_array = 0;
	_query_tree = 0;
	_dynamic_hood = 0;
	_variation = 0;
	_full_mls = _use_mls = false;
}

LocalParticleSimulation::~LocalParticleSimulation() {
	if (_positions_array) {
		_MCdelete((unsigned long)_positions_array);
		delete[] _positions_array;
	}
	if (_new_positions_array) {
		_MCdelete((unsigned long)_new_positions_array);
		delete[] _new_positions_array;
	}
	if (_normals_array) {
		_MCdelete((unsigned long)_normals_array);
		delete[] _normals_array;
	}
	if (_surfel_indices_uIntPtr) {
		for (unsigned int i=0; i<_surfel_indices->size(); i++) {
			_MCdelete((unsigned long)_surfel_indices_uIntPtr[i]);
			delete _surfel_indices_uIntPtr[i];
		}
			
		_MCdelete((unsigned long)_surfel_indices_uIntPtr);
		delete[] _surfel_indices_uIntPtr;
	}

	if (_surfel_indices) {
		_MCdelete((unsigned long)_surfel_indices);
		delete _surfel_indices;
	}

	if (_old_positions_array) {
		_MCdelete((unsigned long)_old_positions_array);
		delete[] _old_positions_array;
	}
	if (_query_tree) {
		_MCdelete((unsigned long)_query_tree);
		delete _query_tree;
	}
	if (_dynamic_hood) {
		_MCdelete((unsigned long)_dynamic_hood);
		delete _dynamic_hood;
	}
	if (_variation) {
		_MCdelete((unsigned long)_variation);
		delete[] _variation;
	}
	
	
}

/**
 * get positions + border
 */
void LocalParticleSimulation::initPositions() {
	Selector *selector;
	std::vector<Vector3D> *positions;
	std::set<unsigned int> tmp_border;	// using map to avoid double 
	std::set<unsigned int>::iterator tmp_it;
	Vector3D tmp_vec;
	KdTree *tmp_tree;
	int i;
	unsigned int ui;
	
	selector = Selector::getInstance();
	tmp_tree = selector->returnKdTree();

	min = Vector3D(1e10, 1e10, 1e10);
	max = Vector3D(-1e10, -1e10, -1e10);

	// get selected positions
	positions = selector->getSelectionPositionsWorldStdVec(selector->intToFlag(_selection_index));
	// get selected indices
	_surfel_indices = selector->getSelectionIndicesStdVec(selector->intToFlag(_selection_index));	
	// set border index
	_border_index = _surfel_indices->size();

	// get border indices
	tmp_border.clear();

	for (i=0; i<_border_index; i++) {
		tmp_vec = (*positions)[i];
		tmp_tree->setNOfNeighbours(_border_k + 1);
		tmp_tree->queryPosition(tmp_vec);

		for (int j=1; j<_border_k+1; j++) {
			unsigned int found_index = tmp_tree->getNeighbourPositionIndex(j);
			if (!std::binary_search(_surfel_indices->begin(), _surfel_indices->end(),found_index))  
				if (selector->isValid(found_index))  // found a point not inside selection.
					tmp_border.insert(found_index);
				
		}
	}
	
	// append border indices to _surfel_indices.
	for (tmp_it = tmp_border.begin(); tmp_it!=tmp_border.end(); tmp_it++)
		_surfel_indices->push_back(*tmp_it);
		
	// create _positions_array
	_positions_array = new Vector3D[_surfel_indices->size()];
	_MCnew( (unsigned long)_positions_array, sizeof(Vector3D)*_surfel_indices->size(), __FILE__, __LINE__);
	
	_old_positions_array = new Vector3D[_surfel_indices->size()];
	_MCnew( (unsigned long)_old_positions_array, sizeof(Vector3D)*_surfel_indices->size(), __FILE__, __LINE__);

	_normals_array = new Vector3D[_surfel_indices->size()];
	_MCnew( (unsigned long)_normals_array, sizeof(Vector3D)*_surfel_indices->size(), __FILE__, __LINE__);
	
	// insert inside positions
	for (i=0; i<_border_index; i++) {
		_old_positions_array[i] = _positions_array[i] = (*positions)[i];
		selector->getSurfelWorldNormal((*_surfel_indices)[i], _normals_array[i]);
		// calculate bounding box
		min[0] = _positions_array[i][0] < min[0] ? _positions_array[i][0] : min[0];
		min[1] = _positions_array[i][1] < min[1] ? _positions_array[i][1] : min[1];
		min[2] = _positions_array[i][2] < min[2] ? _positions_array[i][2] : min[2];

		max[0] = _positions_array[i][0] > max[0] ? _positions_array[i][0] : max[0];
		max[1] = _positions_array[i][1] > max[1] ? _positions_array[i][1] : max[1];
		max[2] = _positions_array[i][2] > max[2] ? _positions_array[i][2] : max[2];

	}
	// append border positions
	for (ui=_border_index; ui<_surfel_indices->size(); ui++) {
		_old_positions_array[ui] = _positions_array[ui] = selector->getSurfelWorldPosition((*_surfel_indices)[ui]);
		selector->getSurfelWorldNormal((*_surfel_indices)[ui], _normals_array[ui]);
		// calculate bounding box
		min[0] = _positions_array[ui][0] < min[0] ? _positions_array[i][0] : min[0];
		min[1] = _positions_array[ui][1] < min[1] ? _positions_array[i][1] : min[1];
		min[2] = _positions_array[ui][2] < min[2] ? _positions_array[i][2] : min[2];

		max[0] = _positions_array[ui][0] > max[0] ? _positions_array[i][0] : max[0];
		max[1] = _positions_array[ui][1] > max[1] ? _positions_array[i][1] : max[1];
		max[2] = _positions_array[ui][2] > max[2] ? _positions_array[i][2] : max[2];
	}

	_MCdelete((unsigned long)positions);
	delete positions;
		
}

void LocalParticleSimulation::getBorder(int k_next) {

}

void LocalParticleSimulation::buildTree() {
	unsigned int i;

	_surfel_indices_uIntPtr = new uIntPtr[_surfel_indices->size()];
	_MCnew((unsigned long) _surfel_indices_uIntPtr, sizeof(_surfel_indices_uIntPtr), __FILE__, __LINE__);
	
	for (i=0; i<_surfel_indices->size(); i++) {
		_surfel_indices_uIntPtr[i] = new unsigned int;
		_MCnew((unsigned long)_surfel_indices_uIntPtr[i], sizeof(unsigned int), __FILE__, __LINE__);
		*_surfel_indices_uIntPtr[i] = (*_surfel_indices)[i];
		
	}
	
	// build tree
	_query_tree = new KdTree(_positions_array, _surfel_indices->size(), 10);
	_MCnew( (unsigned long)_query_tree, sizeof(_query_tree), __FILE__, __LINE__);
	
	// build dynamic neighbourhood
	float max_diam = (min - max).getLength();
	_dynamic_hood = new QueryGrid<uInt> (7, 2*max_diam); // 7 used to be 5
	_MCnew((unsigned long)_dynamic_hood, sizeof(_dynamic_hood), __FILE__,__LINE__);
	
	_dynamic_hood->insert(_surfel_indices_uIntPtr, _positions_array,  _surfel_indices->size());
}	
	
void LocalParticleSimulation::getVariation(float variation_constant) {
	unsigned int i,j;
	Vector3D source;
	Vector3D difference;
	Vector3D mean;
	Mgc::Eigen	*eigenSolver;
	Mgc::Real fSumXX = 0.0f, fSumXY = 0.0f, fSumXZ = 0.0f;
	Mgc::Real fSumYY = 0.0f, fSumYZ = 0.0f, fSumZZ = 0.0f;
	
	eigenSolver = new Mgc::Eigen(3);
	_MCnew( (unsigned long)eigenSolver, sizeof(eigenSolver), __FILE__, __LINE__);
	
	_variation = new float[_surfel_indices->size()];
	_MCnew( (unsigned long)_variation, sizeof(float) * _surfel_indices->size(), __FILE__, __LINE__);
	
	_variation_avg = 0;
	for (i=0; i<_surfel_indices->size(); i++) {
		// solve eigenvalue problem of the covariance matrix
		// get current position
		source = _positions_array[i];
		// we got to solve evalue's of the covariance matrix
		_query_tree->setNOfNeighbours(NO_OF_NEIGHBOURS_FOR_VARIATION);
		_query_tree->queryPosition(source);

		mean.makeZero();
		for (j = 0; j < NO_OF_NEIGHBOURS_FOR_VARIATION; j++) {		
			mean += _query_tree->getNeighbourPosition(j);
		}
		mean /= NO_OF_NEIGHBOURS_FOR_VARIATION;
		
		// clear values
		fSumXX = fSumXY = fSumXZ = fSumYY = fSumYZ = fSumZZ = 0.0f;
		for (j = 0; j < NO_OF_NEIGHBOURS_FOR_VARIATION; j++) {

			difference = mean - _query_tree->getNeighbourPosition(j);
			
			fSumXX += difference[0] * difference[0];
			fSumXY += difference[0] * difference[1];
			fSumXZ += difference[0] * difference[2];
			fSumYY += difference[1] * difference[1];
			fSumYZ += difference[1] * difference[2];
			fSumZZ += difference[2] * difference[2];
		}
		
		// setup the eigensolver
		eigenSolver->Matrix(0,0) = fSumXX;
		eigenSolver->Matrix(0,1) = fSumXY;
		eigenSolver->Matrix(0,2) = fSumXZ;
		eigenSolver->Matrix(1,0) = eigenSolver->Matrix(0,1);
		eigenSolver->Matrix(1,1) = fSumYY;
		eigenSolver->Matrix(1,2) = fSumYZ;
		eigenSolver->Matrix(2,0) = eigenSolver->Matrix(0,2);
		eigenSolver->Matrix(2,1) = eigenSolver->Matrix(1,2);
		eigenSolver->Matrix(2,2) = fSumZZ;
		
		// compute eigenstuff, smallest eigenvalue is in last position
		eigenSolver->DecrSortEigenStuff3();
		
		
	//	variation = sqrt (eigenSolver->GetEigenvalue(2)) / (sqrt(eigenSolver->GetEigenvalue(0)) + 
	//		sqrt(eigenSolver->GetEigenvalue(1)) + sqrt(eigenSolver->GetEigenvalue(2)));

		_variation[i] = eigenSolver->GetEigenvalue(2) / (eigenSolver->GetEigenvalue(0) + 
			eigenSolver->GetEigenvalue(1) + eigenSolver->GetEigenvalue(2));
		_variation[i] += variation_constant;
		_variation_avg += _variation[i];
	}

	_MCdelete((unsigned long)eigenSolver);
	delete eigenSolver;
		
}


Vector3D LocalParticleSimulation::doMLS(Vector3D source, int max_neighbours, float query_distance) {
	MovingLeastSquares mls;
	Vector3D *found;
	unsigned i;
	unsigned count;
	KdTree *kdtree;
	kdtree = Selector::getInstance()->returnKdTree();

	_query_tree->setNOfNeighbours(2*max_neighbours);
	_query_tree->queryPosition(source); 

	count = _query_tree->getNOfFoundNeighbours();
	
	found = new Vector3D[count];
	_MCnew( (unsigned long)found, sizeof(Vector3D)*count, __FILE__, __LINE__);

	for (i=0;i<count;i++) {
		found[i] = _query_tree->getNeighbourPosition(i);
	}
	
	mls.setNOfNeighbours(count);
	mls.setGaussParam(query_distance*query_distance);
	//mls.setGaussParam(50);
	mls.setPositions(found, count);
	mls.setSource(source);
	mls.doFullMLS();

	if (_full_mls)
		return mls.getPolyPosition();
	

	_MCdelete((unsigned long)found);
	delete found;
		
	return mls.getPlanePosition();
}

void LocalParticleSimulation::doAllMLS() {
	MovingLeastSquaresSmoother mls;
	mls.setParameters(40, 40, 1.0);
	mls.setSelection(_selection_index);
	
	mls.execute();
}

void LocalParticleSimulation::repel(int max_neighbours, float query_distance, int num_of_iterations, float force_scale, float damping) {
	Selector *selector;
	Vector3D *forces;
	Vector3D direction;
	Vector3D new_position;
	Vector3D normal;
	Vector3D old_position;
	float variation_force_scale;
	float variation_scale;
	
	float weight;
	int no_neighbours;
	int iteration;
	int i;
	int j;
	unsigned int relative_index;

	forces = new Vector3D[_border_index];
	_MCnew( (unsigned long)forces, sizeof(Vector3D)*_border_index, __FILE__, __LINE__);
	
	selector = Selector::getInstance();
	
	// iterate
	for (iteration = 0; iteration<num_of_iterations; iteration++) {
		_dynamic_hood->setNOfNeighbours(max_neighbours+1);
		_dynamic_hood->queryAdvisoryAverageQueryRadius(query_distance);

		// repel all points
		for (i=0; i<_border_index; i++) {
			//_dynamic_hood->queryRange(selector->getSurfelWorldPosition((*_surfel_indices)[i]),query_distance);
			_dynamic_hood->queryRange(_positions_array[i],query_distance);
			forces[i].makeZero();
			
			no_neighbours = _dynamic_hood->getNOfFoundNeighbours();
			if (no_neighbours > max_neighbours)
				no_neighbours = max_neighbours;

			// calc variation_force_scale
			_query_tree->setNOfNeighbours(1);
			_query_tree->queryPosition(_positions_array[i]);
			variation_scale = _variation[_query_tree->getNeighbourPositionIndex(0)];
			variation_force_scale = _variation_avg / variation_scale;
			
			// calc all forces acting on point i
			for (j=1; j<no_neighbours; j++) {
				//direction = selector->getSurfelWorldPosition((*_surfel_indices)[i]) - _dynamic_hood->getNeighbourPosition(j);
				direction = _positions_array[i] - _dynamic_hood->getNeighbourPosition(j);
				
				// get weight (3 versions)
				
				//weight = sqrt (_dynamic_hood->getSquaredDistance (no_neighbours-1)) - 
				//	sqrt (_dynamic_hood->getSquaredDistance (j));
				weight = (query_distance - direction.getLength());
				//weight = query_distance / direction.getLength();

				if (weight < 0.0) 
					weight = 0;
				
				// normalize directon
				direction.normalize();
				// calc force
				forces[i] += variation_force_scale * force_scale * direction * weight;
			}
			
			while(0) {}
		}

		
		for (i=0; i<_border_index; i++) {
			// euler step
			old_position = _positions_array[i];
			new_position = old_position + forces[i];
	
			_query_tree->setNOfNeighbours(1);
			_query_tree->queryPosition(new_position);
			// project new position of closest point's (of original set) tangent plane, 
			relative_index = _query_tree->getNeighbourPositionIndex(0);

			if (!selector->getSurfelWorldNormal((*_surfel_indices)[relative_index], normal)) {
				while(0);
				continue;
			}
			
			new_position = Vector3D::projectPointOntoPlane (new_position, old_position, normal);
			
			_dynamic_hood->move(_surfel_indices_uIntPtr[i],old_position,new_position);
			
			// set repulsion
			selector->setSurfelWorldPosition((*_surfel_indices)[i],new_position, false);
			selector->setSurfelWorldNormal((*_surfel_indices)[i],_normals_array[relative_index], false);
			
			_old_positions_array[i] = _positions_array[i];
			_positions_array[i] = new_position;
		}	
		
		Scene::getInstance()->emitObjectModifiedSignal (true);
		Scene::getInstance()->emitRenderFrameSignal ();
	}
	if (_use_mls) {
		doAllMLS();
	}
	_MCdelete((unsigned long)forces);
	delete[] forces;
}

void LocalParticleSimulation::project() {

}

void LocalParticleSimulation::setParameters(int border_k,int max_neighbours, float query_distance, int num_of_iterations, float force_scale, float damping) {
	_border_k = border_k;
	_max_neighbours = max_neighbours;
	_query_distance = query_distance;
	_num_of_iterations = num_of_iterations;
	_force_scale = force_scale;
	_damping = damping;
}

void LocalParticleSimulation::setMLS(bool on, bool full) {
	_use_mls = on;
	_full_mls = full;
}

void LocalParticleSimulation::execute() {
	QApplication::setOverrideCursor(Qt::waitCursor);

	initPositions();
	buildTree();
	getVariation(1);
	repel(_max_neighbours, _query_distance, _num_of_iterations, _force_scale, _damping);

	QApplication::restoreOverrideCursor();
}

UndoAction* LocalParticleSimulation::executeUndo() {
	QApplication::setOverrideCursor(Qt::waitCursor);

	std::vector<SurfelInterface*> sfls;
	std::vector<Vector3D> poses;
	std::vector<Vector3D> normals;

	Selector *selector;
	SurfelInterface *surfel;
	PositionNormalUndo *undo;

	selector = Selector::getInstance();
	initPositions();
	
	for (int i=0; i<_border_index; i++) {
		surfel = selector->getSurfel( (*_surfel_indices)[i] );
		sfls.push_back(surfel);
		poses.push_back(surfel->getPosition());
		normals.push_back(surfel->getNormal());
	}
	
	undo = new PositionNormalUndo(sfls, poses, normals);
	_MCnew( (unsigned long)undo, sizeof(undo), __FILE__, __LINE__);

	buildTree();
	getVariation(1);
	repel(_max_neighbours, _query_distance, _num_of_iterations, _force_scale, _damping);

	QApplication::restoreOverrideCursor();
	
	return undo;
}
// Some Emacs-Hints -- please don't remove:
//
//  Local Variables:
//  mode:C++
//  tab-width:4
//  End:
