// Title:   MovingLeastSquares.h
// Created: Thu Sep 25 15:16:26 2003
// Authors: Simon Heinzle, Tim Weyrich
//
// Copyright (c) 2001, 2002, 2003, Computer Graphics Lab, ETH Zurich
//
// This file is part of the Pointshop3D system.
// See http://www.pointshop3d.com/ for more information.
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later
// version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
// 
// You should have received a copy of the GNU Lesser General
// Public License along with this library; if not, write to the
// Free Software Foundation, Inc., 59 Temple Place, Suite 330,
// Boston, MA 02111-1307 USA
// 
// Contact info@pointshop3d.com if any conditions of this
// licensing are not clear to you.
//

#ifndef __MOVINGLEASTSQUARES_H
#define __MOVINGLEASTSQUARES_h

#ifdef _WIN32
	// turn off std::map warnings
	# pragma warning(push)
	# pragma warning(disable:4786) 
#endif

/**
 * Moving Least Squares Projector
 *
 *
 * @author Simon Heinzle
 * @version 1.0
 */

#include "PlaneFit.h"

#include "../../../../Core/DataStructures/src/NeighbourHood.h"

#include <vector>
#include <qvector.h>

class MovingLeastSquares {
	Vector3D _source;
	// planefit for normal
	PlaneFit _plane_fit;
	// projected position onto plane
	Vector3D _plane_projected_position;
	// projected position onto mls polynom
	Vector3D _poly_projected_position;
	Vector3D _normal;
	// parameters
	float _param;
	unsigned int _knext;
	float _min_error;
	int _max_iterations;

	// knext query tree
	KdTree *_kdtree;
	Vector3D *_kdtree_positions;
public:
	/**
	 * Constructor
	 *
	 * Parameter <code>float min_error = 1.e-3f</code> defines the minimum error of the plane projection,
	 * <code>int max_iterations = 10</code> defines the maximum number of iterations for plane projection
	 * if <code>min_error </code>is not reached.
	 */
	MovingLeastSquares(float min_error = 1.e-3f, int max_iterations = 10);
	/**
	 * Deconstructor
	 */
	~MovingLeastSquares();
	/**
	 * Sets the number of neighbours to be used for projection.
	 */
	void setNOfNeighbours(unsigned int knext);
	/**
	 * Sets the gauss parameter, weighting is done via exp ^ (- distance^2 / gauss_param^2)
	 */
	void setGaussParam(float gauss_param);
	/**
	 * Sets "adaptive" gauss parameter, <code>float cutoff</code> defines the distance at which the weigthing function
	 * is at f(distance) = 0.05
	 */
	void setAdaptiveGaussParam(float cutoff); 
	/**
	 * Sets the <code>const Vector3D &source</code> as point to be projected.
	 */
	void setSource(const Vector3D &source);
	/**
	 * Sets the positions which are used by the projection.
	 */
	void setPositions(const Vector3D *positions, int no_of_positions);
	/**
	 * Sets the positions which are used by the projection.
	 */
	void setPositions(const std::vector<Vector3D> &positions);
	/**
	 * Sets the positions which are used by the projection.
	 */
	void setPositions(const QVector<Vector3D> &positions);
	/**
	 * Performs the plane projection.
	 */
	void doPlaneProjection();
	/**
	 * Performs the polynomial projection, works only if plane projection is done earlier.
	 * <code> int degree = 3 </code> defines the degree of the fitted polynom, must be > 0.
	 */
	void doPolynomialProjection(int degree = 3);
	/**
	 * Perform full (plane & polynomial) projection. 
	 * <code> int degree = 3 </code> defines the degree of the fitted polynom, must be > 0.
	 */
	void doFullMLS(int degree = 3);
	/**
	 * @return the fitted plane projected position
	 */
	Vector3D getPlanePosition() const;
	/**
	 * @return the fitted plane projected normal
	 */
	Vector3D getPlaneNormal() const;
	/**
	 * @return the fitted polynomial projected position
	 */
	Vector3D getPolyPosition() const;
	/**
	 * @return the fitted polynomial projected normal
	 */
	Vector3D getPolyNormal() const;
	/**
	 * @return a <code>bool</code> to indicate wheter this projection is valid or not.
	 */
	bool isValid() const;

};

#ifdef _WIN32
	# pragma warning(pop)
#endif

#endif

// Some Emacs-Hints -- please don't remove:
//
//  Local Variables:
//  mode:C++
//  tab-width:4
//  End:
