// Title:   MovingLeastSquaresSmoother.cpp
// Created: Thu Sep 25 15:16:24 2003
// Authors: Simon Heinzle, Tim Weyrich
//
// Copyright (c) 2001, 2002, 2003, Computer Graphics Lab, ETH Zurich
//
// This file is part of the Pointshop3D system.
// See http://www.pointshop3d.com/ for more information.
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later
// version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
// 
// You should have received a copy of the GNU Lesser General
// Public License along with this library; if not, write to the
// Free Software Foundation, Inc., 59 Temple Place, Suite 330,
// Boston, MA 02111-1307 USA
// 
// Contact info@pointshop3d.com if any conditions of this
// licensing are not clear to you.
//

#include "MovingLeastSquaresSmoother.h"
#ifdef _WIN32
	// turn off std::map warnings
	//# pragma warning(push)
	# pragma warning(disable:4786) 
#endif

#include "MovingLeastSquares.h"

#include <vector>
#include <qdatetime.h>
#include <gsl/gsl_multifit.h>
#include <gsl/gsl_errno.h>
#include <qapplication.h>

#include "my_defs.h"
#include "UndoRedoSpecialisations.h"

#define NEIGHBOURHOOD_SIZE 10

#include "NewOverload.h"

MovingLeastSquaresSmoother::MovingLeastSquaresSmoother() : NormalAction() {
	_name = "moving least squares smoother";
	_selection_index = 0;

	_neighbourhood_size = NEIGHBOURHOOD_SIZE;
}

void MovingLeastSquaresSmoother::planeProjector(Vector3D source_position, int max_iterations, float min_error) {
	PlaneFit plane;
	float f;
	float epsilon;
	int current_iteration;
	Vector3D new_projected_source;
	
	_plane_projected_position = source_position;
	current_iteration = 0;
	do {
		plane.doPlaneFit(_plane_projected_position, _neighbourhood_size, _factor);
		f = -1.f * plane.getDistance(_plane_projected_position);
		new_projected_source =  _plane_projected_position + plane.getNormal() * f;
		epsilon = Vector3D::distance(new_projected_source, _plane_projected_position) / _plane_projected_position.getLength();
		current_iteration++;
		_plane_projected_position = new_projected_source;
	} while ( (epsilon > min_error) && (current_iteration < max_iterations) );
	_planefit = plane;
}

void MovingLeastSquaresSmoother::setParameters(int nhood_size, float gauss_factor, float alpha, bool useplane, int degree) {
	_neighbourhood_size = nhood_size;
	_factor = gauss_factor * gauss_factor;
	_alpha = alpha;
	_useplane = useplane;
	_degree = degree;
}

/**
 * this version uses the movingleastsquares.cpp implementation
 */

void MovingLeastSquaresSmoother::execute() {
	QApplication::setOverrideCursor (Qt::waitCursor);

	MovingLeastSquares MLS;
	Selector *selector;
	std::vector<unsigned int> *selection;
	std::vector<SurfelInterface*> old_surfels;
	std::vector<Vector3D> old_positions;
	std::vector<Vector3D> old_normals;
	std::vector<Vector3D> *all_positions;
	Vector3D curr_pos;
	Vector3D normal;
	Vector3D new_pos;
	Vector3D new_normal;

	QTime time;

	time.start();

	selector = Selector::getInstance();
	selection = selector->getSelectionIndicesStdVec(Selector::intToFlag(_selection_index));
	if (selection->size() == 0) {
		QApplication::restoreOverrideCursor();
		_MCdelete((unsigned long)selection);
		delete selection;
		return;
	}
	// get all positions
	
	all_positions = Selector::getInstance()->getAllPositionsWorldStdVec();

	MLS.setNOfNeighbours(_neighbourhood_size);
	MLS.setGaussParam(_factor);
	MLS.setPositions(*all_positions);
	
	// do projections
	for (unsigned int i=0; i<selection->size(); i++) {
		curr_pos = selector->getSurfelWorldPosition((*selection)[i]);
		MLS.setSource(curr_pos);
		MLS.doPlaneProjection();
		selector->getSurfelWorldNormal((*selection)[i],normal);
		
		if (_useplane) {
			new_pos = MLS.getPlanePosition() * _alpha + curr_pos * (1.0 - _alpha);
			selector->setSurfelWorldPosition((*selection)[i], new_pos, false);
			if (Vector3D::dotProduct(normal, MLS.getPlaneNormal()) < 0.0f)
				new_normal = MLS.getPlaneNormal() * -1.f;
			else
				new_normal = MLS.getPlaneNormal();
			new_normal = new_normal * _alpha + normal * (1.f - _alpha);
			selector->setSurfelWorldNormal((*selection)[i], new_normal,false);
		} else {
			MLS.doPolynomialProjection(_degree);
			new_pos = MLS.getPolyPosition() * _alpha + curr_pos * (1.0 - _alpha);
			selector->setSurfelWorldPosition((*selection)[i], new_pos, false);
			// using normal of polyfit
			if (Vector3D::dotProduct(normal, MLS.getPolyNormal()) < 0.0f)
				new_normal = MLS.getPolyNormal() * -1.f;
			else
				new_normal = MLS.getPolyNormal();
			new_normal = new_normal *_alpha + normal * (1.0 - _alpha);
			selector->setSurfelWorldNormal((*selection)[i], new_normal,false);

		}
	}		
	
	Scene::getInstance()->emitObjectModifiedSignal();
	Scene::getInstance()->emitRenderFrameSignal();

	_MCdelete((unsigned long)selection);
	delete selection;
	_MCdelete((unsigned long)all_positions);
	delete all_positions;


	old_positions.clear();
	old_normals.clear();
	old_surfels.clear();

	qDebug("%d msecs calculated",time.elapsed());
	QApplication::restoreOverrideCursor();
}


UndoAction* MovingLeastSquaresSmoother::executeUndo() {
	QApplication::setOverrideCursor (Qt::waitCursor);

	MovingLeastSquares MLS;
	Selector *selector;
	std::vector<unsigned int> *selection;
	std::vector<SurfelInterface*> old_surfels;
	std::vector<Vector3D> old_positions;
	std::vector<Vector3D> old_normals;
	std::vector<Vector3D> *all_positions;
	Vector3D curr_pos;
	Vector3D normal;
	Vector3D new_pos;
	Vector3D new_normal;

	QTime time;

	time.start();

	selector = Selector::getInstance();
	selection = selector->getSelectionIndicesStdVec(Selector::intToFlag(_selection_index));
	if (selection->size() == 0) {
		QApplication::restoreOverrideCursor();
		_MCdelete((unsigned long)selection);
		delete selection;
		return 0;
	}
	// get all positions
		
	all_positions = Selector::getInstance()->getAllPositionsWorldStdVec();

	MLS.setNOfNeighbours(_neighbourhood_size);
	MLS.setGaussParam(_factor);
	MLS.setPositions(*all_positions);
	
	// do projections
	for (unsigned int i=0; i<selection->size(); i++) {
		curr_pos = selector->getSurfelWorldPosition((*selection)[i]);
		MLS.setSource(curr_pos);
		MLS.doPlaneProjection();
		selector->getSurfelWorldNormal((*selection)[i],normal);
		
		if (_useplane) {
			new_pos = MLS.getPlanePosition() * _alpha + curr_pos * (1.0 - _alpha);
			
			SurfelInterface *surfel = selector->getSurfel((*selection)[i]);
			old_surfels.push_back(surfel);
			old_positions.push_back(surfel->getPosition());
			old_normals.push_back(surfel->getNormal());
			
			selector->setSurfelWorldPosition((*selection)[i], new_pos, false);
			if (Vector3D::dotProduct(normal, MLS.getPlaneNormal()) < 0.0f)
				new_normal = MLS.getPlaneNormal() * -1.f;
			else
				new_normal = MLS.getPlaneNormal();
			new_normal = new_normal * _alpha + normal * (1.f - _alpha);
			selector->setSurfelWorldNormal((*selection)[i], new_normal,false);
		} else {
			MLS.doPolynomialProjection(_degree);
			new_pos = MLS.getPolyPosition() * _alpha + curr_pos * (1.0 - _alpha);
		
			SurfelInterface *surfel = selector->getSurfel((*selection)[i]);
			old_surfels.push_back(surfel);
			old_positions.push_back(surfel->getPosition());
			old_normals.push_back(surfel->getNormal());

			selector->setSurfelWorldPosition((*selection)[i], new_pos, false);
			// using normal of polyfit
			if (Vector3D::dotProduct(normal, MLS.getPolyNormal()) < 0.0f)
				new_normal = MLS.getPolyNormal() * -1.f;
			else
				new_normal = MLS.getPolyNormal();
			new_normal = new_normal *_alpha + normal * (1.0 - _alpha);
			selector->setSurfelWorldNormal((*selection)[i], new_normal,false);

		}
	}		
	
	Scene::getInstance()->emitObjectModifiedSignal();
	Scene::getInstance()->emitRenderFrameSignal();

	qDebug("%d msecs calculated",time.elapsed());
	QApplication::restoreOverrideCursor();

	PositionNormalUndo *undo = new PositionNormalUndo(old_surfels, old_positions, old_normals);
	_MCnew( (unsigned long)undo, sizeof(undo), __FILE__, __LINE__);
	
	_MCdelete((unsigned long)selection);
	delete selection;
	_MCdelete((unsigned long)all_positions);
	delete all_positions;

	old_positions.clear();
	old_normals.clear();
	old_surfels.clear();
	return undo;
}
// Some Emacs-Hints -- please don't remove:
//
//  Local Variables:
//  mode:C++
//  tab-width:4
//  End:
