// Title:   OutlierClassificator.cpp
// Created: Thu Sep 25 15:15:59 2003
// Authors: Simon Heinzle, Tim Weyrich
//
// Copyright (c) 2001, 2002, 2003, Computer Graphics Lab, ETH Zurich
//
// This file is part of the Pointshop3D system.
// See http://www.pointshop3d.com/ for more information.
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later
// version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
// 
// You should have received a copy of the GNU Lesser General
// Public License along with this library; if not, write to the
// Free Software Foundation, Inc., 59 Temple Place, Suite 330,
// Boston, MA 02111-1307 USA
// 
// Contact info@pointshop3d.com if any conditions of this
// licensing are not clear to you.
//

#ifdef _WIN32
	// turn off std::map warnings
	//# pragma warning(push)
	# pragma warning(disable:4786 4275 4800) 
#endif

#include <algorithm>
#include <set>
#include <qapplication.h>

#include "OutlierClassificator.h"
#include "PlaneFit.h"
#include "Selector.h"
#include "miniball/miniball.H"

#include "../../../../Core/DataStructures/src/SurfelCollection.h"
#include "../../../../Core/Interfaces/src/SurfelInterface.h"
#include "../../../../Core/Scene/src/Object.h"
#include "../../../../Core/Scene/src/Scene.h"

#include "NewOverload.h"

OutlierClassificator::OutlierClassificator(int k) {
	_indices = 0;
	_kNext_graph = 0;
	_kNext_graph_sorted = 0;
	_classification_mb = 0;
	_classification_pf = 0;
	_confidence_pf = 0;
	_classification_bi = 0;
	_k = k;
	_count = 0;
	_surfels.clear();
	_old_diffuse.clear();
	_old_specular.clear();

}

OutlierClassificator::~OutlierClassificator() {
	if (_indices) {
		_MCdelete((unsigned long)_indices);
		delete(_indices);
	}
	if (_kNext_graph) {
		_MCdelete((unsigned long)_kNext_graph);
		delete[](_kNext_graph);
	}
	if (_kNext_graph_sorted) {
		_MCdelete((unsigned long)_kNext_graph_sorted);
		delete[](_kNext_graph_sorted);
	}
	if (_classification_mb) {
		_MCdelete((unsigned long)_classification_mb);
		delete[](_classification_mb);
	}
	if (_classification_pf) {
		_MCdelete((unsigned long)_classification_pf);
		delete[](_classification_pf);
	}
	if (_confidence_pf) {
		_MCdelete((unsigned long)_confidence_pf);
		delete[](_confidence_pf);
	}
	if (_classification_bi) {
		_MCdelete((unsigned long)_classification_bi);
		delete[](_classification_bi);
	}
}

bool OutlierClassificator::buildKNextGraph(int selection_index) {
	QApplication::setOverrideCursor (Qt::waitCursor);

	Selector *selector;
	int selection_count;
	int i;
	int knext;
	Vector3D position;
	KdTree *kdtree;
		
	qDebug("[OutlierClassificator] starting...");
	_time_overall.start();
	
	// clean up old arrays
	if (_indices) {
		_MCdelete((unsigned int)_indices);
		delete _indices;
		_indices = 0;
	}
	if (_classification_mb) {
		_MCdelete((unsigned int)_classification_mb);
		delete[] _classification_mb;
		_classification_mb = 0;
	}
	if (_classification_pf) {
		_MCdelete((unsigned int)_classification_pf);
		delete[] _classification_pf;
		_classification_pf = 0;
	}
	if (_confidence_pf) {
		_MCdelete((unsigned int)_confidence_pf);
		delete[] _confidence_pf;
		_confidence_pf = 0;
	}
	if (_classification_bi) {
		_MCdelete((unsigned int)_classification_bi);
		delete[] _classification_bi;
		_classification_bi = 0;
	}
	if (_kNext_graph) {
		_MCdelete((unsigned int)_kNext_graph);
		delete[] _kNext_graph;
		_kNext_graph = 0;
	}
	_surfels.clear();
	_old_diffuse.clear();
	_old_specular.clear();
	restoreColor();
	
	// create new arrays
	selector = Selector::getInstance();
	selector->buildTree();
	if (selection_index = -1)
		// get all points
		_indices = selector->getAllIndicesStdVec();
	else
		_indices = selector->getSelectionIndicesStdVec(Selector::intToFlag(selection_index));
	
	// check if we got a selection
	_count = selection_count = _indices->size();
	if (selection_count == 0) {
		QApplication::restoreOverrideCursor();
		return false;
	}
	
	_classification_mb = new float[selection_count];
	_MCnew( (unsigned long)_classification_mb, sizeof(_classification_mb), __FILE__, __LINE__);

	_classification_pf = new float[selection_count];
	_MCnew( (unsigned long)_classification_pf, sizeof(_classification_pf), __FILE__, __LINE__);

	_confidence_pf = new float[selection_count];
	_MCnew( (unsigned long)_confidence_pf, sizeof(_confidence_pf), __FILE__, __LINE__);
	
	_classification_bi = new float[selection_count];
	_MCnew( (unsigned long)_classification_bi, sizeof(_classification_bi), __FILE__, __LINE__);
	
	_kNext_graph = new std::vector<unsigned int>[selection_count];
	_MCnew( (unsigned long)_kNext_graph, sizeof(_kNext_graph), __FILE__, __LINE__);

	_kNext_graph_sorted = new std::vector<unsigned int>[selection_count];
	_MCnew( (unsigned long)_kNext_graph_sorted, sizeof(_kNext_graph_sorted), __FILE__, __LINE__);

	// fill _kNext_graph and init _classification_mb to 0.5
	kdtree = selector->returnKdTree();
	for (i=0; i<selection_count; i++) {
		_classification_mb[i] = -1.f;
		_classification_pf[i] = -1.f;
		_classification_bi[i] = -1.f;
		_confidence_pf[i] = -1.f;

		kdtree->setNOfNeighbours(_k+1);
		kdtree->queryPosition(selector->getSurfelWorldPosition( (*_indices)[i]));

		_kNext_graph[i].clear();
		_kNext_graph_sorted[i].clear();
		for (knext = 0; knext < _k; knext++) {
			_kNext_graph[i].push_back(kdtree->getNeighbourPositionIndex(knext+1));
			_kNext_graph_sorted[i].push_back(kdtree->getNeighbourPositionIndex(knext+1));
		}
		
		// sort _kNext_graph_sorted
		std::sort(_kNext_graph_sorted[i].begin(), _kNext_graph_sorted[i].end());
	}
#ifdef OUTLIERTIMING
	qDebug("[OutlierClassificator] graph built, done in %d msec", _time_overall.elapsed());
#endif
	QApplication::restoreOverrideCursor();
	return true;
}

bool OutlierClassificator::doMiniballClassification(int selection_index) {
	
	QApplication::setOverrideCursor (Qt::waitCursor);

	Vector3D pos;
	Selector *selector;
	std::vector<unsigned int> *selection;
	float diameter;
	float distance;
	float ratio;
	unsigned int index,j;
	unsigned int real_index;
	QTime mytime;
	
	mytime.start();

	_selection_index = selection_index;

	selector = Selector::getInstance();
	selection = selector->getSelectionIndicesStdVec(Selector::intToFlag(selection_index));
	
	_max_c_mb = -1.e10;
	_min_c_mb = 1.e10;

	for (index = 0; index<selection->size(); index++) {
		Miniball<3> mb;
		Miniball<3>::Point mb_point;

		real_index = (*selection)[index];
		// create miniball
		for (j = _k*DISCARD_RATIO; j<_k; j++) {
			if (selector->isValid( (_kNext_graph[real_index])[j])) {
				pos = selector->getSurfelWorldPosition( (_kNext_graph[real_index])[j] );
				mb_point[0] = pos[0];
				mb_point[1] = pos[1];
				mb_point[2] = pos[2];
				mb.check_in(mb_point);
			}
		}
		mb.build();
		// get diameter and center
		diameter = sqrt(mb.squared_radius()) * 2.f;
		mb_point = mb.center();
		// calc distance & ratio
		pos = selector->getSurfelWorldPosition( real_index );
		distance = Vector3D::distance(pos, Vector3D(mb_point[0], mb_point[1], mb_point[2]));
#if 0
		if (distance != 0) {
			ratio = distance / diameter;
			// scale ratio with
			ratio /= scale_ratio;
			if (ratio>1.f)
				ratio = 1.f;
			if (ratio<0.f)
				ratio = 0.f;
		} else 
			ratio = 0.f;
#else
		if (distance != 0)
			ratio = distance / diameter;
		else ratio = 0.f;

		/*_min_c_mb = ratio < _min_c_mb ? ratio : _min_c_mb;
		_max_c_mb = ratio > _max_c_mb ? ratio : _max_c_mb;*/
		if (ratio < _min_c_mb)
			_min_c_mb = ratio;
		if (ratio > _max_c_mb)
			_max_c_mb = ratio;

#endif
		_classification_mb[real_index] = ratio;

	}

#ifdef OUTLIERTIMING
	qDebug("[OutlierClassificator] classified / max = %f , min = %f / done in %d msec, total %d msec", _max_c_mb, _min_c_mb, mytime.elapsed(), _time_overall.elapsed());
#endif
	
	_MCdelete((unsigned long)selection);
	delete selection;

	QApplication::restoreOverrideCursor();
	return true;
}

bool OutlierClassificator::doPlaneClassification(int selection_index) {
	QApplication::setOverrideCursor (Qt::waitCursor);
	Selector *selector;
	std::vector<unsigned int> *selection;
	PlaneFit plane;
	float distance;
	float testdistance;
	float ratio;
	float confidence;//,classifies,makes,you,believing,the,political,lies;
	float eigenvals[3];
	int index;
	int real_index;
	int j;
	QTime mytime;

	mytime.start();
	
	_max_c_pf = _max_conf_pf = -1.e10;
	_min_c_pf = _min_conf_pf = 1.e10;
	selector = Selector::getInstance();
	selection = selector->getSelectionIndicesStdVec(Selector::intToFlag(selection_index));
	
	_selection_index = selection_index;
	

	for (index = 0; index<selection->size(); index++) {
		// fit plane
		real_index = (*selection)[index];
		plane.doPlaneFitByIndicesStdVec(&_kNext_graph[real_index],_k*DISCARD_RATIO,_k);
		plane.getEigenValues(eigenvals);
		// get distances
		distance  = 0;
		for (j=_k*DISCARD_RATIO; j<_k; j++)  
			if (selector->isValid((_kNext_graph[real_index])[j]))
				distance += fabs(plane.getDistance(selector->getSurfelWorldPosition( (_kNext_graph[real_index])[j] )));
		
		distance /= _k * ( 1.f - DISCARD_RATIO);
		
		testdistance = fabs(plane.getDistance(selector->getSurfelWorldPosition(real_index)));
		
		if (distance != 0)
			ratio = testdistance / distance;
		else ratio = 0.f;

		
		// when ev[2] == 0 -> handling
		confidence = eigenvals[2] / (eigenvals[0] + eigenvals[1] + eigenvals[2]);

		if (ratio < _min_c_pf)
			_min_c_pf = ratio;
		if (ratio > _max_c_pf)
			_max_c_pf = ratio;
		
		if (confidence < _min_conf_pf)
			_min_conf_pf = confidence;
		if (confidence > _max_conf_pf)
			_max_conf_pf = confidence;
		
		_classification_pf[real_index] = ratio;
		_confidence_pf[real_index] = confidence;
	}

	if (_min_c_pf == _max_c_pf )
		_max_c_pf = _min_c_pf = -1.0f;

#ifdef OUTLIERTIMING
	qDebug("[OutlierClassificator] classified / max = %f , min = %f / done in %d msec, total %d msec", _max_c_pf, _min_c_pf, mytime.elapsed(), _time_overall.elapsed());
#endif

	_MCdelete((unsigned long)selection);
	delete selection;

	QApplication::restoreOverrideCursor();
	return true;
}

bool OutlierClassificator::doBidirectionClassification(int selection_index) {
	QApplication::setOverrideCursor (Qt::waitCursor);

	Selector *selector;
	std::vector<unsigned int> *selection;
	float ratio;
	int uni_count, bi_count;
	int neighbour;
	unsigned int index;
	unsigned int real_index;
	int j;
	QTime mytime;
	
	mytime.start();
	
	_max_c_bi = -1.e10;
	_min_c_bi = 1.e10;

	selector = Selector::getInstance();
	selection = selector->getSelectionIndicesStdVec(Selector::intToFlag(selection_index));
	_selection_index = selection_index;
	
	
	for (index = 0; index<selection->size(); index++) {
		uni_count = 0;
		bi_count = 0;
		
		real_index = (*selection)[index];
		// get number of bidirectional and unidirectional "edges"
		for (j=0; j<_k; j++) {
			neighbour = (_kNext_graph_sorted[real_index])[j];
			if (selector->isValid(neighbour))
				if (std::binary_search(_kNext_graph_sorted[neighbour].begin(), _kNext_graph_sorted[neighbour].end(),real_index))
					bi_count++;
				else
					uni_count++;
		}

		if (bi_count != 0)
			ratio = (float) uni_count /  (float) bi_count;
		else
			ratio = _k;

		if (ratio < _min_c_bi)
			_min_c_bi = ratio;
		if (ratio > _max_c_bi)
			_max_c_bi = ratio;
		
		_classification_bi[real_index] = ratio;
	}
//#ifdef OUTLIERTIMING
	qDebug("[OutlierClassificator] classified / max = %f , min = %f / done in %d msec, total %d msec", _max_c_bi, _min_c_bi, mytime.elapsed(), _time_overall.elapsed());
//#endif
	
	_MCdelete((unsigned long)selection);
	delete selection;

	QApplication::restoreOverrideCursor();
	return true;
}

bool OutlierClassificator::scaleClassification() {
	/* currently not provided */
	return 0;
	
	QTime mytime;
	mytime.start();

	for (int i=0; i<_count; i++) {
		_classification_mb[i] = (_classification_mb[i] - _min_c_mb) / (_max_c_mb - _min_c_mb);
		if (_classification_mb[i] > 1.f) 
			_classification_mb[i] = 1.f;
		if (_classification_mb[i] < 0.f)
			_classification_mb[i] = 0.f;
	}

	qDebug("[OutlierClassificator] scaled, done in %d msec, total %d msec", mytime.elapsed(), _time_overall.elapsed());
	return true;
}

bool OutlierClassificator::colorSurfelsMB(float lower, float upper) {
	QApplication::setOverrideCursor(Qt::waitCursor);

	Scene *scene;
	Selector *selector;
	std::vector<unsigned int> *selection;
	SurfelInterface *surfel;
	float angle;

	int real_index;	
	int index;
	
	QTime mytime;
	mytime.start();

	scene = Scene::getInstance();
	selector = Selector::getInstance();
	
	selection = selector->getSelectionIndicesStdVec(Selector::intToFlag(_selection_index));

	for (index=0; index<selection->size(); index++) {
		real_index = (*selection)[index];
		surfel = selector->getSurfel(real_index);
	
		angle = 1.f - (_classification_mb[real_index] - lower) / (upper - lower);
		if (angle > 1.f) 
			angle = 1.f;
		if (angle < 0.f)
			angle = 0.f;
			
		QColor col;
		col.setHsv (angle * 240.0f, 255, 255);
		
		scene->setSurfelDiffuseColor(surfel, col.rgb(),false);
		scene->setSurfelSpecularColor(surfel, col.rgb(), false);

	}

	scene->emitObjectModifiedSignal();
	scene->emitRenderFrameSignal();

#ifdef OUTLIERTIMING
	qDebug("[OutlierClassificator] colored, done in %d msec, total %d msec", mytime.elapsed(), _time_overall.elapsed());
#endif	
	
	_MCdelete((unsigned long)selection);
	delete selection;

	QApplication::restoreOverrideCursor();
	return true;
}

bool OutlierClassificator::colorSurfelsMB()	{
	return colorSurfelsMB(_min_c_mb,_max_c_mb);
}

bool OutlierClassificator::colorSurfelsCutoffMB(float cutoff) {
	QApplication::setOverrideCursor(Qt::waitCursor);

	Scene *scene;
	Selector *selector;
	SurfelInterface *surfel;
	std::vector<unsigned int> *selection;
	int index;
	int real_index;

	scene = Scene::getInstance();
	selector = Selector::getInstance();
	selection = selector->getSelectionIndicesStdVec(Selector::intToFlag(_selection_index));
	
	restoreColor();
	_surfels.clear();
	_old_diffuse.clear();
	_old_specular.clear();
	
	for (index=0; index<selection->size(); index++) {
		real_index = (*selection)[index];
		surfel = selector->getSurfel(real_index);
		// save old color
		_surfels.push_back(surfel);
		_old_diffuse.push_back(surfel->getDiffuseColor());
		_old_specular.push_back(surfel->getSpecularColor());
		if (_classification_mb[real_index] > cutoff) {
			scene->setSurfelDiffuseColor(surfel, qRgb(255,0,0),false);
			scene->setSurfelSpecularColor(surfel, qRgb(255,0,0), false);
		} else {
			scene->setSurfelDiffuseColor(surfel, qRgb(255,255,255),false);
			scene->setSurfelSpecularColor(surfel, qRgb(255,255,255), false);
		}
	}
	
	scene->emitObjectModifiedSignal();
	scene->emitRenderFrameSignal();
	
	_MCdelete((unsigned long)selection);
	delete selection;

	QApplication::restoreOverrideCursor();
	return true;
}

UndoAction* OutlierClassificator::deleteSurfelsCutoffMB(float cutoff) {
	QApplication::setOverrideCursor(Qt::waitCursor);

	Scene *scene;
	Selector *selector;
	SurfelInterface *surfel;
	std::vector<unsigned int> *selection;		
	std::vector<unsigned int> del_surfels;
	int index;
	int real_index;
	int del_count;

	del_surfels.resize(_count);
	
	restoreColor();

	scene = Scene::getInstance();
	selector = Selector::getInstance();
	selection = selector->getSelectionIndicesStdVec(Selector::intToFlag(_selection_index));
	
	del_count = 0;
	for (index=0; index<selection->size(); index++) {
		real_index = (*selection)[index];
		surfel = selector->getSurfel(real_index);
		if (_classification_mb[real_index] > cutoff) {
			del_surfels.push_back(real_index);
			scene->setSurfelDiffuseColor(surfel, qRgb(255,0,0),false);
			scene->setSurfelSpecularColor(surfel, qRgb(255,0,0), false);
			del_count++;
		}
	}

	
	UndoAction *undo;
	undo = selector->deleteSurfelsUndo(del_surfels);
	
	_MCdelete((unsigned long)selection);
	delete selection;

	QApplication::restoreOverrideCursor();
	return undo;
}

bool OutlierClassificator::colorSurfelsPF(float lower, float upper) {
	QApplication::setOverrideCursor(Qt::waitCursor);

	Scene *scene;
	Selector *selector;
	SurfelInterface *surfel;
	std::vector<unsigned int> *selection;		
	float angle;
	int index;
	int real_index;
	
	QTime mytime;
	mytime.start();

	scene = Scene::getInstance();
	selector = Selector::getInstance();
	selection = selector->getSelectionIndicesStdVec(Selector::intToFlag(_selection_index));
	
	//selector->deselectAll();
	
	for (index=0; index<selection->size(); index++) {
		real_index = (*selection)[index];
		surfel = selector->getSurfel(real_index);
	
		angle = 1.f - (_classification_pf[real_index] - lower) / (upper - lower);
		if (angle > 1.f) 
			angle = 1.f;
		if (angle < 0.f)
			angle = 0.f;
	
		QColor col;
		col.setHsv (angle * 240.0f, 255, 255);
		
		scene->setSurfelDiffuseColor(surfel, col.rgb(),false);
		scene->setSurfelSpecularColor(surfel, col.rgb(), false);

	}

	scene->emitObjectModifiedSignal();
	scene->emitRenderFrameSignal();

#ifdef OUTLIERTIMING	
	qDebug("[OutlierClassificator] colored, done in %d msec, total %d msec", mytime.elapsed(), _time_overall.elapsed());
#endif
	
	_MCdelete((unsigned long)selection);
	delete selection;

	QApplication::restoreOverrideCursor();
	return true;
}

bool OutlierClassificator::colorSurfelsPF() {
	return colorSurfelsPF(_min_c_pf, _max_c_pf);
}

bool OutlierClassificator::colorSurfelsConfidencePF(float lower, float upper) {
	QApplication::setOverrideCursor(Qt::waitCursor);

	Scene *scene;
	Selector *selector;
	SurfelInterface *surfel;
	std::vector<unsigned int> *selection;
	float angle;
	int real_index;			
	int index;
	
	QTime mytime;
	mytime.start();

	scene = Scene::getInstance();
	selector = Selector::getInstance();
	selection = selector->getSelectionIndicesStdVec(Selector::intToFlag(_selection_index));
	
	//selector->deselectAll();
	
	for (index=0; index<selection->size(); index++) {
		real_index = (*selection)[index];
		surfel = selector->getSurfel(real_index);
	
		angle = 1.f - (_confidence_pf[real_index] - lower) / (upper - lower);
		if (angle > 1.f) 
			angle = 1.f;
		if (angle < 0.f)
			angle = 0.f;
	
		QColor col;
		col.setHsv (angle * 240.0f, 255, 255);
		
		scene->setSurfelDiffuseColor(surfel, col.rgb(),false);
		scene->setSurfelSpecularColor(surfel, col.rgb(), false);

	}

	scene->emitObjectModifiedSignal();
	scene->emitRenderFrameSignal();

#ifdef OUTLIERTIMING
	qDebug("[OutlierClassificator] colored, done in %d msec, total %d msec", mytime.elapsed(), _time_overall.elapsed());
#endif
	
	_MCdelete((unsigned long)selection);
	delete selection;
	
	QApplication::restoreOverrideCursor();
	return true;
}

bool OutlierClassificator::colorSurfelsConfidencePF() {
	return colorSurfelsConfidencePF(_min_conf_pf, _max_conf_pf);
}

bool OutlierClassificator::colorSurfelsCutoffPF(float cutoff, float cutoff_confidence) {
	QApplication::setOverrideCursor(Qt::waitCursor);
	Scene *scene;
	Selector *selector;
	std::vector<unsigned int>* selection;
	SurfelInterface *surfel;
	int index;
	int real_index;

	scene = Scene::getInstance();
	selector = Selector::getInstance();
	selection = selector->getSelectionIndicesStdVec(Selector::intToFlag(_selection_index));

	restoreColor();
	_surfels.clear();
	_old_diffuse.clear();
	_old_specular.clear();
	
	for (index=0; index<selection->size(); index++) {
		real_index = (*selection)[index];
		surfel = selector->getSurfel(real_index);
		// save old color
		_surfels.push_back(surfel);
		_old_diffuse.push_back(surfel->getDiffuseColor());
		_old_specular.push_back(surfel->getSpecularColor());
	
		if (_classification_pf[real_index] > cutoff && _confidence_pf[real_index] > cutoff_confidence) {
			scene->setSurfelDiffuseColor(surfel, qRgb(255,0,0),false);
			scene->setSurfelSpecularColor(surfel, qRgb(255,0,0), false);
		} else {
			scene->setSurfelDiffuseColor(surfel, qRgb(255,255,255),false);
			scene->setSurfelSpecularColor(surfel, qRgb(255,255,255), false);
		}
	}
	scene->emitObjectModifiedSignal();
	scene->emitRenderFrameSignal();
	
	_MCdelete((unsigned long)selection);
	delete selection;

	QApplication::restoreOverrideCursor();
	return true;
}

bool OutlierClassificator::colorSurfelsCutoffPF(float cutoff) {
	QApplication::setOverrideCursor(Qt::waitCursor);
	
	Scene *scene;
	Selector *selector;
	std::vector<unsigned int>* selection;
	SurfelInterface *surfel;
	int index;
	int real_index;

	scene = Scene::getInstance();
	selector = Selector::getInstance();
	selection = selector->getSelectionIndicesStdVec(Selector::intToFlag(_selection_index));

	restoreColor();
	_surfels.clear();
	_old_diffuse.clear();
	_old_specular.clear();
	
	for (index=0; index<selection->size(); index++) {
		real_index = (*selection)[index];
		surfel = selector->getSurfel(real_index);
		// save old color
		_surfels.push_back(surfel);
		_old_diffuse.push_back(surfel->getDiffuseColor());
		_old_specular.push_back(surfel->getSpecularColor());
	
		if ( (_classification_pf[real_index] * _confidence_pf[real_index]) > cutoff) {
			scene->setSurfelDiffuseColor(surfel, qRgb(255,0,0),false);
			scene->setSurfelSpecularColor(surfel, qRgb(255,0,0), false);
		} else {
			scene->setSurfelDiffuseColor(surfel, qRgb(255,255,255),false);
			scene->setSurfelSpecularColor(surfel, qRgb(255,255,255), false);
		}
	}
	scene->emitObjectModifiedSignal();
	scene->emitRenderFrameSignal();
	
	_MCdelete((unsigned long)selection);
	delete selection;
	
	QApplication::restoreOverrideCursor();
	return true;
}


UndoAction* OutlierClassificator::deleteSurfelsCutoffPF(float cutoff, float cutoff_confidence) {
	QApplication::setOverrideCursor(Qt::waitCursor);
	Scene *scene;
	Selector *selector;
	std::vector<unsigned int>* selection;
	SurfelInterface *surfel;
	std::vector<unsigned int> del_surfels;
	int index;
	int real_index;
	int del_count;

	del_surfels.resize(_count);
	restoreColor();

	scene = Scene::getInstance();
	selector = Selector::getInstance();
	selection = selector->getSelectionIndicesStdVec(Selector::intToFlag(_selection_index));

	del_count = 0;
	for (index=0; index<selection->size(); index++) {
		real_index = (*selection)[index];
		surfel = selector->getSurfel(real_index);
		if (_classification_pf[real_index] > cutoff && _confidence_pf[real_index] > cutoff_confidence) {
			del_surfels.push_back(real_index);
			scene->setSurfelDiffuseColor(surfel, qRgb(255,0,0),false);
			scene->setSurfelSpecularColor(surfel, qRgb(255,0,0), false);
			del_count++;
		}
	}

	UndoAction *undo;
	undo = selector->deleteSurfelsUndo(del_surfels);
	
	_MCdelete ((unsigned long)selection);
	delete selection;

	QApplication::restoreOverrideCursor();
	return undo;
}

UndoAction* OutlierClassificator::deleteSurfelsCutoffPF(float cutoff) {
	QApplication::setOverrideCursor(Qt::waitCursor);
	Scene *scene;
	Selector *selector;
	std::vector<unsigned int>* selection;
	SurfelInterface *surfel;
	std::vector<unsigned int> del_surfels;
	int index;
	int real_index;
	int del_count;

	del_surfels.resize(_count);
	restoreColor();

	scene = Scene::getInstance();
	selector = Selector::getInstance();
	selection = selector->getSelectionIndicesStdVec(Selector::intToFlag(_selection_index));

	del_count = 0;
	for (index=0; index<selection->size(); index++) {
		real_index = (*selection)[index];
		surfel = selector->getSurfel(real_index);
		if ((_classification_pf[real_index] * _confidence_pf[real_index]) > cutoff) {
			del_surfels.push_back(real_index);
			scene->setSurfelDiffuseColor(surfel, qRgb(255,0,0),false);
			scene->setSurfelSpecularColor(surfel, qRgb(255,0,0), false);
			del_count++;
		}
	}

	UndoAction *undo;
	undo = selector->deleteSurfelsUndo(del_surfels);
	
	_MCdelete((unsigned long)selection);
	delete selection;

	QApplication::restoreOverrideCursor();
	return undo;
}



bool OutlierClassificator::colorSurfelsBI(float lower, float upper) {
	QApplication::setOverrideCursor(Qt::waitCursor);
	Scene *scene;
	Selector *selector;
	SurfelInterface *surfel;
	std::vector<unsigned int> *selection;
	float angle;
	int index;
	int real_index;
	
	QTime mytime;
	mytime.start();

	scene = Scene::getInstance();
	selector = Selector::getInstance();
	selection = selector->getSelectionIndicesStdVec(Selector::intToFlag(_selection_index));
	//selector->deselectAll();
	
	for (index=0; index<selection->size(); index++) {
		real_index = (*selection)[index];
		surfel = selector->getSurfel(real_index);
	
		angle = 1.f - (_classification_bi[real_index] - lower) / (upper - lower);
		if (angle > 1.f) 
			angle = 1.f;
		if (angle < 0.f)
			angle = 0.f;
	
		QColor col;
		col.setHsv (angle * 240.0f, 255, 255);
		
		scene->setSurfelDiffuseColor(surfel, col.rgb(),false);
		scene->setSurfelSpecularColor(surfel, col.rgb(), false);

	}

	scene->emitObjectModifiedSignal();
	scene->emitRenderFrameSignal();

#ifdef OUTLIERTIMING	
	qDebug("[OutlierClassificator] colored, done in %d msec, total %d msec", mytime.elapsed(), _time_overall.elapsed());
#endif
	
	_MCdelete((unsigned long)selection);
	delete selection;

	QApplication::restoreOverrideCursor();
	return true;
}

bool OutlierClassificator::colorSurfelsBI() {
	return colorSurfelsBI(_min_c_bi, _max_c_bi);
}

bool OutlierClassificator::colorSurfelsCutoffBI(float cutoff) {
	QApplication::setOverrideCursor(Qt::waitCursor);
	Scene *scene;
	Selector *selector;
	SurfelInterface *surfel;
	std::vector<unsigned int> *selection;
	int index;
	int real_index;

	scene = Scene::getInstance();
	selector = Selector::getInstance();
	selection = selector->getSelectionIndicesStdVec(Selector::intToFlag(_selection_index));
	
	restoreColor();
	_surfels.clear();
	_old_diffuse.clear();
	_old_specular.clear();
	
	for (index=0; index<selection->size(); index++) {
		real_index = (*selection)[index];
		surfel = selector->getSurfel(real_index);
		// save old color
		_surfels.push_back(surfel);
		_old_diffuse.push_back(surfel->getDiffuseColor());
		_old_specular.push_back(surfel->getSpecularColor());
	
		if (_classification_bi[real_index] > cutoff) {
			scene->setSurfelDiffuseColor(surfel, qRgb(255,0,0),false);
			scene->setSurfelSpecularColor(surfel, qRgb(255,0,0), false);
		} else {
			scene->setSurfelDiffuseColor(surfel, qRgb(255,255,255),false);
			scene->setSurfelSpecularColor(surfel, qRgb(255,255,255), false);
		}

	}
	scene->emitObjectModifiedSignal();
	scene->emitRenderFrameSignal();
	
	_MCdelete((unsigned long)selection);
	delete selection;

	QApplication::restoreOverrideCursor();
	return true;
}

UndoAction* OutlierClassificator::deleteSurfelsCutoffBI(float cutoff) {
	QApplication::setOverrideCursor(Qt::waitCursor);
	Scene *scene;
	Selector *selector;
	SurfelInterface *surfel;
	std::vector<unsigned int> *selection;
	std::vector<unsigned int> del_surfels;
	int index;
	int real_index;
	int del_count;

	del_surfels.resize(_count);
	restoreColor();

	scene = Scene::getInstance();
	selector = Selector::getInstance();
	selection = selector->getSelectionIndicesStdVec(Selector::intToFlag(_selection_index));

	del_count = 0;
	for (index=0; index<selection->size(); index++) {
		real_index = (*selection)[index];
		surfel = selector->getSurfel(real_index);
		if (_classification_bi[real_index] > cutoff) {
			del_surfels.push_back(real_index);
			scene->setSurfelDiffuseColor(surfel, qRgb(255,0,0),false);
			scene->setSurfelSpecularColor(surfel, qRgb(255,0,0), false);
			del_count++;
		}
	}

	UndoAction *undo = selector->deleteSurfelsUndo(del_surfels);
	
	_MCdelete((unsigned long)selection);
	delete selection;

	QApplication::restoreOverrideCursor();
	return undo;
}

float OutlierClassificator::getCriterion(float weight_mb, float cutoff_mb, 
										 float weight_pf, float cutoff_pf,
										 float weight_bi, float cutoff_bi,
										 int real_index) {
	float tmp_class;
	float criterion;
	float upper_d, lower_d;

	criterion = 0;
	tmp_class = 0;

	
	// MB: clamp all [0,1], .5 is cutoff_mb
	upper_d = _max_c_mb - cutoff_mb;
	lower_d = cutoff_mb - _min_c_mb;

	if (lower_d > upper_d) {
		lower_d = cutoff_mb - upper_d;
		upper_d = _max_c_mb;
	} else {
		upper_d = cutoff_mb + lower_d;
		lower_d = _min_c_mb;
	}
	// get clamped classification, may exceed [0,1] but 0.5 is midpoint 0.5
	tmp_class = (_classification_mb[real_index] - lower_d) / (upper_d - lower_d);
	// clamp to [0,1]
	if (tmp_class>1.f)
		tmp_class = 1.f;
	else if (tmp_class<0.f)
		tmp_class = 0.f;
	// update criterion
	criterion += weight_mb * tmp_class;

	// PF: clamp to [0,1], .5 is cutoff_pf
	upper_d = _max_c_pf*_max_conf_pf - cutoff_pf;
	lower_d = cutoff_pf - _min_c_pf*_min_conf_pf;

	if (lower_d > upper_d) {
		lower_d = cutoff_pf - upper_d;
		upper_d = _max_c_pf*_max_conf_pf;
	} else {
		upper_d = cutoff_pf + lower_d;
		lower_d = _min_c_pf*_min_conf_pf;
	}
	// get clamped classification, may exceed.
	tmp_class = (_classification_pf[real_index] - lower_d) / (upper_d - lower_d);
	// clamp to [0,1]
	if (tmp_class>1.f)
		tmp_class = 1.f;
	else if (tmp_class<0.f)
		tmp_class = 0.f;
	// update criterion
	criterion += weight_pf * tmp_class;

	// BI: clamp to [0,1], .5 is cutoff
	upper_d = _max_c_bi - cutoff_bi;
	lower_d = cutoff_bi - _min_c_bi;

	if (lower_d > upper_d) {
		lower_d = cutoff_bi - upper_d;
		upper_d = _max_c_bi;
	} else {
		upper_d = cutoff_bi + lower_d;
		lower_d = _min_c_bi;
	}
	// get clamped classification, may exceed [0,1] but 0.5 is midpoint 0.5
	tmp_class = (_classification_bi[real_index] - lower_d) / (upper_d - lower_d);
	// clamp to [0,1]
	if (tmp_class>1.f)
		tmp_class = 1.f;
	else if (tmp_class<0.f)
		tmp_class = 0.f;
	// update criterion
	criterion += weight_bi * tmp_class;
	
	criterion /= weight_mb + weight_pf + weight_bi;

	return criterion;

}

bool OutlierClassificator::colorSurfelsCutoffALL(float weight_mb, float cutoff_mb, 
	                                             float weight_pf, float cutoff_pf,
						                         float weight_bi, float cutoff_bi) {
	QApplication::setOverrideCursor(Qt::waitCursor);

	Scene *scene;
	Selector *selector;
	SurfelInterface *surfel;
	std::vector<unsigned int> *selection;
	int index;
	int real_index;
	float criterion;

	scene = Scene::getInstance();
	selector = Selector::getInstance();
	selection = selector->getSelectionIndicesStdVec(Selector::intToFlag(_selection_index));
	
	restoreColor();
	_surfels.clear();
	_old_diffuse.clear();
	_old_specular.clear();
	
	for (index=0; index<selection->size(); index++) {
		real_index = (*selection)[index];
		surfel = selector->getSurfel(real_index);
		// save old color
		_surfels.push_back(surfel);
		_old_diffuse.push_back(surfel->getDiffuseColor());
		_old_specular.push_back(surfel->getSpecularColor());
	
		criterion = getCriterion(weight_mb, cutoff_mb,
			                     weight_pf, cutoff_pf,
								 weight_bi, cutoff_bi,
								 real_index);
		
		if (criterion > 0.5f) {
			scene->setSurfelDiffuseColor(surfel, qRgb(255,0,0),false);
			scene->setSurfelSpecularColor(surfel, qRgb(255,0,0), false);
			scene->setSurfelFlagOn(surfel, SurfelInterface::EMPHASISE, false, false);
		} else {
			scene->setSurfelDiffuseColor(surfel, qRgb(255,255,255),false);
			scene->setSurfelSpecularColor(surfel, qRgb(255,255,255), false);
			scene->setSurfelFlagOn(surfel, SurfelInterface::EMPHASISE, false, false);
		}

	}
	scene->emitObjectModifiedSignal();
	scene->emitRenderFrameSignal();
	
	_MCdelete((unsigned long)selection);
	delete selection;
	
	QApplication::restoreOverrideCursor();
	return true;

	
}

UndoAction* OutlierClassificator::deleteSurfelsCutoffALL(float weight_mb, float cutoff_mb, 
														 float weight_pf, float cutoff_pf,
		                       							 float weight_bi, float cutoff_bi) {
	QApplication::setOverrideCursor(Qt::waitCursor);
	Scene *scene;
	Selector *selector;
	SurfelInterface *surfel;
	std::vector<unsigned int> *selection;
	std::vector<unsigned int> del_surfels;
	int index;
	int real_index;
	int del_count;
	float criterion;

	restoreColor();
	_surfels.clear();
	_old_diffuse.clear();
	_old_specular.clear();

	scene = Scene::getInstance();
	selector = Selector::getInstance();
	selection = selector->getSelectionIndicesStdVec(Selector::intToFlag(_selection_index));

	del_count = 0;
	for (index=0; index<selection->size(); index++) {
		real_index = (*selection)[index];
		surfel = selector->getSurfel(real_index);
		
		criterion = getCriterion(weight_mb, cutoff_mb,
			                     weight_pf, cutoff_pf,
								 weight_bi, cutoff_bi,
								 real_index);
		
		if (criterion > 0.5f) {
			del_surfels.push_back(real_index);
			//scene->setSurfelDiffuseColor(surfel, qRgb(255,0,0),false);
			//scene->setSurfelSpecularColor(surfel, qRgb(255,0,0), false);
			del_count++;
		}
	}

	UndoAction *undo = selector->deleteSurfelsUndo(del_surfels);
	
	_MCdelete((unsigned long)selection);
	delete selection;

	QApplication::restoreOverrideCursor();
		
	//return undo;
	return 0;

}

void OutlierClassificator::getCutoffs(float percentage, float &cutoff_mb, float &cutoff_pf, float &cutoff_bi) const {
	std::vector<unsigned int> *selection;
	unsigned int real_index;
	unsigned int size;
	unsigned int stop_index;
	
	float *values_mb, *values_pf, *values_bi;

	selection = Selector::getInstance()->getSelectionIndicesStdVec(Selector::intToFlag(_selection_index));
	size = selection->size();
	values_mb = new float[size];
	_MCnew( (unsigned long)values_mb, sizeof(values_mb), __FILE__, __LINE__);
	
	values_pf = new float[size];
	_MCnew( (unsigned long)values_pf, sizeof(values_pf), __FILE__, __LINE__);

	values_bi = new float[size];
	_MCnew( (unsigned long)values_bi, sizeof(values_bi), __FILE__, __LINE__);

	for (unsigned int i=0; i<size; i++) {
		real_index = (*selection)[i];
		values_mb[i] = _classification_mb[real_index];
		values_pf[i] = _classification_pf[real_index] * _confidence_pf[real_index];
		values_bi[i] = _classification_bi[real_index];
	}

	std::sort(values_mb, values_mb + size);
	std::sort(values_pf, values_pf + size);
	std::sort(values_bi, values_bi + size);
	
	stop_index = (float)(1.f - percentage) * (float)size;
	
	cutoff_mb = values_mb[stop_index];
	cutoff_bi = values_bi[stop_index];
	cutoff_pf = values_pf[stop_index];

	_MCdelete((unsigned int)values_mb);
	delete[] values_mb;
	_MCdelete((unsigned int)values_bi);
	delete[] values_bi;
	_MCdelete((unsigned int)values_pf);
	delete[] values_pf;
	_MCdelete((unsigned int)selection);
	delete selection;
}


class ValueIndex {
public:
	unsigned int index;
	float value;
	ValueIndex(unsigned int i, float v) { index = i; value = v; }
	bool operator< (const ValueIndex& a) const { return value < a.value; }
};

bool OutlierClassificator::colorSurfelsPercentALL(float percentage, float weight_mb, float weight_pf, float weight_bi) {
	QApplication::setOverrideCursor(Qt::waitCursor);

	Scene *scene;
	Selector *selector;
	SurfelInterface *surfel;
	std::vector<unsigned int> *selection;
	std::vector<ValueIndex> index_values;

	int index;
	int real_index;
	float criterion;

	float cutoff_mb, cutoff_pf, cutoff_bi;
	float stop_index;
	float delete_criterion;

	
	scene = Scene::getInstance();
	selector = Selector::getInstance();
	selection = selector->getSelectionIndicesStdVec(Selector::intToFlag(_selection_index));
	
	restoreColor();
	_surfels.clear();
	_old_diffuse.clear();
	_old_specular.clear();

	getCutoffs(percentage,cutoff_mb, cutoff_pf, cutoff_bi);
	index_values.clear();	
	for (index=0; index<selection->size(); index++) {
		real_index = (*selection)[index];
		surfel = selector->getSurfel(real_index);
		
		criterion = getCriterion(weight_mb, cutoff_mb,
			                     weight_pf, cutoff_pf,
								 weight_bi, cutoff_bi,
								 real_index);
		index_values.push_back(ValueIndex(real_index, criterion));
	}

 
	std::sort(index_values.begin(), index_values.end());
	stop_index = (1.f - percentage) * (float)index_values.size();
	delete_criterion = index_values[(int)stop_index].value;

	for (index=0; index<selection->size(); index++) {
		real_index = index_values[index].index;
		surfel = selector->getSurfel(real_index);

		// save old color
		_surfels.push_back(surfel);
		_old_diffuse.push_back(surfel->getDiffuseColor());
		_old_specular.push_back(surfel->getSpecularColor());
	
		if (index_values[index].value >= delete_criterion) {
			scene->setSurfelDiffuseColor(surfel, qRgb(255,0,0),false);
			scene->setSurfelSpecularColor(surfel, qRgb(255,0,0), false);
			scene->setSurfelFlagOn(surfel, SurfelInterface::EMPHASISE, false, false);
		} else {
			scene->setSurfelDiffuseColor(surfel, qRgb(255,255,255),false);
			scene->setSurfelSpecularColor(surfel, qRgb(255,255,255), false);
			scene->setSurfelFlagOn(surfel, SurfelInterface::EMPHASISE, false, false);
		}
		
	}

	scene->emitObjectModifiedSignal();
	scene->emitRenderFrameSignal();
	
	_MCdelete((unsigned long)selection);
	delete selection;
	
	QApplication::restoreOverrideCursor();
	return true;	
}

UndoAction* OutlierClassificator::deleteSurfelsPercentALL(float percentage, float weight_mb, float weight_pf, float weight_bi) {
	QApplication::setOverrideCursor(Qt::waitCursor);

	Scene *scene;
	Selector *selector;
	SurfelInterface *surfel;
	std::vector<unsigned int> *selection;
	std::vector<ValueIndex> index_values;
	
	std::vector<unsigned int> del_surfels;
	int del_count;

	int index;
	int real_index;
	float criterion;

	float cutoff_mb, cutoff_pf, cutoff_bi;
	float stop_index;
	float delete_criterion;

	
	scene = Scene::getInstance();
	selector = Selector::getInstance();
	selection = selector->getSelectionIndicesStdVec(Selector::intToFlag(_selection_index));
	
	restoreColor();
	_surfels.clear();
	_old_diffuse.clear();
	_old_specular.clear();

	getCutoffs(percentage,cutoff_mb, cutoff_pf, cutoff_bi);
	index_values.clear();	
	for (index=0; index<selection->size(); index++) {
		real_index = (*selection)[index];
		surfel = selector->getSurfel(real_index);
		
		criterion = getCriterion(weight_mb, cutoff_mb,
			                     weight_pf, cutoff_pf,
								 weight_bi, cutoff_bi,
								 real_index);
		index_values.push_back(ValueIndex(real_index, criterion));
	}

	std::sort(index_values.begin(), index_values.end());
	stop_index = (1.f - percentage) * (float)index_values.size();
	delete_criterion = index_values[(int)stop_index].value;
	del_count =0;
	for (index=0; index<selection->size(); index++) {
		real_index = index_values[index].index;
		surfel = selector->getSurfel(real_index);

		// save old color
		_surfels.push_back(surfel);
		_old_diffuse.push_back(surfel->getDiffuseColor());
		_old_specular.push_back(surfel->getSpecularColor());
	
		if (index_values[index].value >= delete_criterion) {
			del_surfels.push_back(real_index);
			del_count++;
		} 
	}

	UndoAction *undo = selector->deleteSurfelsUndo(del_surfels);
	QApplication::restoreOverrideCursor();
	scene->emitObjectModifiedSignal();
	scene->emitRenderFrameSignal();

	_MCdelete((unsigned long)selection);
	delete selection;
#if 1
	delete undo;
	return 0;	
#else
	return undo;
#endif
}


bool OutlierClassificator::restoreColor() {
	QApplication::setOverrideCursor(Qt::waitCursor);
	Scene *scene;
	int size;
	int i;
	
	scene = Scene::getInstance();
	size = _surfels.size();
	if (size) {
		for (i=0; i<size; i++) {
			scene->setSurfelDiffuseColor(_surfels[i], _old_diffuse[i], false);
			scene->setSurfelSpecularColor(_surfels[i], _old_specular[i], false);
			scene->setSurfelFlagOn(_surfels[i], SurfelInterface::EMPHASISE, true, false);
		}
	}

	scene->emitRenderFrameSignal();
	scene->emitObjectModifiedSignal();
	QApplication::restoreOverrideCursor();
	return true;
}

bool OutlierClassificator::testColor(int selection_index, int mode, float r, float g, float b, float coeff) {
	SurfelInterface *surfel;
	Selector *selector;
	Scene *scene;

	selector = Selector::getInstance();
	selector->buildTree();
	_indices = selector->getSelectionIndicesStdVec(Selector::intToFlag(_selection_index));
	
	// check if we got a selection
	int selection_count = _indices->size();
	if (selection_count == 0) {
		return false;
	}
	
	scene = Scene::getInstance();

	for (int i=0; i<selection_count; i++) {
		surfel = selector->getSurfel( (*_indices)[i]);
		if (mode&1) 
			scene->setSurfelDiffuseColor(surfel, qRgb(coeff*r, coeff*g, coeff*b),false);
		if (mode&2)
			scene->setSurfelSpecularColor(surfel, qRgb(coeff*r, coeff*g, coeff*b),false);
	}
	return true;
}


float OutlierClassificator::mbMaxC(void) {
	return _max_c_mb;
}

float OutlierClassificator::mbMinC(void) {
	return _min_c_mb;
}

float OutlierClassificator::pfMaxC(void) {
	return _max_c_pf;
}

float OutlierClassificator::pfMinC(void) {
	return _min_c_pf;
}

float OutlierClassificator::pfMaxConf(void) {
	return _max_conf_pf;
}

float OutlierClassificator::pfMinConf(void) {
	return _min_conf_pf;
}

float OutlierClassificator::biMaxC(void) {
	return _max_c_bi;
}

float OutlierClassificator::biMinC(void) {
	return _min_c_bi;
}

/*
#ifdef _WIN32
	#pragma warning(pop)
#endif // _WIN32
*/  


// Some Emacs-Hints -- please don't remove:
//
//  Local Variables:
//  mode:C++
//  tab-width:4
//  End:
