// Title:   ScanTools.cpp
// Created: Thu Sep 25 15:15:46 2003
// Authors: Simon Heinzle, Tim Weyrich
//
// Copyright (c) 2001, 2002, 2003, Computer Graphics Lab, ETH Zurich
//
// This file is part of the Pointshop3D system.
// See http://www.pointshop3d.com/ for more information.
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later
// version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
// 
// You should have received a copy of the GNU Lesser General
// Public License along with this library; if not, write to the
// Free Software Foundation, Inc., 59 Temple Place, Suite 330,
// Boston, MA 02111-1307 USA
// 
// Contact info@pointshop3d.com if any conditions of this
// licensing are not clear to you.
//

#ifdef _WIN32
	// turn off std::map warnings
	//# pragma warning(push)
	# pragma warning(disable:4786) // identifier was truncated to 25 char...
#endif

#include <qtoolbutton.h>
#include <qtoolbar.h>
#include <qcursor.h>
#include <qevent.h>
#include <qpixmap.h>
#include <qwhatsthis.h>
#include <qwidget.h>
#include <qpainter.h>
#include <qcolor.h>
#include <qlineedit.h>
#include <string.h>
#include <qlistbox.h>
#include <qtabwidget.h>
#include <qpushbutton.h>
#include <qcheckbox.h>
#include <qslider.h>
#include <qlabel.h>
#include <qradiobutton.h>

#include "../../../../Core/RendererManager/src/RendererManager.h"
#include "../../../../Core/Interfaces/src/ToolInterface.h"
#include "../UI/ScanToolsWidget.h"
#include "../UI/ScanToolsConfigurationWidget.h"

#include "ScanTools.h"
#include "ScanToolsGlobals.h"

#ifndef _WIN32
#include <stdlib.h>
#define _gcvt gcvt
#endif

// ***************
// * tool includes
#include "Selector.h"
#include "Deletor.h"
#include "RandomNoise.h"
#include "MovingLeastSquaresSmoother.h"
#include "MLSScatterer.h"
#include "LocalParticleSimulation.h"
#include "OutlierClassificator.h"
#include "BadNormalRemover.h"
#include "SurfelResizer.h"
#include "ColorModifier.h"
#include "Help.h"



#include "NewOverload.h"

// ******************
// 'What's this' text
// ******************

static const char *whatsThisText = "<img source=\"template\"> "
"Click this button to activate the Scan Tool Set"
"<ul><li>Scan Tool Set is designed for processing 3d Scans</li>";

// *****************************************************
// Help Text
// *****************************************************

QString selectionString = "Here you can set selections and perform various operations by clicking on the \" * on selection\"  button of the requested operation. (most operations can be applied directly)";
QString deletorString = "Here you can delete surfels either directly (in Deletor mode) with the box selector or you can delete the current selection (in Selection mode) by pressing the \"delete current selection\" button.";
QString noiseString = "Here you can add noise to the object either directly (in Noise mode) with the box selector or on current selection (in Selection mode) by pressing the \"add noise to selection\" button.";
QString mlsString = "Here you can smooth to the object either directly (in MLS Smoother mode) with the box selector or smooth the current selection (in Selection mode) by pressing the \"run on selection button\". \n "\
	"Bigger gauss factor leads to more smoothing. The bigger the gauss factor, the more points influence the smoothing, so for very big gauss factors i recommend increasing the k next neighbour count.\n"\
	"Alpha defines the influence of the smoothing -> (1-a) * old position + a * new position.\n"\
	"Degree of polynom defines the degree of the resulting fitted polynom (must be greater than 0). Use only planefit is equivalent to polynom degree 0.\n";
QString scattererString = "Here you can scatter points directly by using the box selector, for increasing the sampling density and / or fill holes.\n"\
	"MLS gauss param defines the projection parameter, projection is done via MLS (see MLS smoother).";
QString lpsString = "This tool is used to redistribute the surfels on the surface.\n"\
	"Mark the surfels you want to redistribute, but be aware that there must remain some border surfels. Otherwise the surfels will fly away from the object.\n"\
	"If you experience troubles with this tool, try to adjust the \"max query distance\" and \"force scale\", which should be kept low to keep the surfels near the surface.";
QString outlierString = "This tool classifies & removes outliers. Can be performed either directly using the box selector or on the current selection using the classify selection button.\n"\
	"After classifying, adjust the cutoff values (the lower the more surfels will be deleted) , which can also be done automatically, or and the weights of each method. \"Preview\" shows the surfels to be deleted in red, \"Delete\" removes them.\n"\
	"NOTE: kNext Graph often needs to be rebuilt if:\n"\
	"- Object selection changed, Objects added / removed\n"\
	"- A lot of surfels were deleted\n"\
	"Note also, that this can require a lot of time.";
QString normalsString = "This tool removes bad normals by comparing neighbouring normals.\n"\
	"Select a small area where no bad normals occur. Then click on the remove button to remove bad normals from the whole object.";
QString resizeString = "This tool adjusts the size of the object.\n"\
	"Specify the length of the greates bounding box edge, then click on \"resize\" to resize all objects";
QString recolorString = "This tool changes / filters the color of surfels, either directly using the box selector or on the current selection by pressing the \"run on selection\" button.";

// ************
// Box selector
//  this global definition is due to callback passing
// ************

BoxSelector box;

void paint_box_callback() {
	box.paintBox();
}

// **********
// Tool class
// **********

ScanTools::ScanTools() {
	toolCursor = new QCursor (CrossCursor);
	_MCnew( (unsigned long)toolCursor, sizeof(toolCursor), __FILE__, __LINE__);

	scantoolsDialog = ScanToolsDialogImpl::getInstance();
	scantoolsConfigurationDialog = ScanToolsConfigurationDialogImpl::getInstance();
	scantoolsBoxColorDialog = ScanToolsBoxColorDialogImpl::getInstance();
	
	rendererGL = (OpenGLRenderer*)RendererManager::getInstance()->getPreviewRenderer();
	
	undoRedoStack = new ActionManager(4);
	_MCnew( (unsigned long)undoRedoStack, sizeof(undoRedoStack), __FILE__, __LINE__);

	ocl = 0;
	box.reset();
	showGL();
	controlPressed = false;
	shiftPressed = false;
	altPressed = false;

#ifdef _DEBUG
	scantoolsDialog->memleak->setEnabled(true);
#endif
	// pass keys from widget
	this->connect (scantoolsDialog, SIGNAL (passKeyPressEvent(QKeyEvent*)),
				   this, SLOT (handleKeyPressEventFromDialog(QKeyEvent*)));
	this->connect (scantoolsDialog, SIGNAL (passKeyReleaseEvent(QKeyEvent*)),
				   this, SLOT (handleKeyReleaseEventFromDialog(QKeyEvent*)));
}

ScanTools::~ScanTools() {
	_MCdelete((unsigned int)toolCursor);
	delete toolCursor;
	ScanToolsConfigurationDialogImpl::destroyInstance();
	ScanToolsDialogImpl::destroyInstance();
	ScanToolsBoxColorDialogImpl::destroyInstance();
	hideGL();
	MemChecker::getInstance()->DumpUnfreed();
}

QToolButton *ScanTools::addButtonToToolBar (QToolBar *toolBar) {
	QPixmap markIcon ("Resources/SCT_MarkIcon.png"),
	        deleteIcon("Resources/SCT_DeleteIcon.png"),
			noiseIcon("Resources/SCT_NoiseIcon.png"),
			mlsIcon("Resources/SCT_MLSIcon.png"),
			scatterIcon("Resources/SCT_ScatterIcon.png"),
			colorIcon("Resources/SCT_RecolorIcon.png"),
			outlierIcon("Resources/SCT_OutlierIcon.png"),
			normalsIcon("Resources/SCT_FalseNormalsIcon.png"),
			resizeIcon("Resources/SCT_ResizerIcon.png"),
			lpsIcon("Resources/SCT_LPSIcon.png");

	// create icon sets
	markSet.setPixmap(markIcon, QIconSet::Automatic, QIconSet::Normal);
	deleteSet.setPixmap(deleteIcon, QIconSet::Automatic, QIconSet::Normal);
	noiseSet.setPixmap(noiseIcon, QIconSet::Automatic, QIconSet::Normal);
	mlsSet.setPixmap(mlsIcon, QIconSet::Automatic, QIconSet::Normal);
	scatterSet.setPixmap(scatterIcon, QIconSet::Automatic, QIconSet::Normal);
	colorSet.setPixmap(colorIcon, QIconSet::Automatic, QIconSet::Normal);
	outlierSet.setPixmap(outlierIcon, QIconSet::Automatic, QIconSet::Normal);
	normalsSet.setPixmap(normalsIcon, QIconSet::Automatic, QIconSet::Normal);
	resizeSet.setPixmap(resizeIcon, QIconSet::Automatic, QIconSet::Normal);
	lpsSet.setPixmap(lpsIcon, QIconSet::Automatic, QIconSet::Normal);

	// add the button to the toolbar
	toolButton = new QToolButton (markSet, "Scan Cleaning Tool Set", "Scan Cleaning Tool Set for processing 3d Scans",
	                              0, 0, toolBar, "Scan Cleaning Tools");
	_MCnew( (unsigned long)toolButton, sizeof(toolButton), __FILE__, __LINE__);
	toolButton->setToggleButton(true);
	
	QWhatsThis::add (toolButton, whatsThisText);
	QMimeSourceFactory::defaultFactory()->setPixmap ("scan cleaning tools", markSet.pixmap(QIconSet::Automatic, QIconSet::Normal));

	markID = scantoolsMenu.insertItem(markSet.pixmap(QIconSet::Automatic, QIconSet::Normal), this, SLOT(handleToolSelected(int)));
	deleteID = scantoolsMenu.insertItem(deleteSet.pixmap(QIconSet::Automatic, QIconSet::Normal), this, SLOT(handleToolSelected(int)));
	noiseID = scantoolsMenu.insertItem(noiseSet.pixmap(QIconSet::Automatic, QIconSet::Normal), this, SLOT(handleToolSelected(int)));
	mlsID = scantoolsMenu.insertItem(mlsSet.pixmap(QIconSet::Automatic, QIconSet::Normal), this, SLOT(handleToolSelected(int)));
	scatterID = scantoolsMenu.insertItem(scatterSet.pixmap(QIconSet::Automatic, QIconSet::Normal), this, SLOT(handleToolSelected(int)));
	lpsID = scantoolsMenu.insertItem(lpsSet.pixmap(QIconSet::Automatic, QIconSet::Normal), this, SLOT(handleToolSelected(int)));
	outlierID = scantoolsMenu.insertItem(outlierSet.pixmap(QIconSet::Automatic, QIconSet::Normal), this, SLOT(handleToolSelected(int)));
	normalsID = scantoolsMenu.insertItem(normalsSet.pixmap(QIconSet::Automatic, QIconSet::Normal), this, SLOT(handleToolSelected(int)));
	resizeID = scantoolsMenu.insertItem(resizeSet.pixmap(QIconSet::Automatic, QIconSet::Normal), this, SLOT(handleToolSelected(int)));
	colorID = scantoolsMenu.insertItem(colorSet.pixmap(QIconSet::Automatic, QIconSet::Normal), this, SLOT(handleToolSelected(int)));
	
	currentTool = markID;

	toolButton->setEnabled(true);
	
	isToolSetSelected = false;
	
	return toolButton;
}

QToolButton *ScanTools::getToolButton() {
	return toolButton;
}

QCursor *ScanTools::getToolCursor() const {
	return toolCursor;
}

QWidget *ScanTools::getToolConfigurationWidget() const {
	//return scantoolsConfigurationDialog;
	return scantoolsBoxColorDialog;
}

bool ScanTools::isAvailable() const {
	return this->checkToolAvailability();
}

void ScanTools::handleKeyPressEvent (QKeyEvent *ke) {
	if (ke->key() == Key_Control)
		controlPressed = true;
	if (ke->key() == Key_Shift)
		shiftPressed = true;
	if (ke->key() == Key_Alt)
		altPressed = true;
	
	if (controlPressed) {
		if (ke->key() == Key_Z)
			undoRedoStack->undo();
		else if (ke->key() == Key_Y)
			undoRedoStack->redo();
	} 

	if (controlPressed && altPressed) {
		if (ke->key() == Key_1)
			scantoolsDialog->Selection1RadioButton->setChecked(true);
		else if (ke->key() == Key_2)
			scantoolsDialog->Selection2RadioButton->setChecked(true);
		else if (ke->key() == Key_3)
			scantoolsDialog->Selection3RadioButton->setChecked(true);
		else if (ke->key() == Key_E)
			handleSelectAllButtonReleased();
		else if (ke->key() == Key_D)
			handleDeselectButtonReleased();
		else if (ke->key() == Key_O) {
			scantoolsDialog->AlignButton->setChecked(true);
			box.setMouseMode(MOUSE_ALIGN);
		} else if (ke->key() == Key_F) {
			scantoolsDialog->FreeButton->setChecked(true);
			box.setMouseMode(MOUSE_FREEMOVE);
		} else if (ke->key() == Key_P) {
			scantoolsDialog->PickingButton->setChecked(true);
			box.setMouseMode(MOUSE_PICKING);
		} else if (ke->key() == Key_F2) 
			scantoolsDialog->TabWidget->showPage(scantoolsDialog->tab);
		else if (ke->key() == Key_F3) 
			scantoolsDialog->TabWidget->showPage(scantoolsDialog->tab_2);
		else if (ke->key() == Key_F4) 
			scantoolsDialog->TabWidget->showPage(scantoolsDialog->tab_3);
		else if (ke->key() == Key_F5) 
			scantoolsDialog->TabWidget->showPage(scantoolsDialog->tab_4);
		else if (ke->key() == Key_F6) 
			scantoolsDialog->TabWidget->showPage(scantoolsDialog->tab_5);
		else if (ke->key() == Key_F7) 
			scantoolsDialog->TabWidget->showPage(scantoolsDialog->tab_6);
		else if (ke->key() == Key_F8) 
			scantoolsDialog->TabWidget->showPage(scantoolsDialog->tab_7);
		else if (ke->key() == Key_F9) 
			scantoolsDialog->TabWidget->showPage(scantoolsDialog->tab_8);
		else if (ke->key() == Key_F10) 
			scantoolsDialog->TabWidget->showPage(scantoolsDialog->tab_9);
		else if (ke->key() == Key_F11) 
			scantoolsDialog->TabWidget->showPage(scantoolsDialog->tab_10);
		else if (ke->key() == Key_8) {
			scantoolsDialog->ShapeBoxButton->setChecked(true);
			box.setShape(SHAPE_BOX);

		} else if (ke->key() == Key_9 ) {
			scantoolsDialog->ShapeSphereButton->setChecked(true);
			box.setShape(SHAPE_SPHERE);
		}

	}
	
	if (controlPressed && !altPressed ) {
		if (ke->key() == Key_F2) {
			scantoolsDialog->TabWidget->showPage(scantoolsDialog->tab);
			handleToolSelected(markID);
		} else if (ke->key() == Key_F3) {
			scantoolsDialog->TabWidget->showPage(scantoolsDialog->tab_2);
			handleToolSelected(deleteID);
		} else if (ke->key() == Key_F4) {
			scantoolsDialog->TabWidget->showPage(scantoolsDialog->tab_3);
			handleToolSelected(noiseID);
		} else if (ke->key() == Key_F5) {
			scantoolsDialog->TabWidget->showPage(scantoolsDialog->tab_4);
			handleToolSelected(mlsID);
		} else if (ke->key() == Key_F6) {
			scantoolsDialog->TabWidget->showPage(scantoolsDialog->tab_5);
			handleToolSelected(scatterID);
		} else if (ke->key() == Key_F7) {
			scantoolsDialog->TabWidget->showPage(scantoolsDialog->tab_6);
			handleToolSelected(lpsID);
		} else if (ke->key() == Key_F8) {
			scantoolsDialog->TabWidget->showPage(scantoolsDialog->tab_7);
			handleToolSelected(outlierID);
		} else if (ke->key() == Key_F9) {
			scantoolsDialog->TabWidget->showPage(scantoolsDialog->tab_8);
			handleToolSelected(normalsID);
		} else if (ke->key() == Key_F10) {
			scantoolsDialog->TabWidget->showPage(scantoolsDialog->tab_9);
			handleToolSelected(resizeID);
		} else if (ke->key() == Key_F11) {
			scantoolsDialog->TabWidget->showPage(scantoolsDialog->tab_10);
			handleToolSelected(colorID);
		} 
	}

	ke->ignore();
}

void ScanTools::handleKeyReleaseEvent (QKeyEvent *ke) {
	if (ke->key() == Key_Control)
		controlPressed = false;
	if (ke->key() == Key_Shift)
		shiftPressed = false;
	if (ke->key() == Key_Alt)
		altPressed = false;


	ke->ignore();
}

void ScanTools::handleMousePressEvent (QMouseEvent *me) {
	if (!((me->state()&ShiftButton) || (me->state()&ControlButton) || (me->state()&AltButton))) {
		
		if ((currentTool == markID) || (currentTool == lpsID) || (currentTool == normalsID)  ) 
			box.setSelectionIndex(getCurrentSelectionIndex());
		else { 
			Selector::getInstance()->deselectIndex(3);
			box.setSelectionIndex(3);
		}
	}
		
	
	box.handleMousePress(me->x(), me->y(), stateToInt(me->stateAfter()));
	rendererGL->renderFrame();
}

void ScanTools::handleMouseReleaseEvent (QMouseEvent *me) {
	box.handleMouseRelease(me->x(), me->y(), stateToInt(me->state()));
	
	if ( (me->state()&ShiftButton) || (me->state()&ControlButton) || (me->state()&AltButton) )
		return;
	if (! Scene::getInstance()->getActiveObject())
		return;
	if (scantoolsDialog->PickingButton->isChecked()) 
		return;

	if (currentTool == deleteID) {
		Deletor deletor;
		deletor.setSelection(3);
		undoRedoStack->addUndoAction(deletor.executeUndo());
	} else if (currentTool == noiseID) {
		RandomNoise noise;
		bool isnumber;
		noise.setSelection(3);
		noise.setParameters(scantoolsDialog->RandomNoiseEdit->text().toFloat(&isnumber));
		if (isnumber)
			undoRedoStack->addUndoAction(noise.executeUndo());
	} else if (currentTool == mlsID) {
		MovingLeastSquaresSmoother mls;
		bool isnumber1, isnumber2, isnumber3, isnumber4;
		mls.setSelection(3);
		mls.setParameters( scantoolsDialog->MLSnEdit->text().toFloat(&isnumber1),
			scantoolsDialog->MLSgaussEdit->text().toFloat(&isnumber2),
			scantoolsDialog->MLSalphaEdit->text().toFloat(&isnumber3),
			scantoolsDialog->MLSdoplaneBox->isChecked(),
			scantoolsDialog->MLSdegreeEdit->text().toFloat(&isnumber4));
		if (isnumber1 && isnumber2 && isnumber3 && isnumber4)
			undoRedoStack->addUndoAction(mls.executeUndo());
		Selector::getInstance()->deselectIndex(3);
	} else if (currentTool == outlierID) {
		QString string;
		char buf[255];
		if (!ocl)
			return;
		if (ocl->doMiniballClassification(3)) {
			string  = "Miniball, max = ";
			_gcvt(ocl->mbMaxC(), 10, buf);
			string += buf;
			scantoolsDialog->CombinerMBBox->setTitle(string);
		} else
			scantoolsDialog->CombinerMBBox->setTitle("Miniball, error.");
		if (ocl->doPlaneClassification(3)) {
			string  = "Planefit, max = ";
			_gcvt(ocl->pfMaxC(), 10, buf);
			string += buf;
			scantoolsDialog->CombinerPFBox->setTitle(string);
		} else
			scantoolsDialog->CombinerPFBox->setTitle("Miniball, error.");
		if (ocl->doBidirectionClassification(3)) {
			string  = "Bidirectional, max = ";
			_gcvt(ocl->biMaxC(), 10, buf);
			string += buf;
			scantoolsDialog->CombinerBIBox->setTitle(string);
		} else {
			string = "Miniball, error.";
			scantoolsDialog->CombinerBIBox->setTitle(string);
		}
	} else if (currentTool == scatterID && (me->state() == LeftButton)) {
		MLSScatterer mls;

		Vector3D lowerbound(-fabs(box.getEdgeLength()[0]/2.f),-fabs(box.getEdgeLength()[1]/2.f),-fabs(box.getEdgeLength()[2]/2.f));
		Vector3D upperbound(fabs(box.getEdgeLength()[0]/2.f),fabs(box.getEdgeLength()[1]/2.f),fabs(box.getEdgeLength()[2]/2.f));
		
		lowerbound += box.getCenter();
		upperbound += box.getCenter();
		mls.setSelection(3);
		mls.setParameters(scantoolsDialog->ScatterCountEdit->text().toInt(), 30);
		mls.setMLSGauss(scantoolsDialog->MLSParam->text().toFloat());
		mls.useMLS(false);
		//mls.getNewPositionsAxisAlignedBox(lowerbound, upperbound);
		mls.getNewPositionsBox(&box);
		mls.setDensityCheck(scantoolsDialog->HolesCheckBox->isChecked());
		undoRedoStack->addUndoAction(mls.executeUndo());
	} else if (currentTool == colorID) {
		ColorModifier cm;
		QColor diff,spec;
		bool s1,s2,s3;

		cm.setSelection(3);
		if (scantoolsDialog->UniColorButton->isChecked()) 
			cm.setAction(COLOR_UNI);
		else
			cm.setAction(COLOR_FILTER);

		cm.setParamsFilter(scantoolsDialog->KnextEdit->text().toInt(&s1),
			scantoolsDialog->FilterCoefficient->text().toFloat(&s2));
		if (!(s1&&s2))
			return;

		spec.setRgb(255.0*scantoolsDialog->SpecREdit->text().toFloat(&s1),255.0*scantoolsDialog->SpecGEdit->text().toFloat(&s2),255.0*scantoolsDialog->SpecBEdit->text().toFloat(&s3));
		if (! (s1 && s2 && s3))
			return;

		diff.setRgb(255.0*scantoolsDialog->DiffREdit->text( ).toFloat(&s1),255.0*scantoolsDialog->DiffGEdit->text().toFloat(&s2),255.0*scantoolsDialog->DiffBEdit->text().toFloat(&s3));
		if (! (s1 && s2 && s3))
			return;

		cm.setParamsUni(diff.rgb(), spec.rgb());

		undoRedoStack->addUndoAction(cm.executeUndo());
	}
	rendererGL->renderFrame();
}


void ScanTools::handleMouseMoveEvent (QMouseEvent *me) {
	
	box.handleMouseMovementMidpoint(me->x(), me->y(), stateToInt(me->stateAfter()));
	rendererGL->renderFrame();
}

void ScanTools::selectTool() {
	// set menu
	toolButton->setPopup(&scantoolsMenu);
	// update the dialog
	updateWindow();
	// switch to preview renderer
	RendererManager::getInstance()->setPreviewRendererEnabled(true,true);
	// show box
	showGL();
	// show dialog
	scantoolsDialog->show();
	scantoolsDialog->raise();
	// turn mouse tracking on
	rendererGL->getRendererWidget()->setMouseTracking (true);
	// establish connections
	connectSelectionOptions();
	connectOther();

	// update object list
	updateObjects();

	// set scantoolsdialog apropriate
	currentTool = markID;
	toolButton->setIconSet(markSet);
	// enable / disable tabs
	scantoolsDialog->TabWidget->setTabEnabled(scantoolsDialog->tab,true);
	scantoolsDialog->TabWidget->setTabEnabled(scantoolsDialog->tab_2,true);
	scantoolsDialog->TabWidget->setTabEnabled(scantoolsDialog->tab_3,true);
	scantoolsDialog->TabWidget->setTabEnabled(scantoolsDialog->tab_4,true);
	scantoolsDialog->TabWidget->setTabEnabled(scantoolsDialog->tab_5,true);
	scantoolsDialog->TabWidget->setTabEnabled(scantoolsDialog->tab_6,true);
	scantoolsDialog->TabWidget->setTabEnabled(scantoolsDialog->tab_7,true);
	scantoolsDialog->TabWidget->setTabEnabled(scantoolsDialog->tab_8,true);
	scantoolsDialog->TabWidget->setTabEnabled(scantoolsDialog->tab_9, false);
	scantoolsDialog->TabWidget->setTabEnabled(scantoolsDialog->tab_10, true);
	// set tab names (to 
	// enable buttons
	scantoolsDialog->DeleteButton->setEnabled(true);
	scantoolsDialog->RandomNoiseButton->setEnabled(true);
	scantoolsDialog->MLSButton->setEnabled(true);
	scantoolsDialog->ALLClassifyButton->setEnabled(true);
	scantoolsDialog->DoColorButton->setEnabled(true);
	scantoolsDialog->SelectionBox->setEnabled(true);
	scantoolsDialog->SelectAllButton->setEnabled(true);
	scantoolsDialog->DeselectButton->setEnabled(true);
	// set apropriate page
	scantoolsDialog->TabWidget->setCurrentPage(0);

}

void ScanTools::deselectTool() {
	// turn of menu
	toolButton->setPopup(0);
	// turn mouse tracking off
	rendererGL->getRendererWidget()->setMouseTracking (false);
	// hide dialog
	scantoolsDialog->hide();
	// hide box
	hideGL();
	// break connections
	disconnectSelectionOptions();
	disconnectOther();
}


// ************
// public slots
// ************

void ScanTools::handleKeyPressEventFromDialog(QKeyEvent* ke) {
	handleKeyPressEvent(ke);
}

void ScanTools::handleKeyReleaseEventFromDialog(QKeyEvent* ke) {
	handleKeyReleaseEvent(ke);
}
	

void ScanTools::handleRendererWidgetRepainted() {
	// the RenderWidget has repainted its content, so here this tool
	// can draw its content with a QPainter, if necessary
}

void ScanTools::handleToolSelected(int id) {
	currentTool = id;

	if (id == markID) {
		toolButton->setIconSet(markSet);
		// enable / disable tabs
		scantoolsDialog->TabWidget->setTabEnabled(scantoolsDialog->tab,true);
		scantoolsDialog->TabWidget->setTabEnabled(scantoolsDialog->tab_2,true);
		scantoolsDialog->TabWidget->setTabEnabled(scantoolsDialog->tab_3,true);
		scantoolsDialog->TabWidget->setTabEnabled(scantoolsDialog->tab_4,true);
		scantoolsDialog->TabWidget->setTabEnabled(scantoolsDialog->tab_5,true);
		scantoolsDialog->TabWidget->setTabEnabled(scantoolsDialog->tab_6,true);
		scantoolsDialog->TabWidget->setTabEnabled(scantoolsDialog->tab_7,true);
		scantoolsDialog->TabWidget->setTabEnabled(scantoolsDialog->tab_8,true);
		scantoolsDialog->TabWidget->setTabEnabled(scantoolsDialog->tab_9, false);
		scantoolsDialog->TabWidget->setTabEnabled(scantoolsDialog->tab_10, true);
		// set tab names (to 
		// enable buttons
		scantoolsDialog->DeleteButton->setEnabled(true);
		scantoolsDialog->RandomNoiseButton->setEnabled(true);
		scantoolsDialog->MLSButton->setEnabled(true);
		scantoolsDialog->ALLClassifyButton->setEnabled(true);
		scantoolsDialog->DoColorButton->setEnabled(true);
		scantoolsDialog->SelectionBox->setEnabled(true);
		scantoolsDialog->SelectAllButton->setEnabled(true);
		scantoolsDialog->DeselectButton->setEnabled(true);
		// set apropriate page
		scantoolsDialog->TabWidget->setCurrentPage(0);
	} else if (id == deleteID) {
		toolButton->setIconSet(deleteSet);
		// enable / disable tabs
		scantoolsDialog->TabWidget->setTabEnabled(scantoolsDialog->tab,true);
		scantoolsDialog->TabWidget->setTabEnabled(scantoolsDialog->tab_2,true);
		scantoolsDialog->TabWidget->setTabEnabled(scantoolsDialog->tab_3,false);
		scantoolsDialog->TabWidget->setTabEnabled(scantoolsDialog->tab_4,false);
		scantoolsDialog->TabWidget->setTabEnabled(scantoolsDialog->tab_5,false);
		scantoolsDialog->TabWidget->setTabEnabled(scantoolsDialog->tab_6,false);
		scantoolsDialog->TabWidget->setTabEnabled(scantoolsDialog->tab_7,false);
		scantoolsDialog->TabWidget->setTabEnabled(scantoolsDialog->tab_8,false);
		scantoolsDialog->TabWidget->setTabEnabled(scantoolsDialog->tab_9, false);
		scantoolsDialog->TabWidget->setTabEnabled(scantoolsDialog->tab_10, false);
		// disable buttons etc
		scantoolsDialog->DeleteButton->setEnabled(false);
		scantoolsDialog->SelectionBox->setEnabled(false);
		scantoolsDialog->SelectAllButton->setEnabled(false);
		scantoolsDialog->DeselectButton->setEnabled(false);
		// set apropriate page
		scantoolsDialog->TabWidget->setCurrentPage(1);
	} else if (id == noiseID) {
		toolButton->setIconSet(noiseSet);
		// enable / disable tabs
		scantoolsDialog->TabWidget->setTabEnabled(scantoolsDialog->tab,true);
		scantoolsDialog->TabWidget->setTabEnabled(scantoolsDialog->tab_2,false);
		scantoolsDialog->TabWidget->setTabEnabled(scantoolsDialog->tab_3,true);
		scantoolsDialog->TabWidget->setTabEnabled(scantoolsDialog->tab_4,false);
		scantoolsDialog->TabWidget->setTabEnabled(scantoolsDialog->tab_5,false);
		scantoolsDialog->TabWidget->setTabEnabled(scantoolsDialog->tab_6,false);
		scantoolsDialog->TabWidget->setTabEnabled(scantoolsDialog->tab_7,false);
		scantoolsDialog->TabWidget->setTabEnabled(scantoolsDialog->tab_8,false);
		scantoolsDialog->TabWidget->setTabEnabled(scantoolsDialog->tab_9, false);
		scantoolsDialog->TabWidget->setTabEnabled(scantoolsDialog->tab_10, false);
		// disable buttons etc
		scantoolsDialog->RandomNoiseButton->setEnabled(false);
		scantoolsDialog->SelectionBox->setEnabled(false);
		scantoolsDialog->SelectAllButton->setEnabled(false);
		scantoolsDialog->DeselectButton->setEnabled(false);
		// set apropriate page
		scantoolsDialog->TabWidget->setCurrentPage(2);
	} else if (id == mlsID) {
		toolButton->setIconSet(mlsSet);
		// enable / disable tabs
		scantoolsDialog->TabWidget->setTabEnabled(scantoolsDialog->tab,true);
		scantoolsDialog->TabWidget->setTabEnabled(scantoolsDialog->tab_2,false);
		scantoolsDialog->TabWidget->setTabEnabled(scantoolsDialog->tab_3,false);
		scantoolsDialog->TabWidget->setTabEnabled(scantoolsDialog->tab_4,true);
		scantoolsDialog->TabWidget->setTabEnabled(scantoolsDialog->tab_5,false);
		scantoolsDialog->TabWidget->setTabEnabled(scantoolsDialog->tab_6,false);
		scantoolsDialog->TabWidget->setTabEnabled(scantoolsDialog->tab_7,false);
		scantoolsDialog->TabWidget->setTabEnabled(scantoolsDialog->tab_8,false);
		scantoolsDialog->TabWidget->setTabEnabled(scantoolsDialog->tab_9, false);
		scantoolsDialog->TabWidget->setTabEnabled(scantoolsDialog->tab_10, false);
		// disable buttons etc
		scantoolsDialog->MLSButton->setEnabled(false);
		scantoolsDialog->SelectionBox->setEnabled(false);
		scantoolsDialog->SelectAllButton->setEnabled(false);
		scantoolsDialog->DeselectButton->setEnabled(false);
		// set apropriate page
		scantoolsDialog->TabWidget->setCurrentPage(3);
	} else if (id == scatterID) {
		toolButton->setIconSet(scatterSet);
		// enable / disable tabs
		scantoolsDialog->TabWidget->setTabEnabled(scantoolsDialog->tab,true);
		scantoolsDialog->TabWidget->setTabEnabled(scantoolsDialog->tab_2,false);
		scantoolsDialog->TabWidget->setTabEnabled(scantoolsDialog->tab_3,false);
		scantoolsDialog->TabWidget->setTabEnabled(scantoolsDialog->tab_4,false);
		scantoolsDialog->TabWidget->setTabEnabled(scantoolsDialog->tab_5,true);
		scantoolsDialog->TabWidget->setTabEnabled(scantoolsDialog->tab_6,false);
		scantoolsDialog->TabWidget->setTabEnabled(scantoolsDialog->tab_7,false);
		scantoolsDialog->TabWidget->setTabEnabled(scantoolsDialog->tab_8,false);
		scantoolsDialog->TabWidget->setTabEnabled(scantoolsDialog->tab_9, false);
		scantoolsDialog->TabWidget->setTabEnabled(scantoolsDialog->tab_10, false);
		// disable buttons etc
		scantoolsDialog->SelectionBox->setEnabled(false);
		scantoolsDialog->SelectAllButton->setEnabled(false);
		scantoolsDialog->DeselectButton->setEnabled(false);
		// set apropriate page
		scantoolsDialog->TabWidget->setCurrentPage(4);
	} else if (id == lpsID){
		toolButton->setIconSet(lpsSet);
		// enable / disable tabs
		scantoolsDialog->TabWidget->setTabEnabled(scantoolsDialog->tab,true);
		scantoolsDialog->TabWidget->setTabEnabled(scantoolsDialog->tab_2,false);
		scantoolsDialog->TabWidget->setTabEnabled(scantoolsDialog->tab_3,false);
		scantoolsDialog->TabWidget->setTabEnabled(scantoolsDialog->tab_4,false);
		scantoolsDialog->TabWidget->setTabEnabled(scantoolsDialog->tab_5,false);
		scantoolsDialog->TabWidget->setTabEnabled(scantoolsDialog->tab_6,true);
		scantoolsDialog->TabWidget->setTabEnabled(scantoolsDialog->tab_7,false);
		scantoolsDialog->TabWidget->setTabEnabled(scantoolsDialog->tab_8,false);
		scantoolsDialog->TabWidget->setTabEnabled(scantoolsDialog->tab_9, false);
		scantoolsDialog->TabWidget->setTabEnabled(scantoolsDialog->tab_10, false);
		// disable buttons etc
		scantoolsDialog->SelectionBox->setEnabled(true);
		scantoolsDialog->SelectAllButton->setEnabled(true);
		scantoolsDialog->DeselectButton->setEnabled(true);
		// set apropriate page
		scantoolsDialog->TabWidget->setCurrentPage(5);
	} else if (id == outlierID) {
		toolButton->setIconSet(outlierSet);
		// enable / disable tabs
		scantoolsDialog->TabWidget->setTabEnabled(scantoolsDialog->tab,true);
		scantoolsDialog->TabWidget->setTabEnabled(scantoolsDialog->tab_2,false);
		scantoolsDialog->TabWidget->setTabEnabled(scantoolsDialog->tab_3,false);
		scantoolsDialog->TabWidget->setTabEnabled(scantoolsDialog->tab_4,false);
		scantoolsDialog->TabWidget->setTabEnabled(scantoolsDialog->tab_5,false);
		scantoolsDialog->TabWidget->setTabEnabled(scantoolsDialog->tab_6,false);
		scantoolsDialog->TabWidget->setTabEnabled(scantoolsDialog->tab_7,true);
		scantoolsDialog->TabWidget->setTabEnabled(scantoolsDialog->tab_8,false);
		scantoolsDialog->TabWidget->setTabEnabled(scantoolsDialog->tab_9, false);
		scantoolsDialog->TabWidget->setTabEnabled(scantoolsDialog->tab_10, false);
		// disable buttons etc
		scantoolsDialog->ALLClassifyButton->setEnabled(false);
		scantoolsDialog->SelectionBox->setEnabled(false);
		scantoolsDialog->SelectAllButton->setEnabled(false);
		scantoolsDialog->DeselectButton->setEnabled(false);
		// set apropriate page
		scantoolsDialog->TabWidget->setCurrentPage(6);
		// rebuild tree
		//handleBuildButtonReleased();
	} else if (id == normalsID) {
		toolButton->setIconSet(normalsSet);
		// enable / disable tabs
		scantoolsDialog->TabWidget->setTabEnabled(scantoolsDialog->tab,true);
		scantoolsDialog->TabWidget->setTabEnabled(scantoolsDialog->tab_2,false);
		scantoolsDialog->TabWidget->setTabEnabled(scantoolsDialog->tab_3,false);
		scantoolsDialog->TabWidget->setTabEnabled(scantoolsDialog->tab_4,false);
		scantoolsDialog->TabWidget->setTabEnabled(scantoolsDialog->tab_5,false);
		scantoolsDialog->TabWidget->setTabEnabled(scantoolsDialog->tab_6,false);
		scantoolsDialog->TabWidget->setTabEnabled(scantoolsDialog->tab_7,false);
		scantoolsDialog->TabWidget->setTabEnabled(scantoolsDialog->tab_8,true);
		scantoolsDialog->TabWidget->setTabEnabled(scantoolsDialog->tab_9, false);
		scantoolsDialog->TabWidget->setTabEnabled(scantoolsDialog->tab_10, false);
		// disable buttons etc
		scantoolsDialog->SelectionBox->setEnabled(true);
		scantoolsDialog->SelectAllButton->setEnabled(true);
		scantoolsDialog->DeselectButton->setEnabled(true);
		// set apropriate page
		scantoolsDialog->TabWidget->setCurrentPage(7);
	} else if (id == resizeID) {
		toolButton->setIconSet(resizeSet);
		// enable / disable tabs
		scantoolsDialog->TabWidget->setTabEnabled(scantoolsDialog->tab,false);
		scantoolsDialog->TabWidget->setTabEnabled(scantoolsDialog->tab_2,false);
		scantoolsDialog->TabWidget->setTabEnabled(scantoolsDialog->tab_3,false);
		scantoolsDialog->TabWidget->setTabEnabled(scantoolsDialog->tab_4,false);
		scantoolsDialog->TabWidget->setTabEnabled(scantoolsDialog->tab_5,false);
		scantoolsDialog->TabWidget->setTabEnabled(scantoolsDialog->tab_6,false);
		scantoolsDialog->TabWidget->setTabEnabled(scantoolsDialog->tab_7,false);
		scantoolsDialog->TabWidget->setTabEnabled(scantoolsDialog->tab_8,false);
		scantoolsDialog->TabWidget->setTabEnabled(scantoolsDialog->tab_9, true);
		scantoolsDialog->TabWidget->setTabEnabled(scantoolsDialog->tab_10, false);
		// set apropriate page
		scantoolsDialog->TabWidget->setCurrentPage(8);
	} else if (id == colorID) {
		toolButton->setIconSet(colorSet);
		// enable / disable tabs
		scantoolsDialog->TabWidget->setTabEnabled(scantoolsDialog->tab,true);
		scantoolsDialog->TabWidget->setTabEnabled(scantoolsDialog->tab_2,false);
		scantoolsDialog->TabWidget->setTabEnabled(scantoolsDialog->tab_3,false);
		scantoolsDialog->TabWidget->setTabEnabled(scantoolsDialog->tab_4,false);
		scantoolsDialog->TabWidget->setTabEnabled(scantoolsDialog->tab_5,false);
		scantoolsDialog->TabWidget->setTabEnabled(scantoolsDialog->tab_6,false);
		scantoolsDialog->TabWidget->setTabEnabled(scantoolsDialog->tab_7,false);
		scantoolsDialog->TabWidget->setTabEnabled(scantoolsDialog->tab_8,false);
		scantoolsDialog->TabWidget->setTabEnabled(scantoolsDialog->tab_9, false);
		scantoolsDialog->TabWidget->setTabEnabled(scantoolsDialog->tab_10, true);
		// disable buttons etc
		scantoolsDialog->DoColorButton->setEnabled(false);
		scantoolsDialog->SelectionBox->setEnabled(false);
		scantoolsDialog->SelectAllButton->setEnabled(false);
		scantoolsDialog->DeselectButton->setEnabled(false);
		// set apropriate page
		scantoolsDialog->TabWidget->setCurrentPage(9);
	}


}

void ScanTools::handleObjectAdded() {
	updateObjects();
}

void ScanTools::handleObjectRemoved() {
	updateObjects();
}

void ScanTools::handleShapeBoxClicked(int id) {
	switch (id) {
		case 0:
			box.setShape(SHAPE_BOX);
			break;
		case 1:
			box.setShape(SHAPE_SPHERE);
			break;
	}
}

void ScanTools::handleSelectAllButtonReleased() {
	Selector::getInstance()->selectAll(getCurrentSelectionIndex());
}

void ScanTools::handleDeselectButtonReleased() {
	Selector::getInstance()->deselectIndex(getCurrentSelectionIndex());
}

void ScanTools::handleObjectListSelectionChanged() {
	Selector::getInstance()->removeAllObjects();
	
	for (int i=0; i<objects.count(); i++) 
		if (scantoolsDialog->ObjectList->isSelected(i)) 
			Selector::getInstance()->addObjectToSelection(objects[i]);
}

void ScanTools::handleMouseBoxClicked(int id) {
	switch(id) {
		case 0:
			box.setMouseMode(MOUSE_ALIGN);
			break;
		case 1:
			box.setMouseMode(MOUSE_FREEMOVE);
			break;
		case 2:
			box.setMouseMode(MOUSE_PICKING);
			break;
	}
}

void ScanTools::handleDeleteButtonReleased() {
	Deletor deletor;
	deletor.setSelection(getCurrentSelectionIndex());
	undoRedoStack->addUndoAction(deletor.executeUndo());
}

void ScanTools::handleRandomNoiseButtonReleased() {
	RandomNoise noise;
	bool isnumber;

	noise.setSelection(getCurrentSelectionIndex());
	noise.setParameters(scantoolsDialog->RandomNoiseEdit->text().toFloat(&isnumber));
	if (isnumber)
		undoRedoStack->addUndoAction(noise.executeUndo());
}

void ScanTools::handleMLSButtonReleased() {
	MovingLeastSquaresSmoother mls;
	bool isnumber1, isnumber2, isnumber3, isnumber4;

	mls.setSelection(getCurrentSelectionIndex());
	mls.setParameters( scantoolsDialog->MLSnEdit->text().toFloat(&isnumber1),
		scantoolsDialog->MLSgaussEdit->text().toFloat(&isnumber2),
		scantoolsDialog->MLSalphaEdit->text().toFloat(&isnumber3),
		scantoolsDialog->MLSdoplaneBox->isChecked(),
		scantoolsDialog->MLSdegreeEdit->text().toFloat(&isnumber4));
	if (isnumber1 && isnumber2 && isnumber3)
		undoRedoStack->addUndoAction(mls.executeUndo());
}

void ScanTools::handlePerformParticleButtonReleased() {
	LocalParticleSimulation lps;
	bool i1,i2,i3,i4,i5,i6;

	lps.setSelection(getCurrentSelectionIndex());
	lps.setParameters(scantoolsDialog->BorderEdit->text().toInt(&i1),
		scantoolsDialog->NeighboursEdit->text().toInt(&i2),
		scantoolsDialog->DistanceEdit->text().toFloat(&i3),
		scantoolsDialog->IterationsEdit->text().toInt(&i4),
		scantoolsDialog->ForceScaleEdit->text().toFloat(&i5),
		scantoolsDialog->DampingEdit->text().toFloat(&i6));
	lps.setMLS(scantoolsDialog->UseMLSBox->isChecked(),
		!scantoolsDialog->OnlyPlaneBox->isChecked());
	if (i1 && i2 && i3 && i4 && i5 && i6)
		undoRedoStack->addUndoAction(lps.executeUndo());
}

void ScanTools::handleTabWidgetSelected(const QString& str) {
	if (str == "Remove Outliers") {
		scantoolsDialog->CombinerWeightingBox->setEnabled(false);
		scantoolsDialog->ClassificationBox->setEnabled(false);
		scantoolsDialog->OutlierPreviewButton->setEnabled(false);
		scantoolsDialog->OutlierDeletionButton->setEnabled(false);
		scantoolsDialog->RestoreButton->setEnabled(false);
		scantoolsDialog->ALLClassifyButton->setEnabled(false);
		if (scantoolsDialog->BuildLabel->text() == "Status: graph built") {
			scantoolsDialog->BuildLabel->setText("graph should be rebuilt");
		}
	}

}

void ScanTools::handleSkipRebuildButtonReleased() {
	scantoolsDialog->CombinerWeightingBox->setEnabled(true);
	scantoolsDialog->ClassificationBox->setEnabled(true);
	scantoolsDialog->OutlierPreviewButton->setEnabled(true);
	scantoolsDialog->OutlierDeletionButton->setEnabled(true);
	scantoolsDialog->RestoreButton->setEnabled(true);
	if (currentTool == markID)
		scantoolsDialog->ALLClassifyButton->setEnabled(true);
}

void ScanTools::handleALLClassifyButtonReleased() {
	QString string;
	char buf[255];

	if (!ocl)
		return;

	if (ocl->doMiniballClassification(getCurrentSelectionIndex())) {
		string  = "Miniball, max = ";
		_gcvt(ocl->mbMaxC(), 10, buf);
		string += buf;
		scantoolsDialog->CombinerMBBox->setTitle(string);
	} else
		scantoolsDialog->CombinerMBBox->setTitle("Miniball, error.");

	if (ocl->doPlaneClassification(getCurrentSelectionIndex())) {
		string  = "Planefit, max = ";
		_gcvt(ocl->pfMaxC(), 10, buf);
		string += buf;
		scantoolsDialog->CombinerPFBox->setTitle(string);
	} else
		scantoolsDialog->CombinerPFBox->setTitle("Miniball, error.");

	if (ocl->doBidirectionClassification(getCurrentSelectionIndex())) {
		string  = "Bidirectional, max = ";
		_gcvt(ocl->biMaxC(), 10, buf);
		string += buf;
		scantoolsDialog->CombinerBIBox->setTitle(string);
	} else {
		string = "Miniball, error.";
		scantoolsDialog->CombinerBIBox->setTitle(string);
	}
}

void ScanTools::handleOutlierPreviewButtonReleased() {
	float weight_mb, weight_pf, weight_bi;
	float margin_mb, margin_pf, margin_bi;
	float percentage;
	bool success;

	if (!ocl)
		return;

	weight_mb = ((float)scantoolsDialog->CombinerMethod1Slider->value()) / 100.f - 0.01f;
	weight_pf = ((float)scantoolsDialog->CombinerMethod2Slider->value()) / 100.f - 0.01f;
	weight_bi = ((float)scantoolsDialog->CombinerMethod3Slider->value()) / 100.f - 0.01f;

	if (scantoolsDialog->AutomaticalBox_->isChecked()) {	
		ocl->restoreColor();
		percentage = scantoolsDialog->OutlierPercentage_->text().toFloat(&success) / 100.0f;
		if (success) 
			ocl->colorSurfelsPercentALL(percentage, weight_mb, weight_pf, weight_bi);
	} else {
		margin_mb = scantoolsDialog->CombinerMBLineEdit->text().toFloat(&success);
		if (!success)
			margin_mb = 2.f * ocl->mbMaxC();
		margin_pf = scantoolsDialog->CombinerPFLineEdit->text().toFloat(&success);
		if (!success)
			margin_pf = 2.f * ocl->pfMaxC() * ocl->pfMaxConf();
		margin_bi = scantoolsDialog->CombinerBILineEdit->text().toFloat(&success);
		if (!success)
			margin_bi = 2.f * ocl->biMaxC();

		ocl->restoreColor();
		ocl->colorSurfelsCutoffALL(weight_mb, margin_mb, weight_pf, margin_pf, weight_bi, margin_bi);
	}
}

void ScanTools::handleRestoreButtonReleased() {
	if (!ocl)
		return;

	ocl->restoreColor();
}

void ScanTools::handleOutlierDeletionButtonReleased() {
	float weight_mb, weight_pf, weight_bi;
	float margin_mb, margin_pf, margin_bi;
	float percentage;
	bool success;
	

	if (!ocl)
		return;

	weight_mb = ((float)scantoolsDialog->CombinerMethod1Slider->value()) / 100.f - 0.01f;
	weight_pf = ((float)scantoolsDialog->CombinerMethod2Slider->value()) / 100.f - 0.01f;
	weight_bi = ((float)scantoolsDialog->CombinerMethod3Slider->value()) / 100.f - 0.01f;

	if (scantoolsDialog->AutomaticalBox_->isChecked()) {
		ocl->restoreColor();
		percentage = scantoolsDialog->OutlierPercentage_->text().toFloat(&success) / 100.0f;
		if (success) {
			ocl->restoreColor();
			undoRedoStack->addUndoAction(ocl->deleteSurfelsPercentALL(percentage, weight_mb, weight_pf, weight_bi));
		}
	} else {
		margin_mb = scantoolsDialog->CombinerMBLineEdit->text().toFloat(&success);
		if (!success)
			margin_mb = 2.f * ocl->mbMaxC();
		margin_pf = scantoolsDialog->CombinerPFLineEdit->text().toFloat(&success);
		if (!success)
			margin_pf = 2.f * ocl->pfMaxC() * ocl->pfMaxConf();
		margin_bi = scantoolsDialog->CombinerBILineEdit->text().toFloat(&success);
		if (!success)
			margin_bi = 2.f * ocl->biMaxC();
		
		ocl->restoreColor();
		undoRedoStack->addUndoAction(ocl->deleteSurfelsCutoffALL(weight_mb, margin_mb, weight_pf, margin_pf, weight_bi, margin_bi));
	}
}

void ScanTools::handleBuildButtonReleased() {
	QString string;
	int k;

	if ( ocl) {
		_MCdelete((unsigned int)ocl);
		delete ocl;
	}

	ocl = new OutlierClassificator;
	_MCnew( (unsigned long)ocl, sizeof(ocl), __FILE__, __LINE__);

	k = atoi(scantoolsDialog->kNextEdit->text());
	ocl->setK(k);

	if (ocl->buildKNextGraph(-1)) {
		string = "Status: graph built";	
	} else 
		string  = "Status: graph NOT built [error in buildKNextGraph()]";
	
	scantoolsDialog->BuildLabel->setText(string);

	scantoolsDialog->CombinerWeightingBox->setEnabled(true);
	scantoolsDialog->ClassificationBox->setEnabled(true);
	scantoolsDialog->OutlierPreviewButton->setEnabled(true);
	scantoolsDialog->OutlierDeletionButton->setEnabled(true);
	scantoolsDialog->RestoreButton->setEnabled(true);
	if (currentTool == markID)
		scantoolsDialog->ALLClassifyButton->setEnabled(true);
}

void ScanTools::handleCombinerMethod1SliderMoved(int value) {
	float r;
	float v;

	w1 = ((float)scantoolsDialog->CombinerMethod1Slider->value()) / 10000.0f;
	r = 100.f - w1;
	
	if (w2 + w3 != 0.0f)
		v = (w2) / ((w2) + (w3));
	else
		v = 0.5f;

	w2 = v * r;
	w3 = r - w2;
	
	scantoolsDialog->CombinerMethod2Slider->setValue(w2*10000.f);
	scantoolsDialog->CombinerMethod3Slider->setValue(w3*10000.f);
}

void ScanTools::handleCombinerMethod2SliderMoved(int value) {
	float r;
	float v;

	w2 = ((float) scantoolsDialog->CombinerMethod2Slider->value()) / 10000.f;
	r = 100.f - w2;
	
	if (w1 + w3 != 0.0f)
		v = (w1) / ((w1) + (w3));
	else
		v = 0.5f;

	w1 = v * r;
	w3 = r - w1;
	
	scantoolsDialog->CombinerMethod3Slider->setValue(w3*10000.f);
	scantoolsDialog->CombinerMethod1Slider->setValue(w1*10000.f);
}

void ScanTools::handleCombinerMethod3SliderMoved(int value) {
	float r;
	float v;

	w3 = ((float) scantoolsDialog->CombinerMethod3Slider->value()) / 10000.f;
	r = 100.f - w3;
	
	if (w1 + w2 != 0.0f)
		v = (w1) / ((w1) + (w2));
	else
		v = 0.5f;

	w1 = v * r;
	w2 = r - w1;
	
		
	scantoolsDialog->CombinerMethod1Slider->setValue(w1*10000.f);
	scantoolsDialog->CombinerMethod2Slider->setValue(w2*10000.f);
}


void ScanTools::handleAutomaticalBoxToggled(bool toggle) {
	scantoolsDialog->Automatical->setEnabled(toggle);
	scantoolsDialog->CombinerMBBox->setEnabled(!toggle);
	scantoolsDialog->CombinerPFBox->setEnabled(!toggle);
	scantoolsDialog->CombinerBIBox->setEnabled(!toggle);

}

void ScanTools::handleRemoveFalseNormalsButtonReleased() {
	BadNormalRemover bnr;
	std::vector<unsigned int>* selection;
	bool number;

	selection = Selector::getInstance()->getSelectionIndicesStdVec(Selector::intToFlag(getCurrentSelectionIndex()));
	if (selection->size() == 0)
		return;

	bnr.setMaxCost((float)scantoolsDialog->NormalSlider->value() / 100.0);
	bnr.setStartPoint((*selection)[0]);
	bnr.setNeighbours(scantoolsDialog->NormalsKnext->text().toInt(&number));
	
	if (number)
		undoRedoStack->addUndoAction(bnr.executeUndo());

	_MCdelete((unsigned long)selection);
	delete selection;
}

void ScanTools::handleResizeButtonReleased() {
	SurfelResizer sr;
	bool success;

	sr.setLength(scantoolsDialog->ResizeSize->text().toFloat(&success));
	if (success)
		undoRedoStack->addUndoAction(sr.executeUndo());
}


void ScanTools::handleDoColorButtonReleased() {
	ColorModifier cm;
	QColor diff,spec;
	bool s1,s2,s3;

	if (scantoolsDialog->UniColorButton->isChecked()) 
		cm.setAction(COLOR_UNI);
	else
		cm.setAction(COLOR_FILTER);

	cm.setParamsFilter(scantoolsDialog->KnextEdit->text().toInt(&s1),
		scantoolsDialog->FilterCoefficient->text().toFloat(&s2));
	if (!(s1&&s2))
		return;

	spec.setRgb(255.0*scantoolsDialog->SpecREdit->text().toFloat(&s1),255.0*scantoolsDialog->SpecGEdit->text().toFloat(&s2),255.0*scantoolsDialog->SpecBEdit->text().toFloat(&s3));
	if (! (s1 && s2 && s3))
		return;

	diff.setRgb(255.0*scantoolsDialog->DiffREdit->text( ).toFloat(&s1),255.0*scantoolsDialog->DiffGEdit->text().toFloat(&s2),255.0*scantoolsDialog->DiffBEdit->text().toFloat(&s3));
	if (! (s1 && s2 && s3))
		return;

	cm.setParamsUni(diff.rgb(), spec.rgb());

	undoRedoStack->addUndoAction(cm.executeUndo());
}

void ScanTools::handleMemLeak() {
	MemChecker::getInstance()->DumpUnfreed();
}

// ***************
// private methods
// ***************


void ScanTools::updateObjects() {
	Scene *scene;
	Object *obj;
	int count;

	scantoolsDialog->ObjectList->clear();
	Selector::getInstance()->removeAllObjects();
	
	scene = Scene::getInstance();
	
	objects.clear();
	objects.resize(scene->getNofObjects());
	
	count = 0;
	for (obj = scene->getFirstObject(); obj !=0; obj = scene->getNextObject()) {
		objects.insert(count, obj);
		scantoolsDialog->ObjectList->insertItem(obj->getName(), count);
		scantoolsDialog->ObjectList->setSelected(count,true);
		count++;
	}
	
}

bool ScanTools::checkToolAvailability() const {
	
	// TODO: check here any conditions, e.g.
	//       - if an interactive main renderer is present
	//       - if an Object is in the Scene
	//       - etc.
	return true;

}

void ScanTools::updateWindow() {

}


void ScanTools::showGL(void) {
	rendererGL->registerScanToolSetExtension(paint_box_callback);
	rendererGL->renderFrame();
}

void ScanTools::hideGL(void) {
	rendererGL->unregisterScanToolSetExtension();
	rendererGL->renderFrame();
}


int ScanTools::stateToInt(ButtonState state) {
	int return_int = 0;
	
	if (state&LeftButton) 
		return_int = LEFTBUTTON;
	else if (state&MidButton)
		return_int = MIDBUTTON;
	else if (state&RightButton)
		return_int = RIGHTBUTTON;

	if (state&ControlButton)
		return_int |= CTRLBUTTON;
	if (state&AltButton)
		return_int |= ALTBUTTON;
	if (state&ShiftButton)
		return_int |= SHIFTBUTTON;
	
	return return_int;
}

void ScanTools::connectSelectionOptions() {
	// objects added / removed
	Scene *scene = Scene::getInstance();
	this->connect (scene, SIGNAL (objectAdded()),
		           this, SLOT (handleObjectAdded()));
	this->connect (scene, SIGNAL (objectRemoved()),
		           this, SLOT (handleObjectRemoved()));
	
	// ShapeBox clicked
	this->connect (scantoolsDialog->ShapeBox, SIGNAL (clicked(int)),
		           this, SLOT (handleShapeBoxClicked(int)));
	// SelectAllButton released
	this->connect (scantoolsDialog->SelectAllButton, SIGNAL (released()),
		           this, SLOT (handleSelectAllButtonReleased()));
	// DeselectButton released
	this->connect (scantoolsDialog->DeselectButton, SIGNAL (released()),
		           this, SLOT (handleDeselectButtonReleased()));
	// ObjectList selection changed
	this->connect (scantoolsDialog->ObjectList, SIGNAL (selectionChanged()),
		           this, SLOT (handleObjectListSelectionChanged()));
	// MouseBox clicked
	this->connect (scantoolsDialog->MouseBox, SIGNAL (clicked(int)),
		           this, SLOT (handleMouseBoxClicked(int)));

	this->connect (scantoolsDialog->memleak, SIGNAL(released()),
		this, SLOT(handleMemLeak()));

}

void ScanTools::disconnectSelectionOptions() {
	// objects added / removed
	Scene *scene = Scene::getInstance();
	this->disconnect (scene, SIGNAL (objectAdded()),
		              this, SLOT (handleObjectAdded()));
	this->disconnect (scene, SIGNAL (objectRemoved()),
		              this, SLOT (handleObjectRemoved()));

	// ShapeBox clicked
	this->disconnect (scantoolsDialog->ShapeBox, SIGNAL (clicked(int)),
		           this, SLOT (handleShapeBoxClicked(int)));
	// SelectAllButton released
	this->disconnect (scantoolsDialog->SelectAllButton, SIGNAL (released()),
		           this, SLOT (handleSelectAllButtonReleased()));
	// DeselectButton released
	this->disconnect (scantoolsDialog->DeselectButton, SIGNAL (released()),
		           this, SLOT (handleDeselectButtonReleased()));
	// ObjectList selection changed
	this->disconnect (scantoolsDialog->ObjectList, SIGNAL (selectionChanged()),
		           this, SLOT (handleObjectListSelectionChanged()));
	// MouseBox clicked
	this->disconnect (scantoolsDialog->MouseBox, SIGNAL (clicked(int)),
		           this, SLOT (handleMouseBoxClicked(int)));
}

void ScanTools::connectOther() {
	
	this->connect (scantoolsDialog->DeleteButton, SIGNAL (released()),
		           this, SLOT (handleDeleteButtonReleased()));
	this->connect (scantoolsDialog->RandomNoiseButton, SIGNAL (released()),
		           this, SLOT (handleRandomNoiseButtonReleased()));
	this->connect (scantoolsDialog->MLSButton, SIGNAL (released()),
		           this, SLOT (handleMLSButtonReleased()));
	this->connect (scantoolsDialog->PerformParticleButton, SIGNAL (released()),
		           this, SLOT (handlePerformParticleButtonReleased()));
	this->connect (scantoolsDialog->ALLClassifyButton, SIGNAL (released()),
		           this, SLOT (handleALLClassifyButtonReleased()));
	this->connect (scantoolsDialog->OutlierPreviewButton, SIGNAL (released()),
		           this, SLOT (handleOutlierPreviewButtonReleased()));
	this->connect (scantoolsDialog->RestoreButton, SIGNAL (released()),
		           this, SLOT (handleRestoreButtonReleased()));
	this->connect (scantoolsDialog->OutlierDeletionButton, SIGNAL (released()),
		           this, SLOT (handleOutlierDeletionButtonReleased()));
	this->connect (scantoolsDialog->BuildButton, SIGNAL (released()),
		           this, SLOT (handleBuildButtonReleased()));
	this->connect (scantoolsDialog->CombinerMethod1Slider, SIGNAL (sliderMoved(int)),
		           this, SLOT (handleCombinerMethod1SliderMoved(int)));
	this->connect (scantoolsDialog->CombinerMethod2Slider, SIGNAL (sliderMoved(int)),
		           this, SLOT (handleCombinerMethod2SliderMoved(int)));
	this->connect (scantoolsDialog->CombinerMethod3Slider, SIGNAL (sliderMoved(int)),
		           this, SLOT (handleCombinerMethod3SliderMoved(int)));
	this->connect (scantoolsDialog->RemoveFalseNormalsButton, SIGNAL (released()),
		           this, SLOT (handleRemoveFalseNormalsButtonReleased()));
	this->connect (scantoolsDialog->ResizeButton, SIGNAL (released()),
		           this, SLOT (handleResizeButtonReleased()));
	this->connect (scantoolsDialog->DoColorButton, SIGNAL (released()),
		           this, SLOT (handleDoColorButtonReleased()));
	this->connect (scantoolsDialog->AutomaticalBox_, SIGNAL (toggled(bool)),
		           this, SLOT (handleAutomaticalBoxToggled(bool)));
	this->connect (scantoolsDialog->TabWidget, SIGNAL (selected(const QString&)),
				   this, SLOT (handleTabWidgetSelected(const QString&)));
	this->connect (scantoolsDialog->SkipRebuildButton, SIGNAL (released()),
				   this, SLOT (handleSkipRebuildButtonReleased()));

	// connect help
	this->connect (scantoolsDialog->HelpButtonSelection, SIGNAL (released()),
				   this, SLOT (handleHelpSelection()));
	this->connect (scantoolsDialog->HelpButtonDeletor, SIGNAL (released()),
				   this, SLOT (handleHelpDeletor()));
	this->connect (scantoolsDialog->HelpButtonNoise, SIGNAL (released()),
				   this, SLOT (handleHelpNoise()));
	this->connect (scantoolsDialog->HelpButtonMLS, SIGNAL (released()),
				   this, SLOT (handleHelpMLS()));
	this->connect (scantoolsDialog->HelpButtonScatterer, SIGNAL (released()),
				   this, SLOT (handleHelpScatterer()));
	this->connect (scantoolsDialog->HelpButtonLPS, SIGNAL (released()),
				   this, SLOT (handleHelpLPS()));
	this->connect (scantoolsDialog->HelpButtonOutliers, SIGNAL (released()),
				   this, SLOT (handleHelpOutliers()));
	this->connect (scantoolsDialog->HelpButtonNormals, SIGNAL (released()),
				   this, SLOT (handleHelpNormals()));
	this->connect (scantoolsDialog->HelpButtonResize, SIGNAL (released()),
				   this, SLOT (handleHelpResize()));
	this->connect (scantoolsDialog->HelpButtonRecolor, SIGNAL (released()),
				   this, SLOT (handleHelpRecolor()));
	


	
}

void ScanTools::disconnectOther() {
	
	this->disconnect (scantoolsDialog->DeleteButton, SIGNAL (released()),
		           this, SLOT (handleDeleteButtonReleased()));
	this->disconnect (scantoolsDialog->RandomNoiseButton, SIGNAL (released()),
		           this, SLOT (handleRandomNoiseButtonReleased()));
	this->disconnect (scantoolsDialog->MLSButton, SIGNAL (released()),
		           this, SLOT (handleMLSButtonReleased()));
	this->disconnect (scantoolsDialog->PerformParticleButton, SIGNAL (released()),
		           this, SLOT (handlePerformParticleButtonReleased()));
	this->disconnect (scantoolsDialog->ALLClassifyButton, SIGNAL (released()),
		           this, SLOT (handleALLClassifyButtonReleased()));
	this->disconnect (scantoolsDialog->OutlierPreviewButton, SIGNAL (released()),
		           this, SLOT (handleOutlierPreviewButtonReleased()));
	this->disconnect (scantoolsDialog->RestoreButton, SIGNAL (released()),
		           this, SLOT (handleRestoreButtonReleased()));
	this->disconnect (scantoolsDialog->OutlierDeletionButton, SIGNAL (released()),
		           this, SLOT (handleOutlierDeletionButtonReleased()));
	this->disconnect (scantoolsDialog->BuildButton, SIGNAL (released()),
		           this, SLOT (handleBuildButtonReleased()));
	this->disconnect (scantoolsDialog->CombinerMethod1Slider, SIGNAL (sliderMoved(int)),
		           this, SLOT (handleCombinerMethod1SliderMoved(int)));
	this->disconnect (scantoolsDialog->CombinerMethod2Slider, SIGNAL (sliderMoved(int)),
		           this, SLOT (handleCombinerMethod2SliderMoved(int)));
	this->disconnect (scantoolsDialog->CombinerMethod3Slider, SIGNAL (sliderMoved(int)),
		           this, SLOT (handleCombinerMethod3SliderMoved(int)));
	this->disconnect (scantoolsDialog->RemoveFalseNormalsButton, SIGNAL (released()),
		           this, SLOT (handleRemoveFalseNormalsButtonReleased()));
	this->disconnect (scantoolsDialog->ResizeButton, SIGNAL (released()),
		           this, SLOT (handleResizeButtonReleased()));
	this->disconnect (scantoolsDialog->DoColorButton, SIGNAL (released()),
		           this, SLOT (handleDoColorButtonReleased()));
	this->disconnect (scantoolsDialog->AutomaticalBox_, SIGNAL (toggled(bool)),
		           this, SLOT (handleAutomaticalBoxToggled(bool)));
	this->disconnect (scantoolsDialog->TabWidget, SIGNAL (selected(const QString&)),
				   this, SLOT (handleTabWidgetSelected(const QString&)));
	
	this->disconnect (scantoolsDialog->SkipRebuildButton, SIGNAL (released()),
				   this, SLOT (handleSkipRebuildButtonReleased()));


	// disconnect help
	// connect help
	this->disconnect (scantoolsDialog->HelpButtonSelection, SIGNAL (released()),
				   this, SLOT (handleHelpSelection()));
	this->disconnect (scantoolsDialog->HelpButtonDeletor, SIGNAL (released()),
				   this, SLOT (handleHelpDeletor()));
	this->disconnect (scantoolsDialog->HelpButtonNoise, SIGNAL (released()),
				   this, SLOT (handleHelpNoise()));
	this->disconnect (scantoolsDialog->HelpButtonMLS, SIGNAL (released()),
				   this, SLOT (handleHelpMLS()));
	this->disconnect (scantoolsDialog->HelpButtonScatterer, SIGNAL (released()),
				   this, SLOT (handleHelpScatterer()));
	this->disconnect (scantoolsDialog->HelpButtonLPS, SIGNAL (released()),
				   this, SLOT (handleHelpLPS()));
	this->disconnect (scantoolsDialog->HelpButtonOutliers, SIGNAL (released()),
				   this, SLOT (handleHelpOutliers()));
	this->disconnect (scantoolsDialog->HelpButtonNormals, SIGNAL (released()),
				   this, SLOT (handleHelpNormals()));
	this->disconnect (scantoolsDialog->HelpButtonResize, SIGNAL (released()),
				   this, SLOT (handleHelpResize()));
	this->disconnect (scantoolsDialog->HelpButtonRecolor, SIGNAL (released()),
				   this, SLOT (handleHelpRecolor()));
	
}


int ScanTools::getCurrentSelectionIndex() {
	return scantoolsDialog->SelectionBox->id(scantoolsDialog->SelectionBox->selected());
}


/*

	// connect help
	this->connect (scantoolsDialog->HelpButtonSelection, SIGNAL (released()),
				   this, SLOT ());
	this->connect (scantoolsDialog->HelpButtonDeletor, SIGNAL (released()),
				   this, SLOT (handleHelpDeletor(deletorString)));
	this->connect (scantoolsDialog->HelpButtonNoise, SIGNAL (released()),
				   this, SLOT (handleHelpNoise(noiseString)));
	this->connect (scantoolsDialog->HelpButtonMLS, SIGNAL (released()),
				   this, SLOT (handleHelpMLS(mlsString)));
	this->connect (scantoolsDialog->HelpButtonScatterer, SIGNAL (released()),
				   this, SLOT (handleHelpScatterer(scattererString)));
	this->connect (scantoolsDialog->HelpButtonLPS, SIGNAL (released()),
				   this, SLOT (handleHelpLPS(lpsString)));
	this->connect (scantoolsDialog->HelpButtonOutliers, SIGNAL (released()),
				   this, SLOT (handleHelpOutliers(outlierString)));
	this->connect (scantoolsDialog->HelpButtonNormals, SIGNAL (released()),
				   this, SLOT (handleHelpNormals(normalsString)));
	this->connect (scantoolsDialog->HelpButtonResize, SIGNAL (released()),
				   this, SLOT (handleHelpResize(resizeString)));
	this->connect (scantoolsDialog->HelpButtonRecolor, SIGNAL (released()),
				   this, SLOT (handleHelpRecolor(recolorString)));
	
  */


void ScanTools::handleHelpSelection () {
	handleHelp(selectionString);
}

void ScanTools::handleHelpDeletor () {
	handleHelp(deletorString);
}

void ScanTools::handleHelpNoise () {
	handleHelp(noiseString);
}

void ScanTools::handleHelpMLS () {
	handleHelp(mlsString);
}

void ScanTools::handleHelpScatterer () {
	handleHelp(scattererString);
}

void ScanTools::handleHelpLPS () {
	handleHelp(lpsString);

}

void ScanTools::handleHelpOutliers () {
	handleHelp(outlierString);

}

void ScanTools::handleHelpNormals () {
	handleHelp(normalsString);
}

void ScanTools::handleHelpResize () {
	handleHelp(resizeString);
}

void ScanTools::handleHelpRecolor () {
	handleHelp(recolorString);

}
	

void ScanTools::handleHelp(const QString& helpstring) {
	HelpDialogImpl *help = HelpDialogImpl::getInstance();
	help->showHelp(helpstring);
}

// ***************
// private slots
// ***************

void ScanTools::handleToolButtonDialog(void) {
}

// ****************
// access functions
// ****************

ToolInterface *createTool() {
	return new ScanTools();
}

const char *getToolName() {
	return "ScanTools";
}

// Some Emacs-Hints -- please don't remove:
//
//  Local Variables:
//  mode:C++
//  tab-width:4
//  End:
