// Title:   Selector.h
// Created: Thu Sep 25 15:15:19 2003
// Authors: Simon Heinzle, Tim Weyrich
//
// Copyright (c) 2001, 2002, 2003, Computer Graphics Lab, ETH Zurich
//
// This file is part of the Pointshop3D system.
// See http://www.pointshop3d.com/ for more information.
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later
// version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
// 
// You should have received a copy of the GNU Lesser General
// Public License along with this library; if not, write to the
// Free Software Foundation, Inc., 59 Temple Place, Suite 330,
// Boston, MA 02111-1307 USA
// 
// Contact info@pointshop3d.com if any conditions of this
// licensing are not clear to you.
//

#ifndef __SELECTOR_H
#define __SELECTOR_H

#ifdef _WIN32
	// turn off std::map warnings
	# pragma warning(push)
	# pragma warning(disable:4786) // identifier was truncated to 255 char
#endif

#include <qvector.h>
#include <map>
#include <vector>
#include "../../../../Core/Scene/src/Scene.h"
#include "../../../../Core/RendererManager/src/RendererManager.h"
#include "../../../../Renderers/OpenGLRenderer/src/OpenGLRenderer.h"
#include "../../../../Core/DataStructures/src/NeighbourHood.h"
#include "../../../../Core/CoreTools/SelectionTool/src/SelectionTool.h"
#include "Action.h"
#include "MyComparators.h"
#include "BoxSelector.h"

/**
 * This class handles some basic volumetric selection issues like range - selections, 
 * selection of points inside a box etc.
 *
 * @author Simon Heinzle
 * @version 1.0
 */
class Selector {
	
	// surfel related containers
	unsigned int _valid_surfels_count;
	unsigned int _all_surfels_count;
	QVector<SurfelInterface> _all_surfels; 
	Vector3D *_transformed_positions;		// for kd - Tree
	Vector3D *_transformed_normals;
	bool *_valid_surfels;					// mark deleted surfels
	
	// kd - tree
	KdTree *_kdtree;
	
	// object container
	std::map<Object*, tTransformationPair,MyObjectComparator> _selected_objects;	// all selected objects
	std::map<Object*, tTransformationPair,MyObjectComparator>::iterator _current_object_iterator;
	std::map<unsigned int,Object*> _index_bounds;	// which surfel belongs to which object (essential for transformation)
	
	// renderer access pointers
	RendererManager *_render_manager;
	OpenGLRenderer *_render_gl;
	
	// selection tool access
	SelectionTool *_selection_tool;

	// has the object - selection changed?
	bool _object_selection_changed;

	// private constructor/destructor -> singleton pattern
	Selector();
	~Selector();
	
	// **************************
	// * aux. private functions *
	
	/**
	 * This method returns all surfels contained by the sphere with center <code>Vector3D center</code> and 
	 * radius <code>float radius</code>.
	 *
	 * @return pointer to <code>QVector<SurfelInterface></code> containing requested surfels.
	 */
	QVector<SurfelInterface>* getSphere(Vector3D center, float radius);
	
	/**
	 * This method returns all indices to surfels contained by the sphere with center
	 * <code>Vector3D center</code> and radius <code>float radius</code>.
	 * Indices point into <code>QVector<SurfelInterface> _all_surfels</code>.
	 *
	 * @return pointer to <code>std::vector<int></code> containing indices to requested surfels.
	 */
	std::vector<int>* getSphereIndicesStdVec(Vector3D center, float radius);
	
	friend class DeleteUndo;
	friend class InsertUndo;
	
public:
	/**
	 * This method creates an instance of this <code>Selector</code>, if necessary,
	 * and returns it. Only one instance at a time exists (singleton pattern).
	 *
	 * @return an instance of this <code>Selector</code>
	 */
	static Selector* getInstance(void);
	/**
	 * This method deletes the instance of this <code>Selector</code>, if necessary.
	 */
	static void deleteInstance(void);
	/**
	 * This method forces a rebuild of the kd-Tree. (necessary when great amount of surfel changed)
	 */
	void buildTree(void);

	/**
	 * This method invalidates surfels tagged with flag <code>SurfelInterface::Flag invalidate_flag</code>
	 * Necessary when surfels are deleted.
	 */
	void invalidateSurfels(SurfelInterface::Flag invalidate_flag);
	/**
	 * This method invalidates surfel with index <code>int index</code>
	 * Necessary when surfel is deleted.
	 */
	void invalidateSurfel(unsigned int index);
	/**
	 * This method deletes surfels from their corresponding <code>SurfelCollection</code>
	 * Invalidates this surfel too.
	 * NOTE: indices container will be sorted afterwards!!
	 */
	void deleteSurfels(std::vector<unsigned int>& indices);
	/**
	 * This method deletes surfels from their corresponding <code>SurfelCollection</code>
	 * Invalidates this surfel too. 
	 * NOTE: indices container will be sorted afterwards!!
	 *
	 * @return a pointer to <code>UndoAction</code> to reverse the deletion
	 */
	UndoAction* deleteSurfelsUndo(std::vector<unsigned int>& indices);
	/**
	 * This method deletes one <code>SurfelInterface *surfel</code> from it's specified <code>Object *object</code>.
	 * If <code>bool emitsignals</code> is set, RenderFrame & ObjectModified signals are emitted.
	 */
	void deleteSurfelFromObject(SurfelInterface *surfel, Object *object, bool emitsignals);
	/**
	 * This method inserts one <code>SurfelInterface *surfel</code> to it's specified <code>Object *object</code>.
	 * If <code>bool emitsignals</code> is set, RenderFrame & ObjectModified signals are emitted.
	 *
	 * @return pointer to the new inserted surfel.
	 */
	SurfelInterface *insertSurfelToObject(SurfelInterface *surfel, Object *object, bool emitsignals);
	/**
	 * This method checks if surfel with index <code>int index</code> is a valid surfel.
	 *
	 * @return flag to indicate if surfel is valid
	 */
	bool isValid(unsigned int index);
	
	/**
	 * This method performs a sphere selection.
	 * Selects all surfels (of the current selected objects) that are in the sphere with center
	 * <code>Vector3D center</code> and radius <code>float radius</code>.
	 * Found surfels are tagged with <code>int selection_index</code> which must be within [0,2]
	 * If <code>bool reset_selection</code> is set, all surfels with tag <code>int selection_index</code>
	 * are untagged.
	 *
	 * @return number of found surfels.
	 */
	int selectSphere(Vector3D center, float radius, int selection_index, bool reset_selection);
	/**
	 * This method performs a box - selection
	 * Selects all surfels (of the current selected objects) that are contained by the box 
	 * <code>BoxSelectior *boxselector</code>.
	 * Found surfels are tagged with <code>int selection_index</code> which must be within [0,2]
	 * If <code>bool reset_selection</code> is set, all surfels with tag <code>int selection_index</code>
	 * are untagged.
	 *
	 * @return number of found surfels.
	 */
	int selectBox(BoxSelector *boxselector,int selection_index, bool reset_selection);
	/**
	 * This method performs an axis aligned box - selection
	 * Selects all surfels (of the current selected objects) that are contained by the axis aligned box 
	 * spanned by <code>Vector3D lowerbound</code> and <code>Vector3D upperbound</code>.
	 * Note that lowerbound[i] < upperbound[i] for all i=0..2
	 * Found surfels are tagged with <code>int selection_index</code> which must be within [0,2]
	 * If <code>bool reset_selection</code> is set, all surfels with tag <code>int selection_index</code>
	 * are untagged.
	 *
	 * @return number of found surfels.
	 */
	int selectAxisAlignedBox(Vector3D lowerbound, Vector3D upperbound,int selection_index, bool reset_selection);
	/** 
	 * This method selects all surfels.
	 * Selects all surfels (of the current selected objects).
	 * Surfels are tagged with <code>int selection_index</code> which must be within [0,2]
	 *
	 * @return number of found surfels.
	 */
	int selectAll(int selection_index);
	/**
	 * This method deselects all surfels tagged with <code>int selection_index</code> (of the current selected objects)
	 */
	void deselectIndex(int selection_index);
	/**
	 * This method removes tags from all surfels (of the current selected objects).
	 */
	void deselectAll();
	/**
	 * This method returns instantiated <code>SelectionTool</code>
	 *
	 * @return instantiated <code>SelectionTool*</code>
	 */
	SelectionTool* getSelectionTool(void);
	/**
	 * This method converts <code>SurfelInterface::Flag convert_flag</code> to a selection index
	 *
	 * @return <code>int</code> selection index.
	 */
	static int flagToInt(SurfelInterface::Flag convert_flag);
	/**
	 * This method converts <code>int convert_int</code> selection index to <code>SurfelInterface::Flag</code>
	 *
	 * @return <code>SurfelInterface::Flag</code> 
	 */
	static SurfelInterface::Flag intToFlag(int convert_int);
	/**
	 * This method returns a pointer to the current kd-tree.
	 *
	 * @return pointer to current <code>KdTree</code>
	 */
	KdTree* returnKdTree(void);
	
	/**
	 * This method returns all surfels in world space.
	 *
	 * @return pointer to <code>std::vector<Vector3D></code> containing all surfels.
	 */
	std::vector<Vector3D>* getAllPositionsWorldStdVec();
	/**
	 * This method returns selected surfels.
	 * Returns all surfels tagged with <code>SurfelInterface::Flag selection_flag</code>
	 *
	 * @return pointer to <code>QVector<SurfelInterface></code> containing selected surfels.
	 */
	QVector<SurfelInterface>* getSelectionSurfels(SurfelInterface::Flag selection_flag);
	/**
	 * This method returns positions (object space) of selected surfels. 
	 * Returns positions in object space of all surfels tagged with 
	 * <code>SurfelInterface::Flag selection_flag</code>.
	 *
	 * @return pointer to <code>QVector<Vector3D></code> containing positions (object space) of selected surfels
	 */
	QVector<Vector3D>* getSelectionPositions(SurfelInterface::Flag selection_flag);
	/**
	 * This method returns positions (world space) of selected surfels. 
	 * Returns positions in world space of all surfels tagged with 
	 * <code>SurfelInterface::Flag selection_flag</code>.
	 *
	 * @return pointer to <code>QVector<Vector3D></code> containing positions (world space) of selected surfels
	 */
	QVector<Vector3D>* getSelectionPositionsWorld(SurfelInterface::Flag selection_flag);
	/**
	 * This method returns positions (world space) of selected surfels. 
	 * Returns positions in world space of all surfels tagged with 
	 * <code>SurfelInterface::Flag selection_flag</code>.
	 *
	 * @return pointer to <code>std::vector<Vector3D></code> containing positions (world space) of selected surfels
	 */
	std::vector<Vector3D>* getSelectionPositionsWorldStdVec(SurfelInterface::Flag selection_flag);
	/**
	 * This method returns indices into <code>QVector<SurfelInterface> _all_surfels</code> of selected surfels.
	 * Returns indices of all surfels tagged with <code>SurfelInterface::Flag selection_flag</code>.
	 *
	 * @return pointer to <code>QVector<Vector3D>/code> containing indices of selected surfels.
	 */
	QVector<unsigned int>* getSelectionIndices(SurfelInterface::Flag selection_flag);
	/**
	 * This method returns indices into <code>QVector<SurfelInterface> _all_surfels</code> of selected surfels.
	 * Returns indices of all surfels tagged with <code>SurfelInterface::Flag selection_flag</code>.
	 *
	 * @return pointer to <code>std::vector<unsigned int>/code> containing indices of selected surfels.
	 */
	std::vector<unsigned int>* getSelectionIndicesStdVec(SurfelInterface::Flag selection_flag);
	/**
	 * This method returns all selected surfels and returns additionally corresponding objects.
	 * The <code>std::map<unsigned int, Object*> &object_indices</code> identifies which SurfelInterface belongs to which Object. This is essential for undo.
	 * This map contains the boundaries, that means if one element is [i,object], all surfels from the last boundary until i belong to that object.
	 * Returns indices of all surfels tagged with <code>SurfelInterface::Flag selection_flag</code>.
	 */
	void getSelectionSurfelsAndIndexObject(SurfelInterface::Flag selection_flag, std::vector<SurfelInterface*> &selected_surfels, std::map<unsigned int, Object*> &object_indices);
	/**
	 * This method returns all valid indices of surfels
	 *
	 * @return pointer to <code>std::vector<unsigned int></code>, a conversion of the internal 
	 * <code>QVector<SurfelInterface> _all_surfels</code>
	 */
	std::vector<unsigned int>* getAllIndicesStdVec();
	
	/**
	 * This method gives access to surfels.
	 * Note that the returned surfel is in it's object space.
	 *
	 * @return pointer to the surfel at index position <code>int index</code>
	 */
	SurfelInterface* getSurfel(int index);
	/**
	 * This method returns the number of all valid surfels.
	 *
	 * @return number of valid surfels.
	 */
	int getNOfSurfels(void);
	/**
	 * This method returns the position in world space of a surfel.
	 * Calculates the world position of surfel at index position <code>int index</code>
	 * and stores the result in <code>Vector3D& return_vector</code>
	 *
	 * @return true if surfel with index <code>int index</code> exists and is valid; false otherwise
	 */
	bool getSurfelWorldPosition(unsigned int index, Vector3D& return_vector); // should be used instead
	/**
	 * This method returns the position in world space if a surfel
	 * Returns the world position of surfel at index position <code>int index</code>
	 * NOTE: Surfel at <code>int index</code> must be valid, otherwise exception is thrown.
	 *
	 * @return world position of surfel with index <code>int index</code>
	 */
	Vector3D getSurfelWorldPosition(unsigned int index); //!!! should not be used
	/**
	 * This method returns the normal in world space of a surfel.
	 * Calculates the world normal of surfel at index position <code>int index</code>
	 * and stores the result in <code>Vector3D& result</code>
	 *
	 * @return true if surfel with index <code>int index</code> exists and is valid; false otherwise
	 */
	bool getSurfelWorldNormal(unsigned int index, Vector3D &result);
	/** 
	 * This method allows setting the position in world coordinates
	 * Updates the position of the surfel at index position <code>int index</code>, using the
	 * new coordinates <code>Vector3D newposition</code>.
	 * Flag <code>bool repaintRenderer</code> can be set false if a series of surfels is going to be set.
	 */
	void setSurfelWorldPosition(unsigned int index, Vector3D newposition, bool repaintRenderer = true);
	/** 
	 * This method allows setting the normal in world coordinates
	 * Updates the normal of the surfel at index position <code>int index</code>, using the
	 * new normal <code>Vector3D newnormal</code>.
	 * Flag <code>bool repaintRenderer</code> can be set false if a series of surfels is going to be set.
	 */
	void setSurfelWorldNormal(unsigned int index, Vector3D newnormal, bool repaintRenderer = true);
	/**
	 * This method adds all currently available object to this class' object selection.
	 */
	void addAllObjectsToSelection();
	/**
	 * This method adds an object to this class' object selection.
	 * Adds <code>Object *obj</code> to the selection.
	 */
	void addObjectToSelection(Object *obj);
	/**
	 * This method removes an object from this class' object selection.
	 * Removes <code>Object *obj</code> from selection (if selected)
	 */
	void removeObjectFromSelection(Object *obj);
	/**
	 * This method removes all objects from this class' object selection.
	 */
	void removeAllObjects();
	/**
	 * "Iterator" access to the selected objects.
	 * 
	 * @return the first <code>Object*</code> in the object selection.
	 */
	Object* getFirstSelectedObject();
	/**
	 * "Iterator" access to the selected objects.
	 * 
	 * @return the next <code>Object*</code> in the object selection, null
	 */
	Object* getNextSelectedObject();
};

#ifdef _WIN32
	#pragma warning(pop)
#endif // _WIN32
#endif //__SELECTOR_H

// Some Emacs-Hints -- please don't remove:
//
//  Local Variables:
//  mode:C++
//  tab-width:4
//  End:
