// Title:   UndoRedoSpecialisations.h
// Created: Thu Sep 25 15:14:52 2003
// Authors: Simon Heinzle, Tim Weyrich
//
// Copyright (c) 2001, 2002, 2003, Computer Graphics Lab, ETH Zurich
//
// This file is part of the Pointshop3D system.
// See http://www.pointshop3d.com/ for more information.
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later
// version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
// 
// You should have received a copy of the GNU Lesser General
// Public License along with this library; if not, write to the
// Free Software Foundation, Inc., 59 Temple Place, Suite 330,
// Boston, MA 02111-1307 USA
// 
// Contact info@pointshop3d.com if any conditions of this
// licensing are not clear to you.
//

#ifndef __UNDOREDOSPECIALISATION_H
#define __UNDOREDOSPECIALISATION_H

/**
 * This file contains specialised undo/redo's
 *
 * Implemented are:
 *	- surfel deletion undo
 *  - surfel insertion undo
 *  - position & normal change undo
 *  - flag change undo
 *  - diffuse color and specular color change undo
 *
 * @author Simon Heinzle
 * @version 1.0
 */

#ifdef _WIN32
	// turn off std::map warnings
	# pragma warning(push)
	# pragma warning(disable:4786) // identifier was truncated to 255 char
#endif

// some definitions for cross including
class InsertUndo;
class DeleteUndo;
class PositionNormalUndo;
class FlagUndo;
class ColorUndo;

#include <vector>
#include <map>
#include <qvector.h>
#include "../../../../Core/Scene/src/Scene.h"
#include "Action.h"

/**
 * class DeleteUndo
 * defines an undo operation for surfel deletions
 */
class DeleteUndo : public UndoAction {
	friend InsertUndo;

	SurfelCollection _deleted_collection;
	std::vector<SurfelInterface*> _deleted_surfels;
	std::map<unsigned int, Object*> _object_indices;
public:
	/**
	 * Constructor
	 * <code>const std::vector<SurfelInterface*> &deleted_surfels</code> are the surfels which are going to be deleted
	 * <code>const std::map<unsigned int, Object*> &object_indices</code> are the index borders, that means:
	 * Let (x,Obj1) and (y,Obj2) be two succesive pairs. The surfel at index i belongs to Obj2 if x <= i < y.
	 */
	DeleteUndo(const std::vector<SurfelInterface*> &deleted_surfels, const std::map<unsigned int, Object*> &object_indices);
	/**
	 * Constructor
	 * <code>const QVector<SurfelInterface> &deleted_surfels</code> are the surfels which are going to be deleted
	 * <code>const std::map<unsigned int, Object*> &object_indices</code> are the index borders, that means:
	 * Let (x,Obj1) and (y,Obj2) be two succesive pairs. The surfel at index i belongs to Obj2 if x<=i<y.
	 */
	DeleteUndo(const QVector<SurfelInterface> &deleted_surfels, const std::map<unsigned int, Object*> &object_indices);
	/**
	 * Deconstructor
	 */
	virtual ~DeleteUndo();
	/**
	 * Performs the undo.
	 *
	 * @return inverse <code>UndoAction*</code>
	 */
	UndoAction* undo();
};


class InsertUndo : public UndoAction { 
	friend DeleteUndo;

	std::vector<SurfelInterface*> _inserted_surfels;
	std::map<unsigned int, Object*> _object_indices;
public:
	/**
	 * Constructor
	 * <code>const std::vector<SurfelInterface*> &inserted_surfels</code> are the surfels which are going to be inserted
	 * <code>const std::map<unsigned int, Object*> &object_indices</code> are the index borders, that means:
	 * Let (x,Obj1) and (y,Obj2) be two succesive pairs. The surfel at index i belongs to Obj2 if x <= i < y.
	 */
	InsertUndo(const std::vector<SurfelInterface*> &inserted_surfels, const std::map<unsigned int, Object*> &object_indices);
	/**
	 * Deconstructor
	 */
	virtual ~InsertUndo();
	
	/**
	 * Performs the undo.
	 *
	 * @return inverse <code>UndoAction*</code>
	 */
	UndoAction* undo();
};

/**
 * PositionNormalRedo
 */
class PositionNormalUndo : public UndoAction {
	std::vector<SurfelInterface*> _surfels;
	std::vector<Vector3D> _old_positions;
	std::vector<Vector3D> _old_normals;
public:
	/**
	 * Constructor
	 * <code>const std::vector<SurfelInterface*> &surfels</code> are the surfels which are going to be moved
	 * <code>const std::map<unsigned int, Object*> &object_indices</code> are the index borders, that means:
	 * Let (x,Obj1) and (y,Obj2) be two succesive pairs. The surfel at index i belongs to Obj2 if x <= i < y.
	 */
	PositionNormalUndo(const std::vector<SurfelInterface*> &surfels, const std::vector<Vector3D> &positions, const std::vector<Vector3D> &normals);
	/**
	 * Performs the undo.
	 *
	 * @return inverse <code>UndoAction*</code>
	 */
	UndoAction *undo();
};


/**
 * FlagUndo
 */
class FlagUndo : public UndoAction {
	std::vector<SurfelInterface*> _surfels;
	std::vector<int> _flags;
public:
	/**
	 * Constructor
	 * <code>const std::vector<SurfelInterface*> &surfels</code> are the surfels which flags are going to be changed
	 * <code>const std::map<unsigned int, Object*> &object_indices</code> are the index borders, that means:
	 * Let (x,Obj1) and (y,Obj2) be two succesive pairs. The surfel at index i belongs to Obj2 if x <= i < y.
	 */
	FlagUndo(const std::vector<SurfelInterface*> &surfels, const std::vector<int> &flags);
	/**
	 * Adds another <code>FlagUndo</code> to <code>this</code>.
	 */
	void addToThis(FlagUndo *undo);
	/**
	 * Performs the undo.
	 *
	 * @return inverse <code>UndoAction*</code>
	 */
	UndoAction *undo();
};

/**
 * ColorUndo
 * Implements only diffuse color and specular color change, no coefficients change etc
 */
class ColorUndo : public UndoAction {
	std::vector<SurfelInterface*> _surfels;
	std::vector<QRgb> _old_diff;
	std::vector<QRgb> _old_spec;
public:
	/**
	 * Constructor
	 * <code>const std::vector<SurfelInterface*> &surfels</code> are the surfels which colors are going to be changed
	 * <code>const std::map<unsigned int, Object*> &object_indices</code> are the index borders, that means:
	 * Let (x,Obj1) and (y,Obj2) be two succesive pairs. The surfel at index i belongs to Obj2 if x <= i < y.
	 */
	ColorUndo(const std::vector<SurfelInterface*> &surfels,	const std::vector<QRgb>& diffuse, const std::vector<QRgb>& specular);
	/**
	 * Performs the undo.
	 *
	 * @return inverse <code>UndoAction*</code>
	 */
	UndoAction *undo();

};

#ifdef _WIN32
	#pragma warning(pop)
#endif // _WIN32

#endif
// Some Emacs-Hints -- please don't remove:
//
//  Local Variables:
//  mode:C++
//  tab-width:4
//  End:
