// Title:   cQuaternion.h
// Created: Thu Sep 25 15:14:48 2003
// Authors: Simon Heinzle, Tim Weyrich
//
// Copyright (c) 2001, 2002, 2003, Computer Graphics Lab, ETH Zurich
//
// This file is part of the Pointshop3D system.
// See http://www.pointshop3d.com/ for more information.
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later
// version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
// 
// You should have received a copy of the GNU Lesser General
// Public License along with this library; if not, write to the
// Free Software Foundation, Inc., 59 Temple Place, Suite 330,
// Boston, MA 02111-1307 USA
// 
// Contact info@pointshop3d.com if any conditions of this
// licensing are not clear to you.
//

#ifndef __CQUATERNION_H
#define __CQUATERNION_H
/**
 * class cQuaternion
 *
 * implements a basic quaternion data type.
 *
 * @author Simon Heinzle
 * @version 1.0
 */
class cQuaternion {
	float _elements[4];

	// *********************
	// * private functions *
	
	/**
	 * @return three component dot product
	 */
	float dot3(float *a, float *b);
	/**
	 * three component cross product, result stored int <code>float *res</code>
	 */
	void cross3(float *a, float *b, float *res);
	/**
	 * three component scalar multiplication, result stored int <code>float *res</code>
	 */
	void scalmult3(float a, float *b, float *res);
	/**
	 * three component add, result stored int <code>float *res</code>
	 */
	void add3(float *a, float *b, float *res);
	/**
	 * three component normalization
	 */
	static void normalize3(float *a);
	
public:
	/**
	 * Constructor
	 */
	cQuaternion(float x=0, float y=0, float z=0, float c=0);
	/**
	 * adds two quaternions
	 *
	 * @return sum of <code>this</code> and <code>cQuaternion c</code>
	 */
	cQuaternion operator+ (cQuaternion c);
	/**
	 * adds two quaternions
	 *
	 * @return sum of <code>this</code> and <code>cQuaternion c</code>, stored int <code>this</code>
	 */
	cQuaternion operator+= (cQuaternion c);
	/**
	 * multiplies two quaternions
	 *
	 * @return product of <code>this</code> and <code>cQuaternion c</code>
	 */
	cQuaternion operator* (cQuaternion c);
	/**
	 * element access, where x=0,y=1,z=2,c=3.
	 *
	 * @return quaternion element
	 */
	float& operator[] (int index);
	/**
	 * element access, where x=0,y=1,z=2,c=3.
	 *
	 * @return quaternion element
	 */
	const float& operator[] (int index) const;
	/**
	 * gets conjugate version of <code>this</code> quaternion
	 *
	 * @return conjugate of <code>this</code>
	 */
	cQuaternion getConjugate(void);
	/**
	 * conjugate <code>this</code> quaternion
	 */
	void makeConjugate(void);
	/**
	 * rotate <code>this</code> quaternion, angle in radian
	 * rotation axis is passed via <code>float* xyz_axis</code>
	 */
	void rotateRAD(float angle, float* xyz_axis);  
	/**
	 * rotate <code>this</code> quaternion, angle in degree
	 * rotation axis is passed via <code>float* xyz_axis</code>
	 */
	void rotateDEG(float angle, float* xyz_axis);  
	/**
	 * rotate <code>this</code> quaternion, angle in radian
	 * rotation axis is passed via <code>cQuaternion xyz_axis</code>. 
	 * Only xyz of the rotation axis are used.
	 */
	void rotateRAD(float angle, cQuaternion xyz_quat); 
	/**
	 * rotate <code>this</code> quaternion, angle in degree
	 * rotation axis is passed via <code>cQuaternion xyz_axis</code>. 
	 * Only xyz of the rotation axis are used.
	 */
	void rotateDEG(float angle, cQuaternion xyz_quat);  
	/**
	 * access pointer to the element array.
	 *
	 * @return pointer to the first element
	 */
	float* getElements(void);
	/**
	 * gets the rotation matrix by rotating around axis <code>cQuaternion quat</code> by 
	 * <code>float angle</code> in radian.
	 * result is stored in <code>float matrix[4][4]</code>, in column-major order.
	 */
	static void getRotationMatrixRAD(float angle, cQuaternion quat, float matrix[4][4]);
	/**
	 * gets the rotation matrix by rotating around axis <code>cQuaternion quat</code> by 
	 * <code>float angle</code> in degree.
	 * result is stored in <code>float matrix[4][4]</code>, in column-major order.
	 */
	static void getRotationMatrixDEG(float angle, cQuaternion quat, float matrix[4][4]);

};

#endif

// Some Emacs-Hints -- please don't remove:
//
//  Local Variables:
//  mode:C++
//  tab-width:4
//  End:
